#!/bin/bash

SCRIPTDIR=$(pwd)/${0%/bstxml.sh}
FILEBASE=${1%.bib}      #remove .bib extension, if present


### Create the .aux file that bibtex needs
AUXFILE=$FILEBASE.aux
echo "" > $AUXFILE
echo "\citation{*}" >> $AUXFILE
echo "\bibstyle{bstxml}" >> $AUXFILE
echo "\bibdata{$FILEBASE}" >> $AUXFILE

### Preprocess the .bib file to change XML-problematic characters < and &
###  to their entity references
mv $FILEBASE.bib $FILEBASE.bib~orig
sed  -e 's/&/\&amp;/g' \
     -e 's/</\&lt;/g'  \
    $FILEBASE.bib~orig > $FILEBASE.bib

### Run bibtex
BSTINPUTS=$SCRIPTDIR
export BSTINPUTS
#echo "BSTINPUTS is $BSTINPUTS"
bibtex $FILEBASE


### Post-process generated .xml file to remove %-comments
###   that were inserted by bibtex to split long lines
sed -e ':Percent
        /%$/ {
              N
              s/%\n//g
              b Percent
             }'  $FILEBASE.bbl > $FILEBASE.tmp-xml

###The sed line above costs a few hours...  Explanation: (newlines instead ;)
###
#   :Percent             #label to jump
#   /%$/ {               #address: regexp, selects a line trailed by %
#         N              #join with the next line, so that \n is in-text
#         s/%\n//g       #replace trailing %\n by nothing
#         b Percen       #branch to the label, to catch % on the next line
#        }
# The problem is that sed normally matches line-by-line, so the command
# s/%\n//g, by itself, cannot remove \n !!!
###(inspired by the example at http://bama.ua.edu/cgi-bin/man-cgi?sed+1)

echo "<?xml version='1.0' encoding='ISO-8859-1'?>" > $FILEBASE.xml
cat $FILEBASE.tmp-xml >> $FILEBASE.xml

### Cleanup
mv -f $FILEBASE.bib~orig $FILEBASE.bib
rm -f $FILEBASE.{aux,blg,log,bbl,tmp-xml}

