using System;
using System.IO;
using System.Collections;
using System.Collections.Specialized;
using Xtatic;
using Xtatic.Value;

public class Sym {
  public enum Tag : int {
    a = 1,
    b,
    c
  }

  static Hashtable tbl = new Hashtable();
  static Tag cnt = 0;

  public static void InternNew(string s, Tag sym) {
    if (sym <= cnt) throw new SystemException("InternNew: symbol already in use" + sym);
    tbl[s] = sym;
    cnt = sym;
  }

  public static Tag Intern(string s) {
    if (tbl.Contains(s)) return (Tag) tbl[s];
    cnt++;
    Tag tag = (Tag) cnt;
    tbl[s] = tag;
    return tag;
  }

  public static string Name(Tag x) {
    foreach (DictionaryEntry de in tbl) if ((int)de.Value == (int)x) return (string)de.Key;
    throw new SystemException("Sym.Name: unknown tag " + x);
  }

  static Sym() {
    InternNew("a", Tag.a);
    InternNew("b", Tag.b);
    InternNew("c", Tag.c);
  }

}

class Test {

  public static Seq f (Seq x0) {
    if (x0 == null) return null;
    switch (x0.kind) {
      case Seq.Kind.Element: {
        SeqElement x1 = (SeqElement) x0;
        Seq x = x1.contents;
        Seq y = null;                       // should not be initialized. Because of the mono compiler glitch
        Seq tail = null;                    // same as above
        Seq x2 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.a: {
            switch (x2.kind) {
              case Seq.Kind.Element: {
                SeqElement x3 = (SeqElement) x2;
                y = x3.contents;
                tail = x3.next;
                switch (x3.tag) {
                  case Sym.Tag.a: goto same_labels;
                  case Sym.Tag.b: goto diff_labels;
                  default: return null;
                } 
              } default: return null;
            }
          }
          case Sym.Tag.b: {
            switch (x2.kind) {
              case Seq.Kind.Element: {
                SeqElement x3 = (SeqElement) x2;
                y = x3.contents;
                tail = x3.next;
                switch (x3.tag) {
                  case Sym.Tag.a: goto diff_labels;
                  case Sym.Tag.b: goto same_labels;
                  default: return null;
                } 
              } default: return null;
            }
          }
          default: return null;
        };
        diff_labels:
          if (tail != null) return null;
          else return Util.Append(x,(Util.Append(x,Util.Append(x,y))));
        same_labels: 
          if (tail != null) return null;
          else return Util.Append(y,(Util.Append(y,Util.Append(y,x))));
      } default: return null;
    }
  }

  public static void Main () {
    Seq l1 = new SeqInt(1,null);
    Seq l2 = new SeqInt(2,null);
    Seq l3 = new SeqInt(3,null);
    Seq l4 = new SeqInt(4,null);
    Seq l5 = new SeqInt(5,null);
    Seq l6 = new SeqInt(6,null);
    Seq l7 = new SeqInt(7,null);

    Seq x1 = f (null);
    Seq x2 = f (new SeqElement(Sym.Tag.a, l1, new SeqElement(Sym.Tag.b,l2,null)));
    Seq x3 = f (new SeqElement(Sym.Tag.b, l3, new SeqElement(Sym.Tag.a,l4,null)));
    Seq x4 = f (new SeqElement(Sym.Tag.a, l5, new SeqElement(Sym.Tag.a,l6,null)));
    Seq x5 = f (new SeqElement(Sym.Tag.b, l6, new SeqElement(Sym.Tag.b,l7,null)));
    Seq x6 = f (new SeqElement(Sym.Tag.a, l1, new SeqElement(Sym.Tag.b,l2,l1)));
    Seq x7 = f (new SeqElement(Sym.Tag.c, l6, new SeqElement(Sym.Tag.b,l7,null)));

    Console.WriteLine ("x1 = " + Util.ToString(x1));
    Console.WriteLine ("x2 = " + Util.ToString(x2));
    Console.WriteLine ("x3 = " + Util.ToString(x3));
    Console.WriteLine ("x4 = " + Util.ToString(x4));
    Console.WriteLine ("x5 = " + Util.ToString(x5));
    Console.WriteLine ("x6 = " + Util.ToString(x6));
    Console.WriteLine ("x7 = " + Util.ToString(x7));
  }

}                                                                            

