open Printf
open Str

type error =
    Unmatched of string

exception Error of error

let report_error = function
    Unmatched(paren) ->
      printf "unmatched %s" paren

let read_input ip =
  let len = in_channel_length ip in
  let str = String.make len ' ' in
  really_input ip str 0 len;
  len,str

let at pat str start =
  try
    let pos = search_forward (regexp pat) str start in
    pos, String.length (matched_string str) + pos
  with Not_found ->
    -1,0

let preinc r =
  let res = !r in
  incr r;
  res

let _buffer_ = ref ""
let _buffer_pos_ = ref 0

let init_buffer n =
  _buffer_ := String.make n ' '

let into_buffer str pos next =
  for i=pos to next do
    !_buffer_.[preinc _buffer_pos_] <- str.[i]
  done

let get_buffer () = string_before !_buffer_ !_buffer_pos_

let filter ip o_br c_br =
  let len,str = read_input ip in
  init_buffer len;
  let rec outside pos =
    let (o1,o2) = at o_br str pos in
    let (c1,_) = at c_br str pos in
    if o1<0 && c1>=0 then raise (Error(Unmatched(c_br)));
    if o1<0 then into_buffer str pos (len-1)
    else
      let next = inside o2 in
      into_buffer str pos (o1-1);
      outside next
  and inside pos =
    let (c1,c2) = at c_br str pos in
    if c1<0 then raise (Error(Unmatched(o_br)));
    let (o1,o2) = at o_br str pos in
    if o1<0 || c1<o1 then c2
    else inside (inside o2) in
  outside 0;
  print_string (get_buffer ())

let _ =
  let fns = ref [] in
  let usage = "Usage: striptag tag file_name" in
  try
    Arg.parse [] (fun fn -> fns := fn::!fns) usage;      
    match !fns with
      | [fn;tag] ->
          let ip = open_in fn in
          filter ip ("<" ^ tag) ("</" ^ tag ^ ">\\|<\\\\/" ^ tag ^ ">");
          close_in ip
      | _ ->
          raise (Arg.Bad "two parameters required")
  with Arg.Bad(msg) ->
    Arg.usage [] (msg ^ "\n" ^ usage)
