/** Grouper -- similar to assoc list, but collects a "set" (stack) 
    of elements for each distinct key.  */
/** NB: List's keys are strings, and compared structurally. */

class GEntry {
  string key;
  Stack data;
  GEntry next;
  GEntry (string k, Stack d, GEntry n) {
    key = k;  data = d;  next = n; 
  }
}


class Iterator {
  GEntry curr;
  Iterator (GEntry e) {
    curr = e;
  }
  bool isFinished () { return (null == curr); }
  string key()  { return curr.key; }
  Stack data() { return curr.data; }
  Iterator moveNext() { curr = curr.next;  return this;} 
}


class Grouper {
  GEntry fakeTop;

  Grouper(){
    fakeTop = new GEntry("!!FAKE!!", null, null);
  }

  /** Starting from start, finds the entry in the list _previous_ to the entry
     keyed by k.  If k is not in the list, return.next is null.  */
  GEntry findPrevious(GEntry start, string k){
    if (null == start.next) return start;
    else if (k.Equals(start.next.key)) return start;
    else return findPrevious(start.next, k);
  }

  void add(string k, object x) {
    GEntry prev = findPrevious(fakeTop, k);
    if (null == prev.next) 
      prev.next = new GEntry (k, new Stack(), null);
    prev.next.data.push(x); 
  }

  Stack get(string k) {
    GEntry prev = findPrevious(fakeTop, k);
    if (null == prev.next) return null;
    else return (prev.next.data);
  }

  Iterator iterator () {
    return new Iterator(fakeTop.next);
  }

}

