using System;
using System.IO;
using System.Collections;
using System.Collections.Specialized;
using Xtatic;
using Xtatic.Value;

public class Sym {
  public enum Tag : int {
    a = 1,
    b,
    big,
    body,
    br,
    center,
    code,
    div,
    em,
    h1,
    h2,
    h3,
    h4,
    h5,
    h6,
    head,
    hr,
    html,
    i,
    img,
    li,
    ol,
    p,
    pre,
    small,
    strike,
    sub,
    sup,
    title,
    tt,
    ul,
    href,
    name
  }

  static Hashtable tbl = new Hashtable();
  static Tag cnt = 0;

  public static void InternNew(string s, Tag sym) {
    if (sym <= cnt) throw new SystemException("InternNew: symbol already in use" + sym);
    tbl[s] = sym;
    cnt = sym;
  }

  public static Tag Intern(string s) {
    if (tbl.Contains(s)) return (Tag) tbl[s];
    cnt++;
    Tag tag = (Tag) cnt;
    tbl[s] = tag;
    return tag;
  }

  public static string Name(Tag x) {
    foreach (DictionaryEntry de in tbl) if ((int)de.Value == (int)x) return (string)de.Key;
    throw new SystemException("Sym.Name: unknown tag " + x);
  }

  static Sym() {
    InternNew("a", Tag.a);
    InternNew("b", Tag.b);
    InternNew("big", Tag.big);
    InternNew("body", Tag.body);
    InternNew("br", Tag.br);
    InternNew("center", Tag.center);
    InternNew("code", Tag.code);
    InternNew("div", Tag.div);
    InternNew("em", Tag.em);
    InternNew("h1", Tag.h1);
    InternNew("h2", Tag.h2);
    InternNew("h3", Tag.h3);
    InternNew("h4", Tag.h4);
    InternNew("h5", Tag.h5);
    InternNew("h6", Tag.h6);
    InternNew("head", Tag.head);
    InternNew("hr", Tag.hr);
    InternNew("html", Tag.html);
    InternNew("i", Tag.i);
    InternNew("img", Tag.img);
    InternNew("li", Tag.li);
    InternNew("ol", Tag.ol);
    InternNew("p", Tag.p);
    InternNew("pre", Tag.pre);
    InternNew("small", Tag.small);
    InternNew("strike", Tag.strike);
    InternNew("sub", Tag.sub);
    InternNew("sup", Tag.sup);
    InternNew("title", Tag.title);
    InternNew("tt", Tag.tt);
    InternNew("ul", Tag.ul);
    InternNew("href", Tag.href);
    InternNew("name", Tag.name);
  }

}

class Test {
  static int nextanchor; 
  static Seq links;

  private static bool check_pcdata(Seq x0) {
    if (x0 == null) return true;
    switch (x0.kind) {
      case Seq.Kind.Char: return check_pcdata(x0.next);
      default: throw new SystemException("check_pcdata: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    }
  }

  private static bool check_a_content(Seq x0) {
    if (x0 == null) return true;
    switch (x0.kind) {
      case Seq.Kind.Char: return check_a_content(x0.next);
      case Seq.Kind.Element: {
        SeqElement x1 = (SeqElement) x0;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.br:
          case Sym.Tag.img: goto check_empty_Flow;
          case Sym.Tag.b:
          case Sym.Tag.big:
          case Sym.Tag.code:
          case Sym.Tag.em:
          case Sym.Tag.i:
          case Sym.Tag.small:
          case Sym.Tag.strike:
          case Sym.Tag.sub:
          case Sym.Tag.sup:
          case Sym.Tag.tt: goto check_Inline_Flow;
          default: throw new SystemException("check_a_content: wrong tag; " + Sym.Name(x1.tag) + " in " + Util.XmlToString(x1));
        } // switch tag
        check_Inline_Flow: return check_Inline(x2) && check_Flow(x3);
        check_empty_Flow: return x2 == null && check_Flow(x3);
      } //case
      default: throw new SystemException("check_a_content: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    } //switch kind
  }

  private static bool check_pre_content(Seq x0) {
    if (x0 == null) return true;
    switch (x0.kind) {
      case Seq.Kind.Char: return check_pcdata(x0.next);
      case Seq.Kind.Element: {
        SeqElement x1 = (SeqElement) x0;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.a: return x3 == null && check_a_content(x2);
          case Sym.Tag.br: return x2 == null && x3 == null;
          case Sym.Tag.b:
          case Sym.Tag.code:
          case Sym.Tag.em:
          case Sym.Tag.i:
          case Sym.Tag.sub:
          case Sym.Tag.sup:
          case Sym.Tag.tt: goto check_empty_Inline;
          default: throw new SystemException("check_pre_content: wrong tag; " + Sym.Name(x1.tag) + " in " + Util.XmlToString(x1));
        } // switch tag
        check_empty_Inline: return x2 == null && check_Inline(x3);
      } //case
      default: throw new SystemException("check_pre_content: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    } //switch kind
  }

  private static bool check_li_star(Seq x0) {
    if (x0 == null) return true;
    switch (x0.kind) {
      case Seq.Kind.Element: {
        SeqElement x1 = (SeqElement) x0;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.li: return check_Flow(x2) && check_li_star(x3);
          default: throw new SystemException("check_li_star: wrong tag; " + Sym.Name(x1.tag) + " in " + Util.XmlToString(x1));
        } // switch tag
      } //case
      default: throw new SystemException("check_li_star: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    } //switch kind
  }

  private static bool check_Inline(Seq x0) {
    if (x0 == null) return true;
    switch (x0.kind) {
      case Seq.Kind.Char: return check_a_content(x0.next);
      case Seq.Kind.Element: {
        SeqElement x1 = (SeqElement) x0;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.a: return check_a_content(x2) && check_Inline(x3);
          case Sym.Tag.br:
          case Sym.Tag.img: goto check_empty_Inline;
          case Sym.Tag.b:
          case Sym.Tag.big:
          case Sym.Tag.code:
          case Sym.Tag.em:
          case Sym.Tag.i:
          case Sym.Tag.small:
          case Sym.Tag.strike:
          case Sym.Tag.sub:
          case Sym.Tag.sup:
          case Sym.Tag.tt: goto check_Inline_Inline;
          default: throw new SystemException("check_Inline: wrong tag; " + Sym.Name(x1.tag) + " in " + Util.XmlToString(x1));
        } // switch tag
        check_Inline_Inline: return check_Inline(x2) && check_Inline(x3);
        check_empty_Inline: return x2 == null && check_Flow(x3);
      } //case
      default: throw new SystemException("check_Inline: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    } //switch kind
  }

  private static bool check_Flow(Seq x0) {
    if (x0 == null) return true;
    switch (x0.kind) {
      case Seq.Kind.Char: return check_Flow(x0.next);
      case Seq.Kind.Element: {
        SeqElement x1 = (SeqElement) x0;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.a: return (check_a_content(x2) && check_Flow(x3));
          case Sym.Tag.div:
          case Sym.Tag.center: goto check_Flow_Flow;
          case Sym.Tag.b:
          case Sym.Tag.big:
          case Sym.Tag.code:
          case Sym.Tag.em:
          case Sym.Tag.h1:
          case Sym.Tag.h2:
          case Sym.Tag.h3:
          case Sym.Tag.h4:
          case Sym.Tag.h5:
          case Sym.Tag.h6:
          case Sym.Tag.i:
          case Sym.Tag.p:
          case Sym.Tag.small:
          case Sym.Tag.strike:
          case Sym.Tag.sub:
          case Sym.Tag.sup:
          case Sym.Tag.tt: goto check_Inline_Flow;
          case Sym.Tag.pre: return check_pre_content(x2) && check_Flow(x3);
          case Sym.Tag.ol:
          case Sym.Tag.ul: goto check_List_Flow;
          case Sym.Tag.br:
          case Sym.Tag.hr:
          case Sym.Tag.img: goto check_empty_Flow;
          default: throw new SystemException("check_Flow: wrong tag; " + Sym.Name(x1.tag) + " in " + Util.XmlToString(x1));
        }
        check_List_Flow:
          switch (x2.kind) {
            case Seq.Kind.Element: {
              SeqElement x4 = (SeqElement) x2;
              Seq x5 = x4.contents;
              Seq x6 = x4.next;
              switch (x4.tag) {
                case Sym.Tag.li: return check_Flow(x5) && check_li_star(x6) && check_Flow(x3);
                default: throw new SystemException("check_Flow: wrong tag; " + Sym.Name(x4.tag) + " in " + Util.XmlToString(x4));
              }
            } default: throw new SystemException("check_Flow: wrong kind; " + x2.kind + " in " + Util.XmlToString(x2)); 
          }
        check_Inline_Flow: return check_Inline(x2) && check_Flow(x3);
        check_Flow_Flow: return check_Flow(x2) && check_Flow(x3);
        check_empty_Flow: return x2 == null && check_Flow(x3);
      } default: throw new SystemException("check_Flow: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    }
  }

  private static Seq skip_pcdata(Seq x0) {
    if (x0 == null) return null;
    switch (x0.kind) {
      case Seq.Kind.Char: return skip_pcdata(x0.next);
      default: throw new SystemException("check_Flow: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    }
  }

  private static bool check_html(Seq x0) {
    switch (x0.kind) {
      case Seq.Kind.Element:
        SeqElement x1 = (SeqElement) x0;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.html:
            switch (x2.kind) {
              case Seq.Kind.Element:
                SeqElement x4 = (SeqElement) x2;
                Seq x5 = x4.contents;
                Seq x6 = x4.next;
                switch (x4.tag) {
                  case Sym.Tag.head:
                    switch (x5.kind) {
                      case Seq.Kind.Element:
                        SeqElement x7 = (SeqElement) x5;
                        Seq x8 = x7.contents;
                        Seq x9 = x7.next;
                        switch (x7.tag) {
                          case Sym.Tag.title:
                            if (!check_pcdata(x8)) return false;
                            if (x9 != null) return false;
                            break;
                          default: throw new SystemException("check_html1: wrong tag; " + Sym.Name(x7.tag) + " in " + Util.XmlToString(x4));
                        } 
                        break;
                      default: throw new SystemException("check_html2: wrong kind; " + x5.kind + " in " + Util.XmlToString(x5)); 
                    };
                    switch (x6.kind) {
                      case Seq.Kind.Element:
                        SeqElement x7 = (SeqElement) x6;
                        Seq x8 = x7.contents;
                        Seq x9 = x7.next;
                        switch (x7.tag) {
                          case Sym.Tag.body:
                            if (!check_Flow(x8)) return false;
                            if (x9 != null) return false;
                            break;
                          default: throw new SystemException("check_html3: wrong tag; " + Sym.Name(x7.tag) + " in " + Util.XmlToString(x7));
                        } 
                        break;
                      default: throw new SystemException("check_html4: wrong kind; " + x6.kind + " in " + Util.XmlToString(x6)); 
                    }
                    break;
                  default: throw new SystemException("check_html5: wrong tag; " + Sym.Name(x4.tag) + " in " + Util.XmlToString(x4));
                } 
                break;
              default: throw new SystemException("check_html6: wrong kind; " + x2.kind + " in " + Util.XmlToString(x2)); 
            };
            if (x3 != null) return false;
            break;
          default: throw new SystemException("check_html7: wrong tag; " + Sym.Name(x1.tag) + " in " + Util.XmlToString(x1));
        }
        break;
      default: throw new SystemException("check_html8: wrong kind; " + x0.kind + " in " + Util.XmlToString(x0)); 
    }
    return true;
  }

  public static void Main () {
    string file = "pierce";
    Seq doc = Util.LoadXml(file + ".html");
    if (!check_html(doc)) {
      Util.PrintVal(doc);
      throw new SystemException("Main: cast failure 1");
    }
    Seq newdoc = addheaders(doc);
    if (!check_html(newdoc)) {
      Util.PrintVal(newdoc);
      throw new SystemException("Main: cast failure2");
    }
    string newdoc_str = Util.XmlToString(newdoc);
    StreamWriter w = new StreamWriter(file + "-out.html");
    w.WriteLine(newdoc_str);
    w.Close();
    Console.WriteLine (newdoc_str);
 }


  /**********************************************************************/ 
  /** The method encapsulating the main functionality.  */

  public static Seq addheaders (Seq doc) {
    SeqElement x1 = (SeqElement) doc;
    Attributes pageattrs = x1.attributes;
    Seq x6 = x1.contents; // html contents
    Seq header = Util.FixedPrefix(x6,1);
    Seq x9 = x6.next;
    SeqElement x4 = (SeqElement) x9;
    Attributes bodyattrs = x4.attributes;
    Seq body = x4.contents;
    nextanchor = 1;
    links = null;
    Seq newbody = traverse(body);
    SeqElement x7 = new SeqElement(Sym.Tag.br,null,newbody);
    SeqElement x2 = new SeqElement(Sym.Tag.hr,null,x7);
    Seq x3 = Util.Append(links,x2);
    SeqElement x8 = new SeqElement(Sym.Tag.body,x3,null);
    x8.attributes = bodyattrs;
    Seq x5 = Util.Append(header,x8);
    SeqElement res = new SeqElement(Sym.Tag.html,x5,null);
    res.attributes = pageattrs;
    return res;
  }                                                              

  /**********************************************************************/ 
  /** HTML traversal code */
    
  public static Seq traverse (Seq inp) {
//      Console.WriteLine("---");
//      Util.PrintVal(inp);
//      Console.WriteLine();
    if (inp == null) return null;
    switch (inp.kind) {
      case Seq.Kind.Element:
        SeqElement x1 = (SeqElement) inp;
        Seq x2 = x1.contents;
        Seq x3 = x1.next;
        switch (x1.tag) {
          case Sym.Tag.h2: {
            Seq h = x1.contents;
            Seq hd = Util.FixedPrefix(inp,1);
            Seq rest = inp.next;
            string anch = "" + nextanchor;
            string refer = "#" + nextanchor;
            nextanchor++;
            Seq x4 = new SeqChar('\n',null);
            Seq x5 = new SeqElement(Sym.Tag.br,null,x4);
            SeqElement x6 = new SeqElement(Sym.Tag.a,h,x5);
            x6.attributes = Util.CreateAttributes(Sym.Tag.href,refer);
            links = Util.Append(links,x6);
            Seq x7 = traverse(rest);
            Seq x8 = Util.Append(hd,x7);
            SeqElement res = new SeqElement(Sym.Tag.a,null,x8);
            res.attributes = Util.CreateAttributes(Sym.Tag.name,anch);
            return res;
          } default: {
            Sym.Tag tg = x1.tag;
            Attributes atts = x1.attributes;
            Seq cont = x1.contents;
            Seq rest = x1.next;
            Seq x4 = traverse(cont);
            Seq x5 = traverse(rest);
            SeqElement res = new SeqElement(tg,x4,x5);
            res.attributes = atts;
            return res;
          }
        } 
      default: {
        Seq x4 = inp.next;
        Seq rest = skip_pcdata(x4);
        Seq txt = Util.Prefix(inp,rest);
        Seq x5 = traverse(rest);
        Seq res = Util.Append(txt,x5);
        return res;
      }
    }
  }
}                                                                            

