/*  Wrappers for .NET library classes.
    (This is temporary, until support for interfaces and properties
     gets implemented in kerlib.ml. )
     Interfaces cannot be declared -- invent classes instead
     (e.g. see Enum, which is a substitute for IEnumerator).
*/

using System;
using System.Collections;
using System.Text.RegularExpressions;

namespace Poorman {


  public class Enum {
    private IEnumerator real_enum;

    internal Enum(IEnumerator ie) {
      real_enum = ie;
    }

    internal Enum() {}

    public bool MoveNext() {
      return real_enum.MoveNext();
    }

    public object Current() {
      return real_enum.Current;
    }
  }

  public class DictEnum : Enum {
   private IDictionaryEnumerator real_dictenum;

   internal DictEnum(IDictionaryEnumerator e) {
     real_dictenum = e;
   }

    public new bool MoveNext() {
      return real_dictenum.MoveNext();
    }

    public new object Current() {
      return real_dictenum.Current;
    }

    public object Key() {
      return real_dictenum.Key;
    }   

    public object Value() {
      return real_dictenum.Value;
    }
  }

  public class HTbl {
    private Hashtable real_htbl;

    public HTbl () {
      real_htbl = new Hashtable();
    }

    public void Add(object key, object value) {
      real_htbl.Add(key, value);
    }

    public void Remove(object key) {
      real_htbl.Remove(key);
    }

    public bool ContainsKey(object key) {
        return real_htbl.ContainsKey(key);
    }
   
    public bool Contains(object key) {
        return real_htbl.Contains(key);
    }
     public bool ContainsValue (object value) {
        return real_htbl.ContainsValue(value);
    }

    public object Item(object key) {
     return real_htbl[key];
    }

    public DictEnum GetEnumerator() {
      return new DictEnum(real_htbl.GetEnumerator());
    }

    public int Count() {
      return real_htbl.Count;
    }
  }


  public abstract class IComp {    //a la System.Collections.ICompare
    virtual public int Compare(object x, object y) {
      throw new Exception("Poorman.IComp.Compare: abstract method got called");
    }
  }

  public class Comp : IComp {  //a la System.Collections.Compare
    private Comparer c;

    public Comp() {
      c = Comparer.Default;
    }

    override public int Compare (object x, object y) {
      return c.Compare(x,y);
    }
  }

  internal class IComp2IComparer : IComparer {
    private IComp wc;
    internal IComp2IComparer(IComp c) {
      wc = c;
    }
    public int Compare(object x, object y) {
      return wc.Compare(x,y);
    }
  }

  public class ArrList {  //a la System.Collections.ArrayList
    private ArrayList real_arr;

    public ArrList() {
      real_arr = new ArrayList();
    }

    public int Count() {
      return real_arr.Count;
    }

    public int Add(object x) {
      return real_arr.Add(x);
    }

    public Enum GetEnumerator() {
      return new Enum(real_arr.GetEnumerator());
    }

    public void Sort(IComp c) {
      real_arr.Sort(new IComp2IComparer(c));
    }

  public int IndexOf(object x) {
    return real_arr.IndexOf(x);
  }

  public bool Contains(object x) {
    return real_arr.Contains(x);
  }

  }

  public class Date { 
    private System.DateTime real_date; 

    public Date(int year, int month, int day) { 
      real_date = new System.DateTime(year, month, day); 
    }

    private Date(System.DateTime dt) { 
      real_date = dt; 
    }

    public int Year {
      get { return real_date.Year; }
    }    

    public int Month {
      get { return real_date.Month; }
    }    

    public int Day {
      get { return real_date.Day; }
    }    

    public string ToString(string format) { 
      return real_date.ToString(format); 
    }

    public static Date Parse(string datestr, string format) { 
      System.DateTime dt = System.DateTime.ParseExact
        (datestr, format, new System.Globalization.DateTimeFormatInfo());
      return new Date(dt); 
    }
  } 

} //namespace Poorman
