module L = List
module C = Compile
open Il

let stmt t =
  let rec opt e inswitch = match e with
    | None -> None, false
    | Some e ->
        let e, flag = loop e inswitch in
        Some e, flag
  and clauses cls = match cls with
    | [] -> [], false
    | (p,e)::cls ->
        let e,flag1 = loop e true in
        let cls,flag2 = clauses cls in
        (p,e)::cls, flag1 || flag2
  and catch e = match e with
    | IECatch(e1,l,params,e2) ->
        let e1,flag1 = catch e1 in
        let e2,flag2 = loop e2 true in
        IECatch(e1,l,params,e2), flag1 || flag2
    | _ -> loop e true
  and loop e inswitch = match e with
    | IEConst(_)
    | IENew(_,_)
    | IENewArray(_,_,_)
    | IEArrayAccess(_,_)
    | IELocal(_)
    | IEField(_,_)
    | IEUnOp(_,_)
    | IEBinOp(_,_,_)
    | IECast(_,_)
    | IETypeStamp(_,_)
    | IERegexField(_,_,_,_)
    | IEStaticField(_,_) ->
        assert (not inswitch);
        e, false
    | IECall(_,_,_)
    | IEStaticCall(_,_,_)
    | IEAssign(_,_)
    | IEEmpty ->
        if inswitch then C.make_seq e IEBreak, true
        else e, false
    | IEWhile(e1,e2) ->
        let e2,_ = loop e2 false in
        let e = IEWhile(e1,e2) in
        if inswitch then C.make_seq e IEBreak, true
        else e, false
    | IELet(p,e1,e2) ->
        let e1,flag1 = opt e1 false in
        let e2,flag2 = loop e2 inswitch in
        IELet(p,e1,e2), flag1 || flag2
    | IESeq(e1,e2) ->
        let e1,flag1 = loop e1 false in
        let e2,flag2 = loop e2 inswitch in
        IESeq(e1,e2), flag1 || flag2
    | IEIf(e1,e2,e3) ->
        let e2,flag2 = loop e2 inswitch in
        let e3,flag3 = loop e3 inswitch in
        let e = IEIf(e1,e2,e3) in
        if inswitch && (flag2 || flag3) then IESeq(e,IEBreak), true
        else e, false
    | IEProcCall(m,x,params,e1,e2) ->
        let e1,flag1 = loop e1 inswitch in
        let e2,flag2 = opt e2 inswitch in
        let e = IEProcCall(m,x,params,e1,e2) in
        if inswitch && (flag1 || flag2) then IESeq(e,IEBreak), true
        else e, false
    | IECatch(e1,l,params,e2) ->
        let e1,flag1 = catch e1 in
        let e2,flag2 = loop e2 true in
        let e = IECatch(e1,l,params,e2) in
        if inswitch && (flag1 || flag2) then IESeq(e,IEBreak), true
        else e, false
    | IEReturn _
    | IEExit(_,_)
    | IEError _ ->
        e, false
    | IEBreak ->
        e, true
    | IESwitchK(e,cls) ->
        let cls,flag = clauses cls in
        let e = IESwitchK(e,cls) in
        if inswitch && flag then IESeq(e,IEBreak), true
        else e, false
    | IESwitchV(e,cls) ->
        let cls,flag = clauses cls in
        let e = IESwitchV(e,cls) in
        if inswitch && flag then IESeq(e,IEBreak), true
        else e, false
    | IESwitchO(e,cls) ->
        let cls,flag = clauses cls in
        let e = IESwitchO(e,cls) in
        if inswitch && flag then IESeq(e,IEBreak), true
        else e, false
    | IESwitchP(e,cls) ->
        let cls,flag = clauses cls in
        let e = IESwitchP(e,cls) in
        if inswitch && flag then IESeq(e,IEBreak), true
        else e, false
    | IESwitchA(e,cls) ->
        let cls,flag = clauses cls in
        let e = IESwitchA(e,cls) in
        if inswitch && flag then IESeq(e,IEBreak), true
        else e, false
  in loop t false

let meth_decl (n,stat,mods,params,retty,e) =
  (n,stat,mods,params,retty,fst (stmt e))

let clmember = function
  | ICMConstr(i,params,e) -> ICMConstr(i,params,fst (stmt e))
  | ICMStaticConstr(e) -> ICMStaticConstr(fst (stmt e))
  | ICMFld(_,_,_,_) as f -> f
  | ICMMeth(d) -> ICMMeth(meth_decl d)

let clas cl = 
  { cl with iCLMems = L.map clmember cl.iCLMems }

let prog (usedirs,intfs,classes) = (usedirs,intfs,L.map clas classes)
