module SD = SynDenot

(* Compute the length of sequences contained in a given type.  If the
   sequences may have different lengths, return `Variable if the length of the
   sequence cannot be determine, and `None if the type is the empty set
   *)
(*VG?? What is the difference b/w `Variable and `None ?  *)
(*VG?? Why extensible datatype (`A | `B ...) is used here?  *)

let rec sequence_length env sty =
  match SD.seqty' sty with
  | SD.TTree(s, atts, sty1) -> 
      `Fixed 1 
  | SD.TAlt l -> 
      List.fold_left merge_length `None (List.map (sequence_length env) l)
  | SD.TCat l ->
 List.fold_left plus_length (`Fixed 0) (List.map (sequence_length env) l)
  | SD.TRep sty1 ->
   if sequence_length env sty1 == `Fixed 0 then `Fixed 0 else `Variable 
  | SD.TSeqRef i ->
      if List.exists (Ident.same i) env then `Variable else
      sequence_length (i :: env) (Ident.find_same !Store.id_to_seqty i) 
  | SD.PSeqBind (x, sty1) ->
      sequence_length env sty1
  | SD.PTailBind (x, sty1) ->
      sequence_length env sty1

and merge_length l1 l2 =
  match l1,l2 with
   `Fixed i1, `Fixed i2 when i1=i2 -> `Fixed i1 
  | `None, _ -> l2
  | _, `None -> l1
  | _ -> `Variable
      
and plus_length l1 l2 =
  match l1,l2 with
    `Fixed i1, `Fixed i2 -> `Fixed (i1+i2) 
  | `None, _ -> `None
  | _, `None -> `None
  | _ -> `Variable

(* translation from hash patterns to pattern automata *)

type exposed =
    D_Final of Ta.pre
  | D_Bin of Ta.pre * (SD.labty * SD.attsty * SD.seqty) * SD.seqty

module SeqtyListSet =
  Set.Make 
    (struct
      type t = SD.seqty list
      let compare = Sortedlist.compare SD.compare_seqty
    end)

let add_preactions pre es =
  List.map 
    (function 
      |	D_Final pre' -> D_Final (pre @ pre')
      |	D_Bin (pre', (s, a, sty1), sty2) -> 
	  D_Bin (pre @ pre', (s, a, sty1), sty2))
    es

let rec expose' env l =
  if SeqtyListSet.mem l env then [] else
  let env = SeqtyListSet.add l env in
  match l with
    [] -> 
      [D_Final []]
  | sty :: rem -> 
      begin 
	match SD.seqty' sty with
	| SD.TTree(s, a, sty1) -> 
	    [D_Bin ([], (s, a, sty1), SD.cat rem)]
	| SD.TAlt l -> 
	    List.fold_right
	      (fun sty1 es -> 
		Nonduplist.merge (=) (expose' env (sty1 :: rem)) es) 
	      l []
	| SD.TCat l -> 
	    expose' env (l @ rem)
	| SD.TRep sty1 ->
	    Nonduplist.merge (=) (expose' env rem) (expose' env (sty1 :: sty :: rem))
	| SD.TSeqRef i ->		
	    expose' env ((Ident.find_same !Store.id_to_seqty i) :: rem)
	| SD.PSeqBind (x, sty1) ->
(*##VG had to hide this to make 'foreach' work correctly...*)
	    if rem = [] then
	      add_preactions [Ta.Rest x] (expose' env [sty1])
	    else 
(**)
	      begin
	      match sequence_length [] sty1 with
	      | `Fixed i -> 
		  add_preactions [Ta.Fixed (x,i)] 
		    (expose' env (sty1 :: rem))
	      | `None -> [] 
              | `Variable -> 
		  add_preactions [Ta.Begin x]
		    (expose' env [sty1; SD.tailbind x (SD.cat rem)])
	      end
	| SD.PTailBind (x, sty1) ->
	    if rem != [] then assert false else
	    add_preactions [Ta.End x] (expose' env [sty1])
      end

let expose sty = expose' (SeqtyListSet.empty) [sty]

(* pattern automata *)

let rec build_ta sty =
  try
    SD.SeqTbl.find Store.seqty_to_states sty
  with Not_found ->
    let st = Ta.create_state false [] (Some sty) in
    SD.SeqTbl.add Store.seqty_to_states sty st;
    List.iter 
      (function
	| D_Final pre -> 
	    st.Ta.final <- true;
	    st.Ta.finpre <- pre
	| D_Bin (pre, (s, a, sty1), sty2) -> 
	  st.Ta.next <-  
	    (s, a, build_ta sty1, build_ta sty2, pre) :: st.Ta.next)
      (expose sty);
    st

let build_ta x =
  let res = build_ta x in
(*   Format.printf "%a@." SynDenot.pr_seqty x; *)
(*   Format.printf "%a@." Ta.print [res]; *)
  res
