module S = SynSource

type modifier = Sealed

type cmem =  cmem' Finfo.finfo
and cmem' = 
  | CmemFields of (((S.decname * S.exp option) list) * SynOp.status * S.ty)
  | CmemMethod of S.meth
  | CmemConstructor of S.constr


let mkClas' mods clasname parents classmembers = 
  let rec traverse (curr_fields, curr_meths, curr_constrs) = function
    | [] -> 
	(curr_fields, curr_meths, curr_constrs)
    | mem::mems ->
	let (inf,mem') =  mem in
	let next = 
	  match mem' with 
	    | (CmemFields flddec) ->
		let (vardecs, static, typ) = flddec in
		let rev_flds = 
		  List.rev_map 
		    (fun (n,eo) -> (inf, (n, None, eo, static, typ))) 
		    vardecs in
		let next_fields = List.rev_append rev_flds curr_fields
		in (next_fields, curr_meths, curr_constrs)
	    | (CmemMethod methdec) ->
		(curr_fields, methdec::curr_meths, curr_constrs)
	    | (CmemConstructor constrdec) -> 
		(curr_fields, curr_meths, constrdec::curr_constrs)
	in traverse next mems 
  in
  let (fields, methods, constructors) = traverse([],[],[]) classmembers 
  in
  let constructor = 
    match constructors with
      | [] -> (failwith "Error: no constructors in class")
      | _::_::_ -> failwith "More than one constructor in class"
      | [c] -> c
  in 
  {S.clas_name = clasname;
   S.clas_extends = parents;
   S.clas_sealed = 
     (match mods with
     | [Sealed] -> true
     | _ -> false)
   ;
   S.clas_constr = constructor;
   S.clas_fields = fields;
   S.clas_meths = methods;
 }

let mkIntf' intfname implements meth_decls = 
  {S.intf_name = intfname;
   S.intf_implements = implements;
   S.intf_meths = meth_decls;
 }

	  
