module SS = SynSource

(****************** Static contexts **********************************)

type cid = Ident.id     (* class id *)
type fid = Ident.id     (* field id *)
type mid = Ident.id     (* method id *)


(** Classing context provides data about class definitions needed for
typechecking expressions. *)
type clsctx = {
    cc_sealed:     cid -> bool;
    cc_fields:     cid -> (fid * (SynOp.status * SynDenot.ty)) list;
    cc_field:      cid -> fid -> (SynOp.status * SynDenot.ty);
      (** raises Not_found *)
    cc_find_field: cid -> string -> fid * (SynOp.status * SynDenot.ty);
      (** raises Not_found *)
    cc_constrtype: cid -> (SynDenot.ty list);
    cc_methrettype:   cid -> Ident.id -> (SynOp.status * SynOp.method_modifier * SynDenot.ty option);
      (** raises Not_found *)
    cc_meththistype: cid -> string -> (Ident.id * SynOp.status * SynOp.method_modifier * (SynDenot.ty option) * (SynDenot.ty list)) list;
      (** returns the list of method types for this class with name, raises Not_found if cid is not a class *)
    cc_mangled_name: string Ident.tbl;
  (** List of all methods that are implementations from interfaces, 
      both directly declared and their ancestors:
      (string is method name, id is generated by MethIdent.getid *)
    cc_implements: cid -> ((string * (Ident.id * (SynDenot.ty option) * (SynDenot.ty list))) list);
    (* list of interfaces implemented, to restore them when compiling *)
    cc_intf_direct: cid -> cid list
  } 



let extend_mangled_names =
  List.fold_left (
    fun mntbl (mnid, (mn : string)) ->
      try
        let s = Ident.find_same mntbl mnid in
        if s <> mn then failwith "BUG: same method sig associated to different mangled names"
        else mntbl
      with
      | Not_found -> Ident.add mntbl (mnid, mn))


