(************************************************)
(******* Target (C#) language data types *******)
(************************************************)
(*--VG  moved here from il2cs.ml *)

type tagsym = string
type locvarsym = string
type membvarsym = string
type methsym = string
type typsym = string
type nssym = string
type labsym = string

type constval =
  | TCNull
  | TCBool of bool
  | TCInt of int
  | TCChar of char
  | TCString of string
  | TCSymbol of string

type basetyp =
  | TBTEmpty
  | TBTElement
  | TBTObject
  | TBTBool
  | TBTInt
  | TBTChar
  | TBTPcdata

type baseval =
  | TBVBool of bool
  | TBVInt of int
  | TBVChar of char
  | TBVSymbol of string

type pat = 
  | TPType of basetyp
  | TPVal of baseval
  | TPLabel of labsym (* in switch; Michael's suggestion *)
  | TPDefault

type predefty =
  | TPTyVoid
  | TPTyBool
  | TPTyInt
  | TPTyChar
  | TPTyString
  | TPTyArray of typ

and typ =
  | TTYPredef of predefty
  | TTYNamed of typsym

type act_param = bool * exp

and exp = 
  | TEConst of constval
  | TENew of typ * exp list
  | TENewArray of typ * exp * int
  | TEArrayAccess of exp * exp
  | TELocal of locvarsym
  | TEField of exp * membvarsym
  | TEStaticField of typsym * membvarsym
  | TECall of exp * methsym * act_param list
  | TEStaticCall of typsym * methsym * act_param list
  | TEAssign of exp * exp
  | TEUnOp of SynOp.unaryop * exp
  | TEBinOp of SynOp.binaryop * exp * exp
  | TECast of typ * exp
  | TEIs of exp * typ
  | TEType of typ

type block = stmt list

and stmt = 
  | TSBlock of block
  | TSExp of exp
  | TSLocals of (locvarsym * exp option) list * typ
  | TSReturn of exp option
  | TSIf of exp * block * block
  | TSWhile of exp * block
  | TSSwitch of exp * clause list
  | TSBreak
  | TSEmpty
  | TSError of string
  | TSLabel of labsym
  | TSGoto of labsym

and clause = pat * block

type form_param = bool * locvarsym * typ

type invoke = 
  | TINVNone
  | TINVThis of act_param list
  | TINVSuper of act_param list

type clmember =
  | TCMConstr of invoke * form_param list * block
  | TCMStaticConstr of block
  | TCMFld of membvarsym * SynOp.status * typ * exp option
  | TCMMeth of
      methsym * SynOp.status * SynOp.method_modifier * form_param list * typ * block

type clas = {
  tCLHomeNS: string;
  tCLName: typsym;
  tCLExt: typsym;
  tCLImpls: typsym list;
  tCLMems: clmember list;
}

type methdecl = methsym * form_param list * typ
 
type intf = {
  tINHomeNS: string;
  tINName: typsym;
  tINImpls: typsym list;
  tINMems: methdecl list;
}
 
(*--VG-- not used!
type ifmember =
  | TIMFld of membvarsym * typ
  | TIMMeth of methsym * form_param list * typ

type iface = {
  tIFName: typsym;
  tIFExts: typsym list;
  tIFMems: ifmember list;
}
*)
      
type usedir =
  | TUSAlias of nssym * nssym
  | TUSNSpace of nssym

(*--VG--
type nsmember =
  | TNMNspace of nspace
  | TNMClas of clas
  | TNMIf of iface

and nspace = {
  tNSName: nssym;
  tNSUsing: usedir list;
  tNSMems: nsmember list;
}

type cunit = tagsym list * usedir list * nsmember list
*)

type prog = usedir list * intf list * clas list

