
(*
  Goes to synDenot.ml
*)

let mapKerclas_clasty' (clty : HT.clasty) : HT.clasty' = 
  match clty with
  | (HT.TClasRef cid) as clsref -> 
      if (cid == Kerids.object_id) then TObject else
      if (cid == Kerids.int_id) then TInt else
      if (cid == Kerids.bool_id) then TBool else
      if (cid == Kerids.string_id) then TString else
      if (cid == Kerids.char_id) then TChar else
      if (cid == Kerids.tag_id) then TTag else
      if (cid == Kerids.seq_id) then TSeq else
      clsref
  | nonclassref -> nonclassref

(*
  Goes to synSource.ml
  *)

type clstbl = (XtNaming.nctx * clas) Ident.tbl
type intftbl = (XtNaming.nctx * intf) Ident.tbl
type seqtbl = (XtNaming.nctx * seqty) Ident.tbl

type prog = 
    { 
      cls: clstbl;
      intfs: intftbl;
      seq: seqtbl; 
    } 

let pr_clas ff clas = 
  let print_field (name, ty) = 
    fpr ff "@[FLD ";  pr_vardecl ff name;
    fpr ff " : ";   pr_ty ff ty;  
    fpr ff "@]@\n"   in 
  let print_constr cns = 
    fpr ff "@[CONSTR : "; pr_constr ff cns;  fpr ff "@]@\n" in
  let print_meth (name, m) = 
    fpr ff "@[METH "; pr_decname ff name;  fpr ff " : "; 
    pr_meth ff m;  fpr ff "@]@\n" 
  in
  let clas' = Finfo.drop clas in
  fpr ff "extends "; pr_refname ff clas'.clas_extends;  fpr ff " ";
  fpr ff "implements "; List.iter (pr_refname ff) clas'.clas_implements; frp ff " ";
  fpr ff "{@\n";
  fpr ff "  @[";
  List.iter print_field clas'.clas_fields;
  fpr ff "@\n";
  print_constr clas'.clas_constr;
  List.iter print_meth clas'.clas_meths;
  fpr ff "@]}"

let pr_intf ff intf = 
  let print_meth (name, m) = 
    fpr ff "@[METH "; pr_decname ff name;  fpr ff " : "; 
    pr_meth ff m;  fpr ff "@]@\n" 
  in
  let intf' = Finfo.drop intf in
  fpr ff "implements "; List.iter (pr_refname ff) intf'.intf_implements; frp ff " ";
  fpr ff "{@\n";
  fpr ff "  @[";
  List.iter print_meth clas'.clas_meths;
  fpr ff "@]}"

let pr_prog ff prog = 
  let prdecl_clas id (nctx, clas) = 
    fpr ff "@[class %a %a@]@\n" Ident.print id pr_clas clas
  in
  let prdecl_intf id (nctx, intf) = 
    fpr ff "@[interface %a %a@]@\n" Ident.print id pr_intf intf
  in
  let prdecl_seqty id (nctx, seqty) = 
    fpr ff "@[regtype %a %a@]@\n" Ident.print id pr_seqty seqty
  in
  Ident.iter prdecl_seqty prog.seq;
  Ident.iter prdecl_clas prog.cls;
  Ident.iter prdecl_intf prog.intfs
  
(*
 In source.ml, don't forget to change add_cunit
*)
    
(************************************************************************
      Host types and operations on them
************************************************************************)

type clasty' =
  | TClasRef of Ident.id 
  | TObject
  | TBool
  | Tbool of bool
  | TInt
  | Tint of int
  | TChar
  | Tchar of char
  | TString
  | Tstring of string
  | TTag 
  | Ttag of string 
  | TSeq
type clasty = clasty'
  
type intfty' = Ident.id
type intfty = intfty'

type hostty' =
  | Empty
  | Sealed of clasty
  | Clas of clasty * intfty list * clasty list * intfty list
  | Intf of intfty list * clasty list * intfty list
type hostty = hostty'
