type 'a t = (Ident.id * 'a) list
      (** Sorted by id *)

let empty = []

let rec add bind (id, dat) = 
  match bind with 
  | [] -> [(id,dat)]
  | (i,d)::rest -> 
      let c = compare id i in
      if c=0 then raise (Invalid_argument "IdBind.add") else
      if c<0 then (id,dat)::bind
      else (i,d)::(add rest (id,dat))

let fold f accum bind = List.fold_left f accum bind


(** peel: 'a t list -> (Ident.id * 'a list * 'a t list) option  *)
(** Separates the first "layer" of elements with the same id.
    Raises Invalid_argument if non-parallelism is discovered. *)

let peel (binds: 'a t list) : (Ident.id * 'a list * 'a t list) option = 
  (* to be used when 1st bind is empty *)
  let rec peel_nil = function 
    | [] -> ()
    | []::rest -> peel_nil rest
    | (_::_)::_ -> raise
	  (Invalid_argument "IdBind.peel -- empty bind was expected")
  in
  (* to be used when 1st bind is non-empty *)
  let rec peel_cons id = function
    | ((i,d)::brest)::rest -> 
	if not (Ident.same id i) then raise 
	    (Invalid_argument "IdBind.peel -- different ids");
	let (_, heads, tails) = peel_cons id rest in
	(id, d::heads, brest::tails)
    | []::rest -> raise
	  (Invalid_argument "IdBind.peel -- non-empty bind was expected")
    | [] -> (id, [], [])
  in
  match binds with
  | [] -> None
  | []::_ -> peel_nil binds;  None
  | ((id,_)::_)::_ -> Some (peel_cons id binds)


let rec fold_parallel f accum binds = 
  match peel binds with 
  | None -> accum
  | Some(id, dats, rests) -> 
      fold_parallel f (f accum (id,dats)) rests
