(* Borrowed from fragments of Ident module of OCaml-3.01  --VG *)

(* $Id: ident.ml,v 1.13 2004/08/19 16:16:04 milevin Exp $ *)

(*TODO: arrange for some sharing between the code of Tbl and Vis; 
   may be through some common more general implementation *)

type id = { stamp: int; name: string; location: string list option}

let dummyid = {stamp= 0; name= "<DUMMYID>"; location = None}

let currentstamp = ref 0             (* grows upwards, into positives  *)
let currentstamp_predefined = ref 0  (* grows downwards, into negatives *)

let create s =
  incr currentstamp;
  { name = s; 
    location = None;
    stamp = !currentstamp;
  }

let create_located path s = 
  incr currentstamp;
  { name = s;
    location = Some path;
    stamp = !currentstamp; }

let create_internal s =
  decr currentstamp_predefined;
  { name = s ^ (string_of_int !currentstamp_predefined); 
    location = None;
    stamp = !currentstamp_predefined;
  }


let bogus_name = "#$%"

let is_bogus i =
  i.name == bogus_name

let unique_name i =
  if i.name = bogus_name then "bogus" ^ (string_of_int i.stamp)
  else i.name ^ (string_of_int i.stamp)

let name i =
  if i.name = bogus_name then "bogus" ^ (string_of_int i.stamp)
  else i.name

(*let equal i1 i2 = i1.name = i2.name*)

let same i1 i2 = i1.stamp = i2.stamp

let hash id = 
  if id.stamp <= 0 then 2*(-id.stamp)
  else 2*id.stamp + 1


let toString id = 
  match id.stamp with
  | 0 -> assert false
  | n when n > 0 -> 
      Format.sprintf "@[%s@]" id.name 
  | n when n < 0 ->  (* internal id *)
      Format.sprintf "@[%s@]" id.name
  | _ -> assert false

let print ff id = 
  Format.fprintf ff "%s" (toString id)

let assoc_same id lst = 
  match List.find (function (i, _) -> same id i) lst with
  | (_, a) -> a

let assoc_name str lst =
  List.find (function (i, _) -> str = (name i)) lst 


let location id = 
  match id.location with 
    | Some p -> p
    | None -> raise (Bug.Bug("Ident.location", "Asked a location of a non-located id: " ^ (name id)))

let location_longname id = 
  let loc = location id in
  let rec dot = function 
    | [] -> ""
    | [n] -> n
    | n::ns -> n ^ "." ^ (dot ns)
  in dot loc
    
let longname id = 
  let loc_name = location_longname id in
    if loc_name = "" then name id
    else loc_name ^ "." ^ (name id)


let compare (x:id) (y:id) =
  compare x.name y.name


(**** Associative tables *****)

type 'a tbl =
    Empty
  | Node of 'a tbl * 'a data * 'a tbl * int

and 'a data =
  { ident: id;
    data: 'a;
    previous: 'a data option }

let empty = Empty

(* Inline expansion of height for better speed
 * let height = function
 *     Empty -> 0
 *   | Node(_,_,_,h) -> h
 *)

let mknode l d r =
  let hl = match l with Empty -> 0 | Node(_,_,_,h) -> h
  and hr = match r with Empty -> 0 | Node(_,_,_,h) -> h in
  Node(l, d, r, (if hl >= hr then hl + 1 else hr + 1))

let balance l d r =
  let hl = match l with Empty -> 0 | Node(_,_,_,h) -> h
  and hr = match r with Empty -> 0 | Node(_,_,_,h) -> h in
  if hl > hr + 1 then
    match l with
    | Node (ll, ld, lr, _)
      when (match ll with Empty -> 0 | Node(_,_,_,h) -> h) >=
           (match lr with Empty -> 0 | Node(_,_,_,h) -> h) ->
        mknode ll ld (mknode lr d r)
    | Node (ll, ld, Node(lrl, lrd, lrr, _), _) ->
        mknode (mknode ll ld lrl) lrd (mknode lrr d r)
    | _ -> assert false
  else if hr > hl + 1 then
    match r with
    | Node (rl, rd, rr, _)
      when (match rr with Empty -> 0 | Node(_,_,_,h) -> h) >=
           (match rl with Empty -> 0 | Node(_,_,_,h) -> h) ->
        mknode (mknode l d rl) rd rr
    | Node (Node (rll, rld, rlr, _), rd, rr, _) ->
        mknode (mknode l d rll) rld (mknode rlr rd rr)
    | _ -> assert false
  else
    mknode l d r


let rec add tbl (id, data) = 
  match tbl with
    Empty ->
      Node(Empty, {ident = id; data = data; previous = None}, Empty, 1)
  | Node(l, k, r, h) ->
      let c = Pervasives.compare id.name k.ident.name in
      if c = 0 then
        Node(l, {ident = id; data = data; previous = Some k}, r, h)
      else if c < 0 then
        balance (add l  (id, data)) k r
      else
        balance l k (add r (id, data))


let rec add_hide tbl (id, data) = 
  match tbl with
    Empty ->
      Node(Empty, {ident = id; data = data; previous = None}, Empty, 1)
  | Node(l, k, r, h) ->
      let c = Pervasives.compare id.name k.ident.name in
      if c = 0 then
        Node(l, {ident = id; data = data; previous = None}, r, h)
      else if c < 0 then
        balance (add l  (id, data)) k r
      else
        balance l k (add r (id, data))



let rec find_stamp s = function
    None ->
      raise Not_found
  | Some k ->
      if k.ident.stamp = s then k.data else find_stamp s k.previous

let rec find_same tbl id = 
  match tbl with
    Empty ->
      raise Not_found
  | Node(l, k, r, _) ->
      let c = Pervasives.compare id.name k.ident.name in
      if c = 0 then
        if id.stamp = k.ident.stamp
        then k.data
        else find_stamp id.stamp k.previous
      else
        find_same (if c < 0 then l else r) id


let rec find_name tbl name = 
  match tbl with
    Empty ->
      raise Not_found
  | Node(l, k, r, _) ->
      let c = Pervasives.compare name k.ident.name in
      if c = 0 then
        k.data
      else
        find_name (if c < 0 then l else r) name

let mem_id tbl id =
  try
    let _ = find_same tbl id in true
  with
  | Not_found -> false

let rec collect_entries d = 
  (d.ident, d.data) ::
  match d.previous with 
  | None -> []
  | Some p -> collect_entries p

let rec find_all tbl name = 
  match tbl with
    Empty -> []
  | Node(l, k, r, _) ->
      let c = Pervasives.compare name k.ident.name in
      if c = 0 then
        collect_entries k
      else
        find_all (if c < 0 then l else r) name
    

let rec map_data f d = 
  { ident = d.ident;
    data = f d.ident d.data;
    previous = match d.previous with
    | None -> None
    | Some p -> Some(map_data f p) }

let rec map f = function 
  | Empty -> Empty
  | Node(l, k, r, h) ->
      Node(map f l,
	   map_data f k,
	   map f r,
	   h)

(*  val fold: ('a -> (id * 'b) -> 'a) -> 'a -> 'b tbl -> 'a  *)

let rec fold_data f accum d = 
  let d_res = f accum (d.ident, d.data) in
  match d.previous with
  | None -> d_res
  | Some p -> fold_data f d_res p

let rec fold f accum = function
  | Empty -> accum
  | Node(l, k, r, _) -> 
      let l_res   = fold f accum l in
      let lk_res  = fold_data f l_res k in
      let lkr_res = fold f lk_res r in
      lkr_res
    

let rec iter_data f d = 
  f d.ident d.data;
  match d.previous with
  | None -> ()
  | Some p -> iter_data f p

let rec iter f = function 
  | Empty -> ()
  | Node(l, k, r, _) ->
      iter f l;
      iter_data f k;
      iter f r


module Set = Set.Make (
  struct
    type t = id
    let compare = Pervasives.compare
  end
)



module Cache = struct
  type t = {
	    mutable store: unit tbl}

  let empty = {
		 store = empty }

  let getID cache str = 
    match find_all cache.store str with
    | _::_::_ -> assert false
    | [(id, _)] -> id
    | [] -> 
	let newid = create str in
	let _ = cache.store <- add_hide cache.store (newid, ()) in
	newid

  let toList cache = 
    fold (fun lst (id, ()) -> id :: lst) [] cache.store
end
