(* Fragments are borrowed from Ident module of OCaml-3.01  --VG *)
(* See also [Xavier, Modular Module System] for details *)

(* $Id: ident.mli,v 1.12 2004/08/19 16:16:04 milevin Exp $ *)

(* Identifiers (unique names) *)

type id
(* for debugging
type id = { stamp: int; name: string; location: string list option}
*)

val create: string -> id
                     (* used for the user's program identifiers *)

val create_internal: string -> id
       (* used for automatically created ids, concats id number at the end *)

val dummyid: id

val bogus_name: string
val is_bogus : id -> bool
val name: id -> string
val unique_name: id -> string
(*val equal: id -> id -> bool*)
        (* Compare identifiers by name. *)      
val same:  id -> id -> bool
        (* Compare identifiers by binding location.
           Two identifiers are the same if they have the same internal stamp *)
val hash: id -> int
        (* (hash id1) = (hash id2)  iff (same id1 id2)   *)

val assoc_name: string -> (id * 'a) list -> (id * 'a)
val assoc_same: id     -> (id * 'a) list -> 'a

val print: Format.formatter -> id -> unit  
val toString: id -> string


type 'a tbl
        (* Association tables from identifiers to type 'a. *)

val empty: 'a tbl
val add: 'a tbl -> (id * 'a) -> 'a tbl
val add_hide: 'a tbl -> (id * 'a) -> 'a tbl
val find_same: 'a tbl -> id     -> 'a        (* raises Not_found *)
val find_name: 'a tbl -> string -> 'a        (* raises Not_found *)
val find_all:  'a tbl -> string -> (id * 'a) list
val mem_id: 'a tbl -> id -> bool
val map: (id -> 'a -> 'b) -> 'a tbl -> 'b tbl
val fold: ('a -> (id * 'b) -> 'a) -> 'a -> 'b tbl -> 'a
val iter: (id -> 'a -> unit) -> 'a tbl -> unit

val compare: id -> id -> int

module Set: Set.S with type elt = id

module Cache : sig
  type t   (* mutable *)
  val empty : t
  val getID : t -> string -> id
      (* finds id if it exists, or creates a new one and adds it to cache *)
  val toList: t -> id list
end



(*****************************************************************)
(* "Located" identifiers contain the location of their declaration.
    The location is a long name (string list) that refers to the namespace 
    where the idenitifier was introduced.  
    This indentifier variety is to be used only with class and type names.  
    The accessor functions below raise Bug error when used with 
    "non-located" ids.
*)

val create_located : string list -> string -> id
val location : id -> string list
val location_longname : id -> string
val longname: id -> string


(*--VG$$ Some architectural comments for the future: 

  (1) "Located" ids appeared to be the simplest way to implement
  namespace-aware code generation.  However this leads to the
  generated code where all type names are fully-expanded.  We might
  want another solution that would preserve user syntax better.

  (2) Ids for classes and types are significantly different in usage
  from other ids (e.g., for variables) and probably deserve their
  separate ADT, probably closely coupled with that of namespaces.
*)
