(* il2cs: translates Michael's intermediate language to C# abstract syntax *)

module SO = SynOp
module ID = Ident
module C = Compile
module L = List
open Cs

(*--VG  this was moved to Cs.mli  

(************************************************)
(******* Target (C#) language data types *******)
(************************************************)

type tagsym = string
type locvarsym = string
type membvarsym = string
type methsym = string
type typsym = string
type nssym = string
type labsym = string

type constval =
  | TCNull
  | TCBool of bool
  | TCInt of int
  | TCChar of char
  | TCString of string
  | TCSymbol of string

type basetyp =
  | TBTEmpty
  | TBTElement
  | TBTObject
  | TBTBool
  | TBTInt
  | TBTChar
  | TBTPcdata

type baseval =
  | TBVBool of bool
  | TBVInt of int
  | TBVChar of char
  | TBVSymbol of string

type pat = 
  | TPType of basetyp
  | TPVal of baseval
  | TPLabel of labsym (* in switch; Michael's suggestion *)
  | TPDefault

type predefty =
  | TPTyVoid
  | TPTyBool
  | TPTyInt
  | TPTyChar
  | TPTyString
  | TPTyArray of typ

and typ =
  | TTYPredef of predefty
  | TTYNamed of typsym
*)

let ty_seq = TTYNamed "Seq"
let ty_seqstring = TTYNamed "SeqString"
let ty_seqsubstring = TTYNamed "SeqSubstring"
let ty_string = TTYPredef TPTyString
let ty_int = TTYPredef TPTyInt
let ty_object = TTYNamed "object"
let ty_tag = TTYNamed "Tag"
let ty_match = TTYNamed "Match"
let ty_att = TTYNamed "Attributes"
let ty_seqtypestamp = TTYNamed "SeqTypestamp"


(*--VG moved to cs.ml

type act_param = bool * exp

and exp = 
  | TEConst of constval
  | TENew of typ * exp list
  | TENewArray of typ * exp * int
  | TEArrayAccess of exp * exp
  | TELocal of locvarsym
  | TEField of exp * membvarsym
  | TEStaticField of typsym * membvarsym
  | TECall of exp * methsym * act_param list
  | TEStaticCall of typsym * methsym * act_param list
  | TEAssign of exp * exp
  | TEUnOp of SO.unaryop * exp
  | TEBinOp of SO.binaryop * exp * exp
  | TECast of typ * exp
  | TEIs of exp * typ
  | TEType of typ

type block = stmt list

and stmt = 
  | TSBlock of block
  | TSExp of exp
  | TSLocals of (locvarsym * exp option) list * typ
  | TSReturn of exp option
  | TSIf of exp * block * block
  | TSWhile of exp * block
  | TSSwitch of exp * clause list
  | TSBreak
  | TSEmpty
  | TSError of string
  | TSLabel of labsym
  | TSGoto of labsym

and clause = pat * block

type form_param = bool * locvarsym * typ

type invoke = 
  | TINVNone
  | TINVThis of act_param list
  | TINVSuper of act_param list

type clmember =
  | TCMConstr of invoke * form_param list * block
  | TCMStaticConstr of block
  | TCMFld of membvarsym * SO.status * typ * exp option
  | TCMMeth of
      methsym * SO.status * SO.method_modifier * form_param list * typ * block

type clas = {
  tCLHomeNS: string;
  tCLName: typsym;
  tCLExt: typsym;
  tCLImpls: typsym list;
  tCLMems: clmember list;
}

type methdecl = methsym * form_param list * typ
 
type intf = {
  tINHomeNS: string;
  tINName: typsym;
  tINImpls: typsym list;
  tINMems: methdecl list;
}
 
(*--VG-- not used!
type ifmember =
  | TIMFld of membvarsym * typ
  | TIMMeth of methsym * form_param list * typ

type iface = {
  tIFName: typsym;
  tIFExts: typsym list;
  tIFMems: ifmember list;
}
*)
      
type usedir =
  | TUSAlias of nssym * nssym
  | TUSNSpace of nssym

(*--VG--
type nsmember =
  | TNMNspace of nspace
  | TNMClas of clas
  | TNMIf of iface

and nspace = {
  tNSName: nssym;
  tNSUsing: usedir list;
  tNSMems: nsmember list;
}

type cunit = tagsym list * usedir list * nsmember list
*)

type prog = usedir list * intf list * clas list



*)



(********** the translation **********)

let gensym_counter = ref 0
let gensym x =
  incr gensym_counter;
  Printf.sprintf "_il2cs_%s_%d" x !gensym_counter

let translate_constval = function
  | Il.ICNull -> TCNull
  | Il.ICBool b -> TCBool b
  | Il.ICInt i -> TCInt i
  | Il.ICChar c -> TCChar c
  | Il.ICString s -> TCString s
  | Il.ICSymbol s -> TCSymbol (Compile.escape_tag_add_prefix s)

let translate_basetyp = function
  | Il.IBTEmpty -> TBTEmpty
  | Il.IBTElement -> TBTElement
  | Il.IBTObject -> TBTObject
  | Il.IBTBool -> TBTBool
  | Il.IBTInt -> TBTInt
  | Il.IBTChar -> TBTChar

let translate_baseval = function
  | Il.IBVBool x -> TBVBool x
  | Il.IBVInt x -> TBVInt x
  | Il.IBVChar x -> TBVChar x
  | Il.IBVSymbol x -> assert false
  | Il.IBVObj x -> assert false
    (* [XX] something is wrong about the data type *)

let rec translate_predefty = function
  | Il.IPTyVoid -> TPTyVoid
  | Il.IPTyBool -> TPTyBool
  | Il.IPTyInt -> TPTyInt
  | Il.IPTyChar -> TPTyChar
  | Il.IPTyString -> TPTyString
  | Il.IPTyArray t -> TPTyArray (translate_typ t)

and translate_typ t =
  if t == Il.ty_string_pcdata then ty_string
  else match t with
    | Il.ITYPredef x -> TTYPredef(translate_predefty x)
    | Il.ITYNamed x -> TTYNamed x


(*** translation of exp ***)

(* [XX] assumes that exp actually obeys some form about exp and stmt; *)
(* however, I don't know what form it exactly is! *)
(* should refine the type of exp to reflect it... *)

(* something like let-insertion in partial evaluation, *)
(* using ref instead of shift/reset *)
(* cf. Eijiro's paper about online PE! *)
let preactions : stmt list ref = ref []

let make_switch e cls =
  if L.exists (fun (pat,_) -> TPDefault==pat) cls then
    TSSwitch(e,cls)
  else
    TSSwitch(e, cls @ [TPDefault,
                       [TSError "unexpected fall-through in switch"]])

let shift statements =
  preactions := !preactions @ statements

let reset stub =
  let previous_preactions = !preactions in
  preactions := [];
  let result = stub () in (* add preactions as side effect *)
  let current_preactions = !preactions in
  preactions := previous_preactions;
  current_preactions @ result

(* [XX] dirty: should be a parameter of translation functions *)
let handler_environment = ref []

let make_match_res_var r = "mr_" ^ r

let make_pos_var_name x = x ^ "_pos"

let make_len_var_name x = x ^ "_len"

let make_plus e1 e2 =
  match (e1,e2) with
    | TEConst(TCInt 0), e -> e
    | e, TEConst(TCInt 0) -> e
    | _ -> TEBinOp(SO.BiPlus, e1, e2)

let make_minus e1 e2 =
  match e2 with
    | TEConst(TCInt 0) -> e1
    | _ -> TEBinOp(SO.BiMinus, e1, e2)

let make_substring_prefix_args varexp matchlenexp =
  let pcenc = !Pref.pcdataencoding in
  match varexp with
    | Il.IELocal x ->
        if pcenc = "substring" || pcenc = "hybrid" then
          let pos = make_pos_var_name x in
          let posexp = TELocal pos in
          [false,posexp; false,matchlenexp]
        else
          [false,TEConst (TCInt 0); false,matchlenexp]
    | _ -> assert false

let make_string_seq strexp rest =
  let pcenc = !Pref.pcdataencoding in
  let t = if pcenc = "substring" then ty_seqsubstring else ty_seqstring in
  TENew(t, [strexp; rest])

let make_pcdata_seq varexp strexp rest =
  let pcenc = !Pref.pcdataencoding in
  if pcenc = "string" then
    TENew(ty_seqstring, [strexp; rest])
  else
    let x = match varexp with Il.IELocal x -> x | _ -> assert false in
    let pos = make_pos_var_name x in
    let len = make_len_var_name x in
    let posexp = TELocal pos in
    let lenexp = TELocal len in
    TENew(ty_seqsubstring, [strexp; posexp; lenexp; rest])

let make_pcdata_seq_args strexp rest args =
  let pcenc = !Pref.pcdataencoding in
  if pcenc = "string" then
    let strexp = TECall(strexp,"Substring",args) in
    TENew(ty_seqstring, [strexp; rest])
  else
    TENew(ty_seqsubstring, strexp :: (L.map snd args) @ [rest])

let make_coalesce x y rest =
  let pcenc = !Pref.pcdataencoding in
  if pcenc = "substring" || pcenc = "hybrid" then
    let pos = make_pos_var_name y in
    let len = make_len_var_name y in
    let posexp = TELocal pos in
    let lenexp = TELocal len in
    let decls = [TSLocals([pos, Some(TEConst (TCInt 0))], ty_int);
                 TSLocals([len, Some(TEConst (TCInt 0))], ty_int)] in
    let args = [false,TELocal x; true,posexp; true,lenexp; true,rest] in
    decls, TEStaticCall("Utils","Coalesce", args)
  else
    let args = [false,TELocal x; true,rest] in
    [], TEStaticCall("Utils", "Coalesce", args)

let make_match_args x =
  let pcenc = !Pref.pcdataencoding in
  if pcenc = "substring" || pcenc = "hybrid" then
    let pos = make_pos_var_name x in
    let len = make_len_var_name x in
    let posexp = TELocal pos in
    let lenexp = TELocal len in
    [false,TELocal x; false,posexp; false,lenexp]
  else
    [false,TELocal x]

let make_suffix_seq k r e1 e1' rest =
  let pcenc = !Pref.pcdataencoding in
  let bogus = TEConst TCNull in
  let mexp = if k != (-1) then bogus else TELocal (make_match_res_var r) in
  let matchpos = if k != (-1) then TEConst (TCInt k) else TEField(mexp,"Length") in
  if pcenc = "string" then
    let args = [false,matchpos] in
    let strexp = if k==0 then e1' else TECall(e1',"Substring",args) in
    TENew(ty_seqstring, [strexp; rest])
  else
    let x = match e1 with Il.IELocal x -> x | _ -> assert false in
    let pos = make_pos_var_name x in
    let len = make_len_var_name x in
    let posexp = TELocal pos in
    let lenexp = TELocal len in
    let newpos = make_plus posexp matchpos in
    let newlen = make_minus lenexp matchpos in
    TENew(ty_seqsubstring, [e1'; newpos; newlen; rest])

let make_simple_regex_test x str =
  let pcenc = !Pref.pcdataencoding in
  if str = "" then
    if pcenc = "string" then
      TEBinOp(SO.BiEq, TELocal x, TEConst (TCString str))
    else 
      let len = make_len_var_name x in
      TEBinOp(SO.BiLeq, TELocal len, TEConst (TCInt 0))
  else if str = Il.wildcard_char_regex then
    if pcenc = "string" then
      let len = TEField(TELocal x, "Length") in
      TEBinOp(SO.BiGt, len, TEConst (TCInt 0))
    else
      let len = make_len_var_name x in
      TEBinOp(SO.BiGt, TELocal len, TEConst (TCInt 0))
  else if pcenc = "string" then
    let str1 = TELocal x in
    let pos = TEConst (TCInt 0) in
    let len = TEField(str1, "Length") in
    let str2 = TEConst (TCString str) in
    let len2 = TEConst (TCInt (String.length str)) in
    let args = [false,str1; false,pos; false,len; false,str2; false,len2] in
    TEStaticCall("Utils", "PrefixString", args)
  else
    let str1 = TELocal x in
    let pos = TELocal (make_pos_var_name x) in
    let len = TELocal (make_len_var_name x) in
    let str2 = TEConst (TCString str) in
    let len2 = TEConst (TCInt (String.length str)) in
    let args = [false,str1; false,pos; false,len; false,str2; false,len2] in
    TEStaticCall("Utils", "PrefixString", args)

let tag_test_clauses =
  L.for_all (function
               | Il.IPDefault, _
               | Il.IPBaseVal (Il.IBVSymbol _), _ -> true
               | _ -> false)

let rec translate_act_param (b, e) = (b, translate_exp e)

and translate_exp' = function
  | None -> None
  | Some(e) -> Some(translate_exp e)

(* doesn't let-insert "real" expressions, *)
(* assuming that they never have important side-effects; *)
(* as a result, evaluation order may change a little *)
and translate_exp e = match e with
  | Il.IEConst c -> TEConst(translate_constval c)
  | Il.IENew(t, es) -> TENew(translate_typ t, L.map translate_exp es)
  | Il.IENewArray (t, es, d) -> TENewArray (translate_typ t, translate_exp es, d)
  | Il.IEArrayAccess (ar, ind) -> TEArrayAccess (translate_exp ar, translate_exp ind)
  | Il.IELocal x -> TELocal x
  | Il.IEField(e, f) -> TEField(translate_exp e, f)
  | Il.IEStaticField(c, f) -> TEStaticField(c, f)
  | Il.IECall(e, m, xs) ->
      TECall(translate_exp e, m, L.map translate_act_param xs)
  | Il.IEStaticCall(t, m, xs) ->
      TEStaticCall(t, m, L.map translate_act_param xs)
  | Il.IEAssign(e1, e2) ->
      TEAssign(translate_exp e1, translate_exp e2)
  | Il.IEUnOp(op, e) -> TEUnOp(op, translate_exp e)
  | Il.IEBinOp(op, e1, e2) ->
      TEBinOp(op, translate_exp e1, translate_exp e2)
  | Il.IECast(t, e) -> TECast(translate_typ t, translate_exp e)
  | Il.IETypeStamp(x,e) ->
      TENew(ty_seqtypestamp, [TEStaticField("Typestamps",x); translate_exp e])
  | Il.IELet(Il.IPVar(x, t), e1, e2) ->
      (match e1 with
        | None -> ()
        | Some e1 -> shift [TSExp (TEAssign (TELocal x, translate_exp e1))]);
      translate_exp e2
  | Il.IELet(_, None, e2) ->
      translate_exp e2
  | Il.IELet(_, Some e1, e2) ->
      shift (translate_block e1);
      translate_exp e2
  | Il.IESeq(e1, e2) ->
      shift (translate_block e1);
      translate_exp e2
  | Il.IERegexField(r,f,e1,e2) ->
      let issimple,str = C.lookup_regexsym r in
      (match issimple, f with
        | C.ISPAny, "__match__" ->
            let e1' = translate_exp e1 in
            let rest = translate_exp e2 in
            make_pcdata_seq e1 e1' rest
        | C.ISPAny, "__rest__" ->
            if str = "" then
              let e1' = translate_exp e1 in
              let rest = translate_exp e2 in
              make_pcdata_seq e1 e1' rest
            else
              translate_exp e2
        | C.ISPFixed k, "__rest__" ->
            let e1' = translate_exp e1 in
            let rest = translate_exp e2 in
            make_suffix_seq k r e1 e1' rest
        | C.ISPStr, "__rest__" ->
            let k = String.length str in
            let e1' = translate_exp e1 in
            let rest = translate_exp e2 in
            make_suffix_seq k r e1 e1' rest
        | C.ISPComplex, "__match__" ->
            let mexp = TELocal (make_match_res_var r) in
            let args = make_substring_prefix_args e1 (TEField(mexp,"Length")) in
            let strexp = translate_exp e1 in
            let rest = translate_exp e2 in
            make_pcdata_seq_args strexp rest args
        | C.ISPComplex, "__rest__" ->
            let e1' = translate_exp e1 in
            let rest = translate_exp e2 in
            make_suffix_seq (-1) r e1 e1' rest
        | _ -> 
            let mexp = TELocal (make_match_res_var r) in
            let strexp =
              TECall(mexp,"Result",[false,TEConst(TCString ("${" ^ f ^ "}"))]) in
            let rest = translate_exp e2 in
            make_string_seq strexp rest)
  | _ -> assert false (* [XX] refine the type *)

and translate_block = function
  | Il.IELet(Il.IPVar(x, t), e1, e2) ->
      reset (fun () -> (match e1 with
        | None -> []
        | Some e1 -> [TSExp(TEAssign(TELocal x, translate_exp e1))])) @
      translate_block e2
  | Il.IELet(_, None, e2) -> translate_block e2
  | Il.IELet(_, Some e1, e2) -> translate_block e1 @ translate_block e2
  | Il.IESeq(e1, e2) -> translate_block e1 @ translate_block e2
  | Il.IEReturn(e') ->
      reset (fun () ->
        [TSReturn(translate_exp' e')])
  | Il.IEIf(e1, e2, e3) ->
      reset (fun () ->
        [TSIf(translate_exp e1, translate_block e2, translate_block e3)])
  | Il.IEProcCall(m, x, xts, e1, Some(e2)) ->
      reset (fun () ->
          [TSIf(TEStaticCall("Matching", m,
                             (false, TELocal x) ::
                             L.map (fun (x, t) -> (true, TELocal x)) xts),
                translate_block e1,
                translate_block e2)])
  | Il.IEProcCall(m, x, xts, e1, None) ->
      reset (fun () ->
          [TSExp
             (TEStaticCall
                ("Matching", m,
                 (false, TELocal x) ::
                 L.map (fun (x, t) -> (true, TELocal x)) xts))]) @
      translate_block e1
  | Il.IEWhile(e1, e2) ->
      reset (fun () ->
        (* [XXX] assuming that e1 has no "important" side effect;
           otherwise, we should perform it at every iteration *)
        [TSWhile(translate_exp e1, translate_block e2)])
  | Il.IEEmpty -> [TSEmpty] (* or just []? *)
  | Il.IEError(s) -> [TSError(s)]
  | Il.IECatch _ as e -> translate_catch e
  | Il.IEExit(k, es) ->
      let xts = L.assoc k !handler_environment in
      (* [XXX] should eliminate unnecessary temporaries *)
      let (assignments, assignments2) =
        L.fold_right2
          (fun (x, t) e (acc, acc2) ->
            match e with (* eliminate identity assignment *)
            | Il.IELocal(y) when x = y -> (acc, acc2)
            | _ ->
                let tmpvar = gensym ("exit_" ^ x) in
                (reset (fun () ->
                    [TSLocals([(tmpvar, Some(translate_exp e))],
                              translate_typ t)]) @
                 acc),
                 TSExp(TEAssign(TELocal x, TELocal tmpvar)) :: acc2)
          xts es ([], []) in
      assignments @ assignments2 @ [TSGoto k]
  | Il.IESwitchV(e, clauses) ->
      if tag_test_clauses clauses then
        translate_switcht e clauses
      else
        reset (fun () ->
          [make_switch (translate_exp e)
             (L.map
                 (fun (pat, e) ->
                    (match pat with
                      | Il.IPDefault -> TPDefault
                      | Il.IPBaseVal bv -> TPVal(translate_baseval bv)
                      | _ -> assert false (* [XX] refine the type *)),
                    translate_block e)
                 clauses)])
  | Il.IESwitchA(e, clauses) -> translate_switcha e clauses
  | Il.IESwitchP(e, clauses) -> translate_switchp e clauses
  | Il.IESwitchO(e, clauses) -> translate_switcho e clauses
  | Il.IESwitchK(e, clauses) -> translate_switchk e clauses
  | Il.IEBreak -> [TSBreak]
  | (Il.IENew _ | Il.IENewArray _ | Il.IEArrayAccess _ | Il.IERegexField(_,_,_,_) |
     Il.IEField _ | Il.IEStaticField _ |
     Il.IECall _ | Il.IEStaticCall _ | Il.IEAssign _ |
     Il.IEUnOp _ | Il.IEBinOp _ | Il.IECast _ | Il.IETypeStamp(_,_)) as e  ->
      reset (fun () -> [TSExp(translate_exp e)])
  | Il.IEConst _ | Il.IELocal _ ->
      reset (fun () -> [])

and translate_switchp e clauses =
  let tmpvar = ref "DUMMY" in
  let r2l_eval_sucks =
    reset (fun () ->
      match translate_exp e with
      | TELocal x ->
          tmpvar := x;
          []
      | e ->
          tmpvar := gensym "pc";
          [TSLocals([(!tmpvar, Some(e))], ty_att)]) in
  r2l_eval_sucks @
  L.fold_right
    (fun (pat, e) acc -> match pat with
      | Il.IPRegex(r,_) ->
          let issimple,str = C.lookup_regexsym r in
          if issimple != C.ISPComplex then
            match (acc,issimple) with
              | [], _
              | _, C.ISPAny -> translate_block e
              | _ ->
                  let test = make_simple_regex_test !tmpvar str in
                  [TSIf(test, translate_block e, acc)]
          else
            let tmpvar2 = make_match_res_var r in
            let mexp = TELocal tmpvar2 in
            let rexp = TEStaticField("Regexs", r) in
            let args = make_match_args !tmpvar in
            let callexp = TECall(rexp, "Match", args) in
            if acc==[] then
              TSLocals([tmpvar2, Some callexp], ty_match) :: translate_block e
            else
              [TSLocals([tmpvar2, Some callexp],ty_match);
               TSIf(TEField(mexp,"Success"), translate_block e, acc)]
      | Il.IPDefault ->
          translate_block e
      | _ -> assert false)
    clauses
    []

and translate_switcha e clauses =
  let tmpvar = ref "DUMMY" in
  let r2l_eval_sucks =
    reset (fun () ->
      match translate_exp e with
      | TELocal x ->
          tmpvar := x;
          []
      | e ->
          tmpvar := gensym "att";
          [TSLocals([(!tmpvar, Some(e))], ty_att)]) in
  r2l_eval_sucks @
  let (default_clause, non_default_clauses) =
    L.partition
      (function
        | (Il.IPDefault, _) -> true
        | (_, _) -> false)
      clauses in
  let error_stmts = [TSError "non-exhaustive SwitchA"] in
  let default_clause =
    match default_clause with
    | [] -> error_stmts
    | [(Il.IPDefault, c)] -> translate_block c
    | _ -> assert false in
  let clauses =
    L.fold_right
      (fun (pat, e) acc ->
        match pat with
        | Il.IPAtt([], Some Il.IPDontCare) -> acc
        | Il.IPAtt([], Some Il.IPVar(x,_)) ->
            let args = [false, TELocal !tmpvar] in
            let call = TEStaticCall("Utils","AttrsToSeq",args) in
            let asgn = TSExp(TEAssign(TELocal x, call)) in
            asgn :: translate_block e
        | Il.IPAtt(tagaps, None) ->
            let attexp = TELocal !tmpvar in
            let cntfield = TEField(attexp,"Count") in
            let n = L.length tagaps in
            let lenexp = TEConst(TCInt n) in
            let nullcmp = TEBinOp(SO.BiEq,attexp,TEConst TCNull) in
            let restcmp = TEBinOp(SO.BiEq,cntfield,lenexp) in
            let restcmp = if n==0 then TEBinOp(SO.BiOr,nullcmp,restcmp) else restcmp in
            let test =
              L.fold_right
                (fun (tag,pat) acc -> match pat with
                  | Il.IPDontCare ->
                      let tag = C.escape_tag_add_prefix tag in
                      let args = [false,TEConst(TCSymbol tag)] in
                      let call = TECall(TELocal !tmpvar, "Contains", args) in
                      TEBinOp(SO.BiAnd,call,acc)
                  | Il.IPVar(x,t) ->
                      let tag = C.escape_tag_add_prefix tag in
                      let attrval = TEArrayAccess(attexp,TEConst(TCSymbol tag)) in
                      let castexp = TECast(translate_typ t, attrval) in
                      let asgn = TEAssign(TELocal x, castexp) in
                      let cmp = TEBinOp(SO.BiNeq,asgn,TEConst TCNull) in
                      TEBinOp(SO.BiAnd,cmp,acc))
                tagaps
                restcmp
            in [TSIf(test, translate_block e, acc)]
        | Il.IPAtt(tagaps, ap') ->
            let tmpvar2 = gensym "tags" in
            let tmpvar3 = gensym "values" in
            TSLocals
              ([(tmpvar2,
                 Some
                   (TENewArray
                      (ty_tag,
                       TEConst(TCInt(L.length tagaps)),
                       0)))],
               TTYPredef(TPTyArray ty_tag)) ::
            TSLocals
              ([(tmpvar3,
                 Some
                   (TENewArray
                      (ty_seq,
                       TEConst(TCInt(L.length tagaps + 1)),
                       0)))],
               TTYPredef(TPTyArray ty_seq)) ::
            snd
              (L.fold_left
                 (fun (n, acc) (tag, _) ->
                   let tag = C.escape_tag_add_prefix tag in
                   (n + 1,
                    acc @
                    [TSExp
                       (TEAssign
                          (TEArrayAccess
                             (TELocal tmpvar2,
                              TEConst(TCInt n)),
                           TEConst(TCSymbol tag)))]))
                 (0, [])
                 tagaps) @
            let call_switcha =
              TEStaticCall("Utils",
                           "switchA",
                           [(false, TELocal !tmpvar);
                            (false, TELocal tmpvar2);
                            (false, TELocal tmpvar3)]) in
            let call_getvalue_last =
              TEArrayAccess(TELocal tmpvar3,
                            TEConst(TCInt(L.length tagaps))) in
            let (_, call_getvalue) =
              L.fold_left
                (fun (n, acc) ->
                  function
                    | (_, Il.IPDontCare) -> (n + 1, acc)
                    | (_, Il.IPVar(x, t)) ->
                        (n + 1,
                         acc @
                         [TSExp
                            (TEAssign
                               (TELocal x,
                                TECast
                                  (translate_typ t,
                                   TEArrayAccess
                                     (TELocal tmpvar3,
                                      TEConst(TCInt n)))))]))
                (0, [])
                tagaps in
            [match ap' with
            | None ->
                TSIf(TEBinOp(SO.BiAnd,
                             call_switcha,
                             TEBinOp(SO.BiEq,
                                     call_getvalue_last,
                                     TEStaticField("Utils","the_empty_seq"))),
                     call_getvalue @
                     translate_block e,
                     acc)
            | Some Il.IPDontCare ->
                TSIf(call_switcha,
                     call_getvalue @
                     translate_block e,
                     acc)
            | Some(Il.IPVar(x, t)) ->
                TSIf(call_switcha,
                     call_getvalue @
                     TSExp(TEAssign(TELocal x, TECast(translate_typ t,
                                                      call_getvalue_last))) ::
                     translate_block e,
                     acc)]
        | _ -> assert false)
      non_default_clauses
      default_clause in
      clauses

and translate_switcho e clauses =
  let tmpvar = ref "DUMMY" in
  let r2l_eval_sucks =
    reset (fun () ->
      match translate_exp e with
      | TELocal x ->
          tmpvar := x;
          []
      | e ->
          tmpvar := gensym "obj";
          [TSLocals([(!tmpvar, Some(e))], ty_object)]) in
  r2l_eval_sucks @
  let (default_clause, non_default_clauses) =
    L.partition
      (function
        | (Il.IPDefault, _) -> true
        | (_, _) -> false)
      clauses in
  let default_clause =
    match default_clause with
    | [] -> [TSError "non-exhaustive SwitchO"]
    | [Il.IPDefault, c] -> translate_block c
    | _ -> assert false in
  let clauses =
    L.fold_left
      (fun acc (pat, e) ->
        match pat with
        | Il.IPBaseVal(Il.IBVObj t) ->
            [TSIf(TEIs(TELocal !tmpvar, TTYNamed t),
                  translate_block e,
                  acc)]
        | _ -> assert false)
      default_clause
      non_default_clauses in
      clauses

and translate_switcht e clauses =
  let tmpvar = ref "DUMMY" in
  let r2l_eval_sucks =
    reset (fun () ->
      match translate_exp e with
      | TELocal x ->
          tmpvar := x;
          []
      | e ->
          tmpvar := gensym "tag";
          [TSLocals([(!tmpvar, Some(e))], ty_tag)]) in
  r2l_eval_sucks @
  let (default_clause, non_default_clauses) =
    L.partition
      (function
        | (Il.IPDefault, _) -> true
        | (_, _) -> false)
      clauses in
  let default_clause =
    match default_clause with
    | [] -> [TSError "non-exhaustive SwitchV"]
    | [Il.IPDefault, c] -> translate_block c
    | _ -> assert false in
  let clauses =
    L.fold_left
      (fun acc (pat, e) ->
        match pat with
        | Il.IPBaseVal(Il.IBVSymbol t) ->
            let t = C.escape_tag_add_prefix t in
            [TSIf(TEBinOp(SO.BiEq, TELocal !tmpvar, TEConst (TCSymbol t)),
                  translate_block e,
                  acc)]
        | _ -> assert false)
      default_clause
      non_default_clauses in
      clauses

and translate_switchk e clauses =
  let tmpvar = ref "DUMMY" in
  let r2l_eval_sucks =
    reset (fun () ->
      match translate_exp e with
      | TELocal x ->
          tmpvar := x;
          []
      | e ->
          tmpvar := gensym "seq";
          [TSLocals([(!tmpvar, Some(e))], ty_seq)]) in
  r2l_eval_sucks @
  (* [XXX] very peculiar treatment necessary for IPAnyAtom;
     maybe the data type is fishy... *)
  let clauses =
    if L.exists
        (function
          | (Il.IPAnyAtom _, _) -> true
          | (_, _) -> false)
        clauses then
      L.map
        (function
          | (Il.IPDefault, e) -> (Il.IPEmpty, e)
          | other -> other)
        clauses
    else clauses in
  let (empty_clause, non_empty_clauses) =
    L.partition
      (function
        | (Il.IPEmpty, _) -> true
        | (_, _) -> false)
      clauses in
  let non_empty_clauses =
    L.map
      (fun (pat, e) ->
        match pat with
        | Il.IPDefault -> (TPDefault, translate_block e)
        | Il.IPAnyAtom Il.IPDontCare -> (TPDefault, translate_block e)
        | Il.IPAnyAtom(Il.IPVar(x, t)) ->
            (TPDefault,
             TSExp(TEAssign(TELocal x, TEField(TELocal !tmpvar, "next"))) ::
             translate_block e)
        | Il.IPPcdata { Il.iTag = tag; Il.iAtt = att;
                       Il.iCont = cont; Il.iRest = rest } ->
            (TPType TBTPcdata,
             let get_fields =
               (match (cont,rest) with
               | Il.IPVar(x,_),Il.IPVar(y, _) ->
                   let rest = TELocal y in
                   let decls, coalesce = make_coalesce !tmpvar x rest in
                   let str = TELocal x in
                   let asgn = TSExp(TEAssign(str,coalesce)) in
                   if y = !tmpvar then decls @ [asgn]
                   else TSExp(TEAssign(rest,TEConst TCNull)) :: decls @ [asgn]
               | Il.IPDontCare,Il.IPVar(y, _) ->
                   let rest = TELocal y in
                   let x = gensym "bogus" in
                   let decls, coalesce = make_coalesce !tmpvar x rest in
                   if y = !tmpvar then decls @ [TSExp coalesce]
                   else decls @ [TSExp(TEAssign(rest,TEConst TCNull)); TSExp coalesce]
               | Il.IPVar(x,_),Il.IPDontCare ->
                   let y = gensym "bogus" in
                   let rest = TELocal y in
                   let decls, coalesce = make_coalesce !tmpvar x rest in
                   let str = TELocal x in
                   let asgn = TSExp(TEAssign(str,coalesce)) in
                   let decl = TSLocals([y,Some(TEConst TCNull)], ty_seq) in
                   decl :: decls @ [asgn]
               | Il.IPDontCare,Il.IPDontCare ->
                   []) in
             get_fields @ translate_block e)
        | (Il.IPElement { Il.iTag = tag; Il.iAtt = att;
                         Il.iCont = cont; Il.iRest = rest } |
           Il.IPBaseType(_, { Il.iTag = tag; Il.iAtt = att;
                             Il.iCont = cont; Il.iRest = rest }) |
           Il.IPObj { Il.iTag = tag; Il.iAtt = att;
                     Il.iCont = cont; Il.iRest = rest }) as tag_or_obj ->
            (TPType
                (match tag_or_obj with
                | Il.IPElement _ -> TBTElement
                | Il.IPObj _ -> TBTObject
                | Il.IPBaseType(Il.IBTBool, _) -> TBTBool
                | Il.IPBaseType(Il.IBTInt, _) -> TBTInt
                | Il.IPBaseType(Il.IBTChar, _) -> TBTChar
                | _ -> assert false),
             let cls =
               (match tag_or_obj with
               | Il.IPElement _ -> "SeqElement"
               | Il.IPObj _ -> "SeqObject"
               | Il.IPBaseType(Il.IBTBool, _) -> "SeqBool"
               | Il.IPBaseType(Il.IBTInt, _) -> "SeqInt"
               | Il.IPBaseType(Il.IBTChar, _) -> "SeqChar"
               | _ -> assert false) in
             let tmpvar2 = gensym "elem" in
             let do_cast =
               TSLocals([(tmpvar2,
                          Some(TECast(TTYNamed cls,
                                      TELocal !tmpvar)))],
                        TTYNamed cls) in
             let get_fields =
               (match rest with
               | Il.IPVar(x, _) ->
                   [TSExp(TEAssign(TELocal x,
                                   TEField(TELocal !tmpvar, "next")))]
               | Il.IPDontCare -> []) @
               (match tag with
               | Il.IPVar(x, _) ->
                   [TSExp(TEAssign(TELocal x,
                                   TEField(TELocal tmpvar2,
                                           "tag")))]
               | Il.IPDontCare -> []) @
               (match att with
               | Il.IPVar(x, _) ->
                   [TSExp(TEAssign(TELocal x,
                                   TEField(TELocal tmpvar2,
                                           "attributes")))]
               | Il.IPDontCare -> []) @
               (match cont with
               | Il.IPVar(x, _) ->
                   [TSExp(TEAssign(TELocal x,
                                   TEField(TELocal tmpvar2,
                                           "contents")))]
               | Il.IPDontCare -> []) in
             (match get_fields with
             | [] -> []
             | _ -> do_cast :: get_fields) @
             translate_block e)
        | _ -> assert false (* [XX] refine the type *))
      non_empty_clauses in
  let normalize =
    match !Pref.appendpolicy with
      | "eager" -> []
      | "lazynormalize"
      | "eagernormalize" ->
          let x = TELocal !tmpvar in
          let call = TEStaticCall("Utils","Normalize",[false,x]) in
          [TSExp(TEAssign(x,call))]
      | _ -> assert false in
  normalize @
  [make_switch (TEField(TELocal !tmpvar, "kind"))
     ((match empty_clause with
       | [Il.IPEmpty, e] -> [(TPType TBTEmpty, translate_block e)]
       | [] -> []
       | _ -> assert false) @
      non_empty_clauses)]

and translate_catch e =
  let rec gather_handlers = function
    | Il.IECatch(body, k, xts, handler) ->
        (k, handler) :: gather_handlers body
    | _ -> [] in
  let handlers = L.rev (gather_handlers e) in
  let rec get_body = function
    | Il.IECatch(body, k, xts, handler) -> get_body body
    | body -> body in
  let body = get_body e in
  [TSSwitch(TEConst(TCInt 0),
            (TPDefault, translate_block body) ::
            L.map
              (fun (k, handler) -> (TPLabel k, translate_block handler))
              handlers)]

let translate_form_param (x, y, z) = (x, y, translate_typ z)

let translate_invoke = function
    (* [XX] no reset, assuming that no stmt appear inside these act_param's *)
  | Il.IINVNone -> TINVNone
  | Il.IINVThis x -> TINVThis(L.map translate_act_param x)
  | Il.IINVSuper x -> TINVSuper(L.map translate_act_param x)

let merge = Sortedlist.merge
let diff = Sortedlist.diff
let compare' (x1, t1) (x2, t2) = compare (t1, x1) (t2, x2)

let rec gather_pattern_variables_in_an_atompat = function
  | Il.IPVar(x, t) -> [(x, t)]
  | Il.IPDontCare -> []
let rec gather_pattern_variables_in_a_pattern = function
  | Il.IPEmpty
  | Il.IPDefault
  | Il.IPRegex(_,_)
  | Il.IPBaseVal _ -> []
  | Il.IPAnyAtom ap -> gather_pattern_variables_in_an_atompat ap
  | Il.IPPcdata { Il.iTag = ap1; Il.iAtt = ap2; Il.iCont = ap3; Il.iRest = ap4 }
  | Il.IPElement { Il.iTag = ap1; Il.iAtt = ap2; Il.iCont = ap3; Il.iRest = ap4 }
  | Il.IPBaseType
      (_, { Il.iTag = ap1; Il.iAtt = ap2; Il.iCont = ap3; Il.iRest = ap4 })
  | Il.IPObj { Il.iTag = ap1; Il.iAtt = ap2; Il.iCont = ap3; Il.iRest = ap4 } ->
      merge compare'
        (gather_pattern_variables_in_an_atompat ap1)
        (merge compare'
           (gather_pattern_variables_in_an_atompat ap2)
           (merge compare'
              (gather_pattern_variables_in_an_atompat ap3)
              (gather_pattern_variables_in_an_atompat ap4)))
  | Il.IPAtt(tagaps, Some ap) ->
      L.fold_left
        (fun acc (tag, ap) ->
          merge compare' acc (gather_pattern_variables_in_an_atompat ap))
        (gather_pattern_variables_in_an_atompat ap)
        tagaps
  | Il.IPAtt(tagaps, None) ->
      L.fold_left
        (fun acc (tag, ap) ->
          merge compare' acc (gather_pattern_variables_in_an_atompat ap))
        []
        tagaps
let rec gather_pattern_variables = function
  | Il.IEConst _
  | Il.IELocal _
  | Il.IEStaticField(_,_)
  | Il.IEEmpty
  | Il.IEBreak
  | Il.IEReturn None
  | Il.IEError _ -> []
  | Il.IEReturn(Some(e))
  | Il.IEField(e, _)
  | Il.IEUnOp(_, e)
  | Il.IETypeStamp(_, e)
  | Il.IECast(_, e) -> gather_pattern_variables e
  | Il.IERegexField(_,_,e1,e2)
  | Il.IEAssign(e1, e2)
  | Il.IEBinOp(_, e1, e2)
  | Il.IEWhile(e1, e2)
  | Il.IESeq(e1, e2) ->
      merge compare'
        (gather_pattern_variables e1)
        (gather_pattern_variables e2)
  | Il.IELet(ap, None, e) ->
      merge compare'
        (gather_pattern_variables e)
        (gather_pattern_variables_in_an_atompat ap)
  | Il.IELet(ap, Some(e1), e2) ->
      merge compare'
        (gather_pattern_variables e1)
        (merge compare'
           (gather_pattern_variables e2)
           (gather_pattern_variables_in_an_atompat ap))
  | Il.IEIf(e1, e2, e3) ->
      merge compare'
        (gather_pattern_variables e1)
        (merge compare'
           (gather_pattern_variables e2)
           (gather_pattern_variables e3))
  | Il.IEExit(_, es)
  | Il.IENew(_, es) ->
      L.fold_left
        (fun acc e ->
          merge compare' acc
            (gather_pattern_variables e))
        []
        es
  | Il.IENewArray(_, e,_) -> gather_pattern_variables e
  | Il.IEArrayAccess (e1, e2) ->
      merge compare'
        (gather_pattern_variables e1)
        (gather_pattern_variables e2)
  | Il.IECall(e, _, es') ->
      L.fold_left
        (fun acc e ->
          merge compare' acc
            (gather_pattern_variables e))
        (gather_pattern_variables e)
        (L.map snd es')
  | Il.IEStaticCall(_, _, es') ->
      L.fold_left
        (fun acc e ->
          merge compare' acc
            (gather_pattern_variables e))
        []
        (L.map snd es')
  | Il.IEProcCall(_, _, xts, e, eo) ->
      merge compare'
        (L.fold_left
           (fun acc xt -> merge compare' acc [xt])
           []
           xts)
        (merge compare'
           (gather_pattern_variables e)
           (gather_pattern_variables' eo))
  | Il.IECatch(body, k, xts, handler) ->
      handler_environment := (k, xts) :: !handler_environment;
      merge compare'
        (L.fold_left
           (fun acc xt -> merge compare' acc [xt])
           []
           xts)
        (merge compare'
           (gather_pattern_variables body)
           (gather_pattern_variables handler))
  | Il.IESwitchK(e, clauses)
  | Il.IESwitchV(e, clauses)
  | Il.IESwitchO(e, clauses)
  | Il.IESwitchP(e, clauses)
  | Il.IESwitchA(e, clauses) ->
      L.fold_left
        (fun acc (pat, e) ->
           merge compare' (gather_pattern_variables e)
             (merge compare' acc (gather_pattern_variables_in_a_pattern pat)))
        (gather_pattern_variables e)
        clauses
and gather_pattern_variables' = function
  | None -> []
  | Some e -> gather_pattern_variables e

let translate_clmember mem =
  let rec declare_pattern_variables = function
    | [] -> []
    | (x, t0) :: rest ->
        let t = translate_typ t0 in
        let vardecl = (x, None) in
        let vardecl = (x, Some(translate_exp (C.bottom_value t0))) in
        match declare_pattern_variables rest with
        | TSLocals(xs, t') :: rest when t' = t ->
            TSLocals(vardecl :: xs, t) :: rest
        | other -> TSLocals([vardecl], t) :: other
  in match mem with
    | Il.ICMConstr(x, form_params, e) ->
        handler_environment := [];
        (* gensym_counter := 0; *) (* do it if you like *)
        let pattern_variables =
          diff compare'
            (gather_pattern_variables e)
            (L.fold_left
               (fun acc (_, x, t) ->
                 merge compare' acc [(x, t)])
               []
               form_params) in
        TCMConstr(translate_invoke x,
                  L.map translate_form_param form_params,
                  declare_pattern_variables pattern_variables @
                  translate_block e)
    | Il.ICMStaticConstr(body) ->
        handler_environment := [];
        let pattern_variables = gather_pattern_variables body in
        TCMStaticConstr(declare_pattern_variables pattern_variables @
                        translate_block body)
    | Il.ICMFld(x, s, y, e) ->
        TCMFld(x, s, translate_typ y, translate_exp' e)
    | Il.ICMMeth(a, isstatic, b, form_params, d, e) ->
        handler_environment := [];
        (* gensym_counter := 0; *) (* do it if you like *)
        let pattern_variables =
          diff compare'
            (gather_pattern_variables e)
            (L.fold_left
               (fun acc (_, x, t) ->
                 merge compare' acc [(x, t)])
               []
               form_params) in
        TCMMeth(a, isstatic, b,
                L.map translate_form_param form_params,
                translate_typ d,
                declare_pattern_variables pattern_variables @
                translate_block e)

let translate_meth_decl (a, form_params, d) =
  (a,
   L.map translate_form_param form_params,
   translate_typ d)
        
let translate_clas {
  Il.iCLHomeNS = iCLHomeNS;
  Il.iCLName = iCLName;
  Il.iCLExt = iCLExt;
  Il.iCLImpls = iCLImpls;
  Il.iCLMems = iCLMems;
} = {
  tCLHomeNS = iCLHomeNS;
  tCLName = iCLName;
  tCLExt = iCLExt;
  tCLImpls = iCLImpls;
  tCLMems = L.map translate_clmember iCLMems;
}

let translate_intf {
  Il.iINHomeNS = iINHomeNS;
  Il.iINName = iINName;
  Il.iINImpls = iINImpls;
  Il.iINMems = iINMems;
} = {
  tINHomeNS = iINHomeNS;
  tINName = iINName;
  tINImpls = iINImpls;
  tINMems = L.map translate_meth_decl iINMems;
}

let translate_usedir = function
  | Il.IUSAlias(x, y) -> TUSAlias(x, y)
  | Il.IUSNSpace x -> TUSNSpace x

let translate_prog (y, intfs, z) =
  (List.map translate_usedir y, List.map translate_intf intfs, List.map translate_clas z)

(* for debugging by hand *)
(*
let pp e =
  let s = Codegen.zblock (translate_block e) in
  Codegen.indent s (String.length s) 0 stdout 0 0
*)
