module SD = SynDenot
module HT = Hashtbl
module SL = Sortedlist
open List

let ta_any = ref Ta.empty   (** this duplicates Store.ta_any *)

let fpr = Format.fprintf
let outf = !Pref.output_formatter

let reset_any_state () =
  let s = Buildta.build_ta !Store.any_seqty_def in
  match s.Ta.next with
    | [_,_,s1,_,_] ->
        ta_any := s1
    | _ -> assert false

let lab (l1 : SD.labty) (l2 : SD.labty) : SD.labty option =
  let (h1,vars1) = SD.crack_labty l1 in
  let (h2,vars2) = SD.crack_labty l2 in
  let vars = vars1 @ vars2 in
  match Sub.minimum_hostty [h1;h2] with
    | None -> None
    | Some h -> Some (SD.slap_labty vars h)

let atts (a1 : SD.attsty) (a2 : SD.attsty) : SD.attsty =
  match (SD.attsty' a1, SD.attsty' a2) with
    | SD.TAttAny, _ -> a2
    | _, SD.TAttAny -> a1
    | SD.PAtt(attpats1,rest1), SD.PAtt(attpats2,rest2) ->
        match (rest1,rest2) with
          | None, _ -> SD.attsty (SD.PAtt(attpats1 @ attpats2, rest2))
          | _ -> SD.attsty (SD.PAtt(attpats1 @ attpats2, rest1))

let _cache_ : (int*int, Ta.state) HT.t = HT.create 97

let rec state (s1 : Ta.state) (s2 : Ta.state) : Ta.state =
  let any_id = (!ta_any).Ta.num in
  if any_id == s1.Ta.num then s2
  else if any_id == s2.Ta.num then s1
  else if s1.Ta.num == s2.Ta.num then s1
  else (
    try HT.find _cache_ (s1.Ta.num,s2.Ta.num) 
    with Not_found ->
      let res = Ta.create_state (s1.Ta.final && s2.Ta.final) [] None in
      res.Ta.finpre <- s1.Ta.finpre @ s2.Ta.finpre;
      HT.add _cache_ (s1.Ta.num, s2.Ta.num) res;
      let trs = trans s1.Ta.next s2.Ta.next in
      res.Ta.next <- rev trs;
      res)

and trans (trs1 : Ta.trans list) (trs2 : Ta.trans list) : Ta.trans list =
  let res = ref [] in
  iter
    (fun (l1, atts1, s11, s12, pre1)  ->
       iter
         (fun (l2, atts2, s21, s22, pre2)  ->
            match (lab l1 l2) with
              | Some l ->
                  let pre = pre1 @ pre2 in
                  let atts = atts atts1 atts2 in
                  let s1 = state s11 s21 in
                  let s2 = state s12 s22 in
                  res := (l, atts, s1, s2, pre) :: !res
              | None -> ())
         trs2)
    trs1;
  !res

let inter s1 s2 =
  match Ta.elim_empty [state s1 s2] with
    | [] -> Ta.empty
    | x::_ -> x

let disjoint s1 s2 =
  let any_id = (!ta_any).Ta.num in
  if (any_id == s1.Ta.num || any_id == s2.Ta.num) then false
  else Ta.is_empty (inter s1 s2)

let _union_cache_ : (int*int, Ta.state) HT.t = HT.create 97

let union2 s1 s2 =
  let any_id = (!ta_any).Ta.num in
  if any_id == s1.Ta.num || any_id == s2.Ta.num then !ta_any
  else if Ta.empty.Ta.num == s1.Ta.num then s2
  else if Ta.empty.Ta.num == s2.Ta.num then s1
  else if s1.Ta.num == s2.Ta.num then s1
  else (
    try HT.find _union_cache_ (s1.Ta.num,s2.Ta.num)
    with Not_found ->
      let res = Ta.create_state (s1.Ta.final || s2.Ta.final) [] None in
      let finpre1 = sort Ta.compare_act s1.Ta.finpre in
      let finpre2 = sort Ta.compare_act s2.Ta.finpre in
      res.Ta.finpre <- SL.merge Ta.compare_act finpre1 finpre2;
      HT.add _union_cache_ (s1.Ta.num, s2.Ta.num) res;
      let trs1 = sort Ta.compare_trans s1.Ta.next in
      let trs2 = sort Ta.compare_trans s2.Ta.next in
      let trs = SL.merge Ta.compare_trans trs1 trs2 in
      res.Ta.next <- trs;
      res)

(*  let union2 s1 s2 = *)
(*    fpr outf "union %d %d = " s1.Ta.num s2.Ta.num; *)
(*    let res = union2 s1 s2 in *)
(*    fpr outf "%d@." res.Ta.num; *)
(*    res *)

let union = fold_left union2 Ta.empty

let reset () =
  reset_any_state ()
