open Format
open DenBoot

let debug = false

let triv_empty z =
  not z.Ta.final && z.Ta.next = [] 

let rec isect_rec boot ihyps x y =
  let res = 
      try Ta.StStMap.find (x,y) !ihyps
      with Not_found ->
	if Sub.subtype boot [x] [y] = Result.Succ then begin
	  ihyps := Ta.StStMap.add (x,y) x !ihyps;
	  x
	end else if Sub.subtype boot [y] [x] = Result.Succ then begin
	  ihyps := Ta.StStMap.add (x,y) y !ihyps;
	  y
	end else
	  let z = Ta.create_state (x.Ta.final && y.Ta.final) [] None in
	  ihyps := Ta.StStMap.add (x,y) z !ihyps;
	  let zl = isectl boot ihyps x.Ta.next y.Ta.next in
	  z.Ta.next <- zl;
	  z
  in
  if debug then printf "isect@ @[%a@]@ and@ @[%a@]@ yielded@ @[%a@]@." Ta.print [x] Ta.print [y] Ta.print [res];
  res


and isectl boot ihyps xl yl =
  let zl = 
    List.concat
      (List.map
	 (fun (c, x1, x2, _) ->
	   List.concat
	     (List.map
		(fun (d, y1, y2, _) ->
		  let e = boot.sb_labdenot.ld_inter c d in
		  if SynDenot.is_emptylab e then []
		  else
		    let z1 = isect_rec boot ihyps x1 y1 in
		    if triv_empty z1 then [] 
		    else
		      let z2 = isect_rec boot ihyps x2 y2 in
		      if triv_empty z2 then []
		      else
			[(e, z1, z2, [])])
		yl))
	 xl) in
  (*Wrong to eliminate empty during calculating intersection *)
  zl
    

let isect boot x y =
  let res = 
    Listutil.union_many Ta.union_state_sets
      (List.map (fun x1 ->
	Listutil.union_many Ta.union_state_sets
	  (List.map (fun y1 ->
	    [isect_rec boot boot.sb_engine.Eng.isect_hyps x1 y1])
	     y))
	 x) in
  res

let isect1 boot x y =
  let res = isect_rec boot boot.sb_engine.Eng.isect_hyps x y in
  res

let isectl boot xl yl =
  let res = isectl boot boot.sb_engine.Eng.isect_hyps xl yl in
  res

let disjoint boot x y =
  List.for_all Ta.is_empty (isect boot x y)

let disjointl boot xl yl =
  Ta.is_empty_l (isectl boot xl yl)

(* diff ********************)

 
let rec diff_rec boot dhyps x y =
  try Ta.StLsMap.find (x,y) !dhyps
  with Not_found ->
    let z = Ta.create_state (x.Ta.final && not(Ta.state_set_final y)) [] None in
    dhyps := Ta.StLsMap.add (x,y) z !dhyps;
    let zl = diffl boot dhyps x.Ta.next (Ta.state_set_next y) in
    z.Ta.next <- zl;
    z

and add_diff boot (p,n) y = (p,Sortedlist.merge Ta.compare_state [y] n)
and is_diff_empty boot (p,n) = (Sub.subtype boot [p] n = Result.Succ)

and diff_shallow boot (c, u1, u2) (d, y1, y2) =
  if boot.sb_labdenot.ld_disjoint c d then [(c, u1, u2)] else
  let e = boot.sb_labdenot.ld_diff c d in
  let u1' = add_diff boot u1 y1 in
  let u2' = add_diff boot u2 y2 in
  (if SynDenot.is_emptylab e then [] else [(e, u1, u2)]) @
  (if is_diff_empty boot u1' then [] else [(c, u1', u2)]) @
  (if is_diff_empty boot u2' then [] else [(c, u1, u2')])
      
and diffl boot dhyps xl yl =
  let l = List.map (fun (c, x1, x2, _) -> (c, (x1, []), (x2, []))) xl in
  let l = 
    List.fold_left
      (fun l (d, y1, y2, _) ->
	Listutil.mapcat
	  (fun (c, u1, u2) ->
	    diff_shallow boot (c, u1, u2) (d, y1, y2))
	  l)
       l yl in
  Listutil.mapcat (fun (c, (x1, y1), (x2, y2)) ->
    if Sub.subtype boot [x1] y1 = Result.Succ then [] else
    let z1 = if disjoint boot [x1] y1 then x1 else diff_rec boot dhyps x1 y1 in
    if triv_empty z1 then []
    else
      if Sub.subtype boot [x2] y2 = Result.Succ then [] else
      let z2 = if disjoint boot [x2] y2 then x2 else diff_rec boot dhyps x2 y2 in
      if triv_empty z2 then [] else
      [(c, z1, z2, [])])
    l
    

let diff boot x y =
  let res = List.map (fun x1 -> diff_rec boot boot.sb_engine.Eng.diff_hyps x1 y) x in
  if debug then printf "diff@ @[%a@]@ from@ @[%a@]@ yielded@ @[%a@]@." Ta.print x Ta.print y Ta.print res;
  res

let diff1 boot x y =
  let res = diff_rec boot boot.sb_engine.Eng.diff_hyps x y in
  res

let diffl boot xl yl =
  let res = diffl boot boot.sb_engine.Eng.diff_hyps xl yl in
  res

