(**  This module is supposed to define ids of core types, 
  i.e. those ids are hard-wired elsewhere in the source.  *)

let mkLibId str = Ident.create_located [] str

let mkSystemId name = Ident.create_located ["System"] name
let mkXtaticId name = Ident.create_located ["Xtatic"] name

(* classes and structs from System namespace *)

let object_id = mkSystemId "Object"
let bool_id = mkSystemId "Boolean"
let int_id = mkSystemId "Int32"   
let string_id = mkSystemId "String"
let char_id = mkSystemId "Char"

(* Xtatic namespace *)
let tag_id = mkXtaticId "Tag"
let seq_id = mkXtaticId "Seq" 

(* Tol-level namespace *)
let any_id = mkLibId "any"
let anyone_id = mkLibId "anyone"
let pcchar_id = mkLibId "pcchar"
let pcdata_id = mkLibId "pcdata" 
let element_id = mkLibId "element"
let xmlone_id = mkLibId "xmlone"
let xml_id = mkLibId "xml"
let attseq_id = mkLibId "attseq"


(**********  Some functions ***********************************)

module SS = SynSource

let ker_info x = (Finfo.make_finfo "<kernellib>" (-1) (-1), x)

let mkLibName (str: string) : SS.longname = 
  [(ker_info str)]

let mkLibLongname (strlist: string list) : SS.longname = 
  List.map ker_info strlist


let mkLibRefname (id: Ident.id) : SS.refname = 
  mkLibLongname [Ident.name id]

let mkLibQname (id : Ident.id) : SS.qualname = 
  (None, ker_info (Ident.name id))
