(**  This module now contains only the library of core types, 
  i.e. those types whose ids are hard-wired elsewhere in the source.
  All other library types are now maintained in src/../lib/*.xti files *)

module I = Ident
module C = Contexts
module SO = SynOp
module SS = SynSource
module SD = SynDenot

module CN = Cnaming
module XN = Xnaming
module N = Naming

open Kerids


(************** nspace  --- Namespace *******************)

let spaceAddClas nspace id = CN.spaceAddChildLeaf nspace (id, CN.LClas)

let nspace =
  let cspace_System = 
    List.fold_left spaceAddClas CN.spaceEmpty
      [object_id; bool_id; int_id; string_id; char_id;] in
  let cspace_Xtatic = 
    List.fold_left spaceAddClas CN.spaceEmpty
      [tag_id; seq_id] in 
  let cspace_global = 
    CN.spaceAddChildSpace CN.spaceEmpty "System" cspace_System in 
  let cspace_global = 
    CN.spaceAddChildSpace cspace_global "Xtatic" cspace_Xtatic 
  in
  let xspace_global = 
    List.fold_left 
      (fun spaces id -> 
	 XN.spacesAddMember spaces (XN.NoUri, Ident.name id) id
      )
      XN.spacesEmpty
      [any_id; anyone_id; xml_id; xmlone_id; element_id;
       pcchar_id; pcdata_id; attseq_id;]    
  in
    { N.cs = cspace_global;
      N.xs = xspace_global }


(************** nimports  --- Aliases for nspace items ********)

let nimports =      (*--VG##  put real contents later... *)
  let cimports = List.map 
		   (fun (a,i) -> CN.AliasMember(a, CN.Leaf(i, CN.LClas)))
		   ["object", object_id;
		    "string", string_id;
		    "bool", bool_id;
		    "int", int_id;
		    "char", char_id;
                    "tag", tag_id;
                    "seq", seq_id;
		   ]
  in
  let ximports = []
  in
    { N.ci = cimports;
      N.xi = ximports; }


(******************** datctx  -- Data context ***********************)

let mkClassLabty cid = SD.basiclab(SD.hostty (SD.TClasRef cid))

let object_labty = mkClassLabty object_id
let tag_labty = mkClassLabty tag_id
let char_labty = mkClassLabty char_id

let anyone_seqty_def = 
  SD.tree object_labty  SD.attsty_any (*##ATT*) SD.sty_any

let any_seqty_def = 
  SD.alt[SD.sty_epsilon;
	 SD.cat[anyone_seqty_def;
		SD.sty_any;
	      ];
       ] 

let element_seqty_def = 
  SD.tree tag_labty  SD.attsty_any (*##ATT*) SD.sty_xml

let pcchar_seqty_def = 
  SD.tree char_labty  SD.attsty_empty (*##ATT*) SD.sty_epsilon

let xmlone_seqty_def =
  SD.alt[element_seqty_def; pcchar_seqty_def]

let xml_seqty_def = 
  SD.alt[SD.sty_epsilon;
	 SD.cat[xmlone_seqty_def;
		SD.sty_xml;
	      ];
       ] 
  
let pcdata_seqty_def = 
  SD.rep (pcchar_seqty_def)

let attseq_seqty_def = 
  let tag_labty = SD.basiclab(SD.hostty (SD.TClasRef Kerids.tag_id)) in
  let sty_pcdata_attr = 
    SD.tree tag_labty (SD.attsty_any) (SD.seqref Kerids.pcdata_id) 
  in SD.rep(sty_pcdata_attr)
  


let seqty_defs = [
  (any_id, any_seqty_def);
  (anyone_id, anyone_seqty_def);
  (pcchar_id, pcchar_seqty_def);
  (pcdata_id, pcdata_seqty_def);
  (element_id, element_seqty_def);
  (xmlone_id, xmlone_seqty_def);
  (xml_id, xml_seqty_def);
  (attseq_id, attseq_seqty_def);
] 

let clas_parents = [
  (bool_id, object_id);
  (int_id, object_id);
  (string_id, object_id);
  (char_id, object_id);
  (tag_id, object_id);
  (seq_id, object_id);
] 
let clas_implements = [
  (bool_id, []);
  (int_id, []);
  (string_id, []);
  (char_id, []);
  (tag_id, []);
  (seq_id, []);
] 

let datctx = {
  Semant.dc_topclass = object_id;
  Semant.dc_parent = List.fold_left Ident.add Ident.empty clas_parents;
  Semant.dc_implements = List.fold_left Ident.add Ident.empty clas_implements;
  Semant.dc_seqdef = List.fold_left Ident.add Ident.empty seqty_defs;
  Semant.dc_seqdef_untranslated = Ident.empty;
  Semant.dc_any_id = any_id;
} 


(********************* clsctx -- Classing context ***********************)

let ty_pcdata = SD.ty (SD.Tseq (SD.seqty (SD.TSeqRef Kerids.pcdata_id)))
let ty_xml    = SD.ty (SD.Tseq (SD.seqty (SD.TSeqRef Kerids.xml_id)))
let ty_any    = SD.ty (SD.Tseq (SD.seqty (SD.TSeqRef Kerids.any_id)))


type kerclas = {
    kc_parent: I.id;
    kc_sealed: bool;
    kc_fields: (Ident.id * (SO.status * SD.ty)) list;   (*only native fields *)
    kc_constr: SD.ty list;
    kc_meths_ret: (Ident.id * (SO.status * SO.method_modifier * SD.ty option)) list;  (* only native meths *)
    kc_meths: (string * 
                 (Ident.id * SO.status * SO.method_modifier * SD.ty option * (SD.ty list))) 
              list;
    kc_meths_names: (Ident.id * string) list;
  } 

type kerclslib = (I.id * kerclas) list


let find_sealed (lib : kerclslib) cid =
  let clas = List.assq cid lib in
  clas.kc_sealed

let rec field (lib: kerclslib) cid fid = 
  let clas = List.assq cid lib in
  try List.assq fid clas.kc_fields
  with Not_found ->
    if I.dummyid == clas.kc_parent then raise Not_found
    else field lib clas.kc_parent fid

let rec find_field (lib: kerclslib) cid fname = 
  let clas = List.assq cid lib in
  try List.find (fun (id, _) -> fname = (Ident.name id)) clas.kc_fields
  with Not_found ->
    if I.dummyid == clas.kc_parent then raise Not_found
    else find_field lib clas.kc_parent fname

let rec fields_all lib cid = 
  let clas = List.assq cid lib in
  let this_fields = clas.kc_fields in
  let parent_fields = 
    if I.dummyid == clas.kc_parent then []
    else fields_all lib clas.kc_parent
  in parent_fields @ this_fields

let constr_type lib cid = 
  let clas = List.assq cid lib in
  clas.kc_constr

let rec meth_ret_type lib cid mname = 
  let clas = List.assq cid lib in
  try List.assoc mname clas.kc_meths_ret
  with Not_found -> 
    if I.dummyid == clas.kc_parent then raise Not_found
    else meth_ret_type lib clas.kc_parent mname

let meth_this_type lib cid name =
  let clas = List.assq cid lib in
  List.fold_left (fun acc (n,m) -> if n = name then m :: acc else acc) []
  clas.kc_meths
  
 let make_meths =
  List.fold_left (
    fun (mr,m,mn) (name, (is_s,mo, rtty, argtys)) ->
      let id = MethIdent.getid name argtys in
      ((id, (is_s, mo, rtty)) :: mr),
      ((name, (id, is_s, mo, rtty, argtys)) :: m),
      ((id, name) :: mn))
  ([],[],[])

let argty_ToString = []
let argty_Equals = [SynDenot.ty_Object]

let mid_ToString = MethIdent.getid "ToString" argty_ToString
let mid_Equals = MethIdent.getid "Equals" argty_Equals

let kerclslib = 
  let object_clas = 
    let meths = [
    ("ToString", (SO.Instance, SO.Virtual, Some SD.ty_String, argty_ToString));
    ("Equals", (SO.Instance, SO.Virtual, Some SD.ty_Bool, argty_Equals));
    ("GetHashCode", (SO.Instance, SO.Virtual, Some SD.ty_Int, []));
    ]
    in
    let mr,m,mn = make_meths meths in
    {
    kc_parent = I.dummyid;
    kc_sealed = false;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = mr;  
    kc_meths = m;
    kc_meths_names = mn;
  }
  in
  let bool_clas = {
    kc_parent = object_id;
    kc_sealed = true;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = [];  
    kc_meths = [];
    kc_meths_names = [];
  }
  in
  let int_clas = {
    kc_parent = object_id;
    kc_sealed = true;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = [];  
    kc_meths = [];
    kc_meths_names = [];
  }
  in
  let char_clas = {
    kc_parent = object_id;
    kc_sealed = true;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = [];  
    kc_meths = [];
    kc_meths_names = [];
  }
  in
  let string_clas =
    let meths = [("CompareTo",
                  (SO.Instance, SO.Virtual, Some SD.ty_Int, [SD.ty_String]))]
    in let mr,m,mn = make_meths meths in
    {
    kc_parent = object_id;
    kc_sealed = true;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = mr;  
    kc_meths = m;
    kc_meths_names = mn;
  }
  in
  let tag_clas = 
    let meths = [("Create", 
                  (SO.Static, SO.Normal, Some SD.ty_Tag, [SD.ty_String]))]
    in let mr,m,mn = make_meths meths in
   {
    kc_parent = object_id;
    kc_sealed = true;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = mr;  
    kc_meths = m;
    kc_meths_names = mn;
  }
  in
  let seq_clas = {
    kc_parent = object_id;
    kc_sealed = true;
    kc_fields = [];
    kc_constr = [];
    kc_meths_ret = [];  
    kc_meths = [];
    kc_meths_names = [];
  }
  in
  [
      (object_id, object_clas);
      (bool_id, bool_clas);
      (int_id, int_clas);
      (char_id, char_clas);
      (string_id, string_clas);
      (tag_id, tag_clas);
      (seq_id, seq_clas);
    ] 

let make_meth_sigs meths = 
  List.map (fun (name, (retty, argtys)) -> 
	      (name, (MethIdent.getid name argtys, retty, argtys)))
    meths



let clsctx = 
  { C.cc_sealed = find_sealed kerclslib;
    C.cc_fields = fields_all kerclslib; 
    C.cc_field =  field kerclslib;
    C.cc_find_field =  find_field kerclslib;
    C.cc_constrtype = constr_type kerclslib;
    C.cc_methrettype = meth_ret_type kerclslib;
    C.cc_meththistype = meth_this_type kerclslib;
    C.cc_mangled_name =
      List.fold_left (
        fun acc (_, cls) ->
          C.extend_mangled_names acc cls.kc_meths_names)
      Ident.empty
      kerclslib;
    C.cc_implements = (fun cid -> []);
    C.cc_intf_direct = (fun cid -> List.assq cid clas_implements)
  } 
