module SS = SynSource

exception DTD_error of string


let bogusfi t = (Finfo.bogus(), t)

let bogusfi_decname str = bogusfi str
let bogusfi_refname str = [bogusfi str]



let typ_of_atttys prefix anames attmap : SS.attsty  =
  bogusfi (SS.TAttAny)    (*##ATT *)
   (*## Currently, attribute declarations are ignored 
	 (patterns extract attributes in untyped way)  *)


let type_of_mixed prefix mixspec_list = 
  let basemix = List.map (function
    | Pxp_types.MPCDATA -> 
	bogusfi(SS.TSeqRef(Kerids.mkLibQname Kerids.pcdata_id))
    | Pxp_types.MChild name -> 
	bogusfi(SS.TSeqRef (None, bogusfi (prefix ^ name)))
			 ) 
		  mixspec_list
in bogusfi(SS.TRep (bogusfi(SS.TAlt basemix)))


let rec type_of_regexp prefix respec = 
  match respec with
  | Pxp_types.Optional spec1 -> 
      bogusfi(SS.TAlt[type_of_regexp prefix spec1;  bogusfi(SS.TCat[])])
  | Pxp_types.Repeated spec1 ->
      bogusfi(SS.TRep (type_of_regexp prefix spec1))
  | Pxp_types.Repeated1 spec1 ->
      let t = type_of_regexp prefix spec1 in
      bogusfi(SS.TCat[t;  bogusfi(SS.TRep (type_of_regexp prefix spec1)) ])
  | Pxp_types.Alt specl ->
      bogusfi(SS.TAlt (List.map (type_of_regexp prefix) specl))
  | Pxp_types.Seq specl ->
      bogusfi(SS.TCat (List.map (type_of_regexp prefix) specl))
  | Pxp_types.Child name ->
      bogusfi(SS.TSeqRef (None, bogusfi (prefix ^ name)))



let type_of_content prefix contmodel : SS.seqty = 
  match contmodel with
  | Pxp_types.Unspecified -> 
      raise (DTD_error "content model: Unspecified?")
  | Pxp_types.Empty -> 
      bogusfi(SS.TCat[])
  | Pxp_types.Any -> 
      bogusfi(SS.TSeqRef(Kerids.mkLibQname Kerids.any_id))
  | Pxp_types.Mixed speclist -> 
      type_of_mixed prefix speclist
  | Pxp_types.Regexp spec ->
      type_of_regexp prefix spec



let incorp_types_of_dtd prefix d = 
  let elem_names = d # element_names in
  List.map
    (fun ename -> 
       let p_ename = prefix ^ ename in
       let elem = d # element ename in
       let attsty = 
	 typ_of_atttys prefix (elem # attribute_names) (elem # attribute) in
       let tagty = bogusfi(SS.TLab (bogusfi(SS.Ttag ename))) in
       let contty = type_of_content prefix (elem # content_model)  in 
       let typdef = bogusfi (SS.TTree(tagty, attsty, contty))
       in (bogusfi_decname p_ename,  typdef)
    ) elem_names
   


let rec print_error e =
  match e with
    Pxp_types.At(where,what) ->
      print_endline where;
      print_error what
  |  _ ->
      raise e

let parse_dtd (prefix: string) (fname: string) 
  : (SS.decname * SS.seqty) list = 
  let prefix = if (prefix = "") then "" else prefix ^ "" in
   try
   let fname = Filing.search_file !Pref.searchpaths fname in
   let d = Pxp_yacc.parse_dtd_entity 
               Pxp_yacc.default_config  (Pxp_yacc.from_file fname)
   in
   incorp_types_of_dtd prefix d
   with
   e -> print_error e
  


