module SD = SynDenot
let fpr = Format.fprintf


(******************* Printing for values  ***************************)
(*--
let rec pr_va ff = function
  | SD.Vhost hva -> pr_hostva ff hva
  | SD.Vseq seq -> fpr ff "@[{|%a|}@]" pr_seqva seq

and pr_hostva ff = function
  | SD.Vobj (cid, objva) -> fpr ff "OBJ:%a" Ident.print cid
  | SD.Vnull -> fpr ff "null"
  | SD.Vbool b -> Format.pp_print_bool ff b
  | SD.Vint n -> Format.pp_print_int ff n
  | SD.Vstring s -> fpr ff "%s" s 
  | SD.Vchar c -> fpr ff "%c" c
  | SD.Vtag t -> fpr ff "@[Tag(%s)@]" t  
      (* Vtag clause is intended to be intercepted by pr_start_marker *)
  | SD.Vclass cid -> fpr ff "CLASS:%a" Ident.print cid

and pr_start_marker ff = function
  | SD.Vtag t -> fpr ff "%s" t
  | SD.Vstring s ->  fpr ff "(\"%s\")" s
  | otherlab -> fpr ff "(%a)" pr_hostva otherlab

and pr_end_marker ff = function
  | SD.Vtag t -> fpr ff "%s" t
  | _ -> ()

and pr_seqva ff seqva = 
  let pr_pcdata ff str = 
    fpr ff "`%s`" str
  in
  let pr_nopcstart ff = function
    | [] -> ()
    | tree::rest -> 
	(match tree with 
	| SD.Vtree(lab, atts, []) ->  (* childless element *)
	    fpr ff "@[%s%a%a%s@]@," 
	      "<" pr_start_marker lab pr_attsva atts "/>"
	| SD.Vtree(lab, atts, cont) -> 
	    fpr ff "@[%s%a%a%s@;<0 2>%a%s%a%s@]@," 
	      "<" pr_start_marker lab pr_attsva atts ">"
	      pr_seqva cont "</" pr_end_marker lab ">"
	);
	pr_seqva ff rest
  in
  match SD.chop_pcdata_as_string seqva with
  | "", [] -> ()
  | "", sva ->  fpr ff "@[%a@]" pr_nopcstart sva
  | pcd, [] ->  fpr ff "@[%a@]" pr_pcdata pcd
  | pcd, sva -> fpr ff "@[%a@]@\n@[%a@]" pr_pcdata pcd pr_nopcstart sva

and pr_attsva ff atts = 
  if Attset.is_empty atts then () else
  Format.fprintf ff " ";
  let pr_aval ff aval = 
    match SD.chop_pcdata_as_string aval with 
    | str_aval, [] -> (* the attr is fully pcdata *)
	Format.fprintf ff "@[\"%s\"@]" str_aval
    | _ ->   (* the attr is non-pcdata *)
	Format.fprintf ff "@[%s(%a)@]" "@" pr_seqva aval
  in Attset.print pr_aval ff atts


let toString_va va = 
  let supposed_empty = Format.flush_str_formatter () in
  if "" <> supposed_empty then 
    raise (Bug.Bug("Pretty.toString_va", 
		   "Non-empty content in string formatter: " ^ supposed_empty))
  else
    (Format.fprintf Format.str_formatter "%a" pr_va va;
     let str = Format.flush_str_formatter () in
     str )


let toString_seqva seqva = 
  let supposed_empty = Format.flush_str_formatter () in
  if "" <> supposed_empty then 
    raise (Bug.Bug("Pretty.toString_seqva", 
		   "Non-empty content in string formatter: " ^ supposed_empty))
  else
    (Format.fprintf Format.str_formatter "%a" pr_seqva seqva;
     let str = Format.flush_str_formatter () in
     str )
*)


(************************ Printing for types ************************)

(** revchars_to_string: char list -> string   -- with order reversed*)
let revchars_to_string revchars = 
  let len = List.length revchars in 
  let str = String.create len in
  let rec charsetter ind = function 
    | [] -> ()
    | c::cs -> 
	str.[ind] <- c;
	charsetter (ind-1) cs
  in let _ = charsetter (len-1) revchars 
  in str

(** Recognizes if given seqtype is a singleton for PCCHAR *)
let check_pccharty seqty : char option = 
  match SD.seqty' seqty with
  | SD.TTree(lab, atts, seq) -> 
      (match SD.labty' lab with
      | SD.TLab hty -> 
	  (match SD.hostty' hty with
	  | SD.Tchar c -> 
	      (match SD.attsty' atts with
	      | SD.PAtt([],None) -> 
		  (match SD.seqty' seq with
		  | SD.TCat [] -> Some c
		  | _ -> None )
	      | _ -> None )
	  | _ -> None )
      | _ -> None )
  | _ -> None 

let chop_pctype_as_string (seqtys: SD.seqty list) : string * (SD.seqty list) = 
  let rec looper revchars = function
    | ((sty :: rest) as seqs) -> 
	(match check_pccharty sty with
	| Some c -> looper (c::revchars) rest
	| None -> (revchars_to_string revchars, seqs)
	)
    | [] -> 
	(revchars_to_string revchars, [])
  in
  looper [] seqtys
  



let pr_attpat ff = function 
  | (aname, id, _) -> fpr ff "@[%s=%a@]" aname Ident.print id


let pr_attsty ff x = 
  match SD.attsty' x with 
  | SD.TAttAny -> ()
  | SD.PAtt(attpats, catchopt) ->
      Printing.print_list pr_attpat " " "" ff attpats;
      match catchopt with
      | None -> ()
      | Some(id, _) -> fpr ff "%s%a" "@a" Ident.print id


(* Host types *)
let rec pr_hostty ff x = 
  match SD.hostty' x with 
  | SD.TClasRef cid -> Ident.print ff cid
  | SD.TIntfRef cid -> Ident.print ff cid
  | SD.TObject -> fpr ff "Object"
  | SD.TNull -> fpr ff "NULL"
  | SD.TBool -> fpr ff "bool"
  | SD.Tbool b -> Format.pp_print_bool ff b
  | SD.TInt -> fpr ff "int"
  | SD.Tint n -> Format.pp_print_int ff n
  | SD.TString -> fpr ff "string"
  | SD.Tstring s -> fpr ff "\"%a\"" Format.pp_print_string s
  | SD.TChar -> fpr ff "char"
  | SD.Tchar c -> Format.pp_print_char ff c
  | SD.TTag -> fpr ff "Tag"
  | SD.Ttag t -> fpr ff "Tag(%s)" t 
  | SD.TSeq -> fpr ff "Seq"
  | SD.TArray t -> fpr ff "%a[]" pr_ty t

and pr_labty ff x = 
  match SD.labty' x with  
  | SD.TLab hty -> 
      (match SD.hostty' hty with
      | SD.Ttag t -> fpr ff "%s" t
      | _ -> fpr ff "(%a)" pr_hostty hty )
  | SD.PLabBind(vid, lty) -> 
      fpr ff "%a %a," pr_labty lty Ident.print vid

and pr_seqty ff seqty = 
  match SD.seqty' seqty with
  | SD.TTree(lab, atts, seq) -> 
      (match SD.seqty' seq with
      | SD.TCat[] -> 
	  fpr ff "@[%s%a%a%s@]" 
	    "<" pr_labty lab pr_attsty atts "/>"
      | _ -> 
	  fpr ff "@[%s%a%a%s@;<0 2>%a%s@]" 
	    "<" pr_labty lab pr_attsty atts ">" pr_seqty seq "</>"
      )
  | SD.TCat seqs -> 
      (match seqs with
      | [] -> fpr ff "()"
      | _ -> pr_seqtys ff seqs  )
  | SD.TAlt seqs -> 
      fpr ff "@[(";
      Printing.print_list pr_seqty " | "  "[EMPTY-ALT]" ff seqs;
      fpr ff ")@]"
  | SD.TRep seq -> fpr ff "@[(%a)*@]" pr_seqty seq
  | SD.TSeqRef id -> Ident.print ff id
  | SD.PSeqBind(id, seqty) -> 
      fpr ff "@[%a@ as@ %a@]" pr_seqty seqty Ident.print id
  | SD.PTailBind(id, seqty) -> 
      fpr ff "@[%a@ as@ %a@]" pr_seqty seqty Ident.print id

and pr_seqtys ff (seqtys: SD.seqty list) = 
  let rec loop lst = 
    let (pctype_str, rest) = chop_pctype_as_string lst in
    begin
      (match pctype_str with
      | "" -> ()  
      | pct -> fpr ff "@[`%s`,@;<1 0>@]" pct );
      (match rest with
      | [] -> ()
      | sq::sqs -> fpr ff "@[@ %a,@;<1 0>%a@]" pr_seqty sq pr_seqtys sqs )
    end
  in 
  loop seqtys


and pr_ty ff x = 
  match SD.ty' x with
  | SD.Thost ht -> pr_hostty ff ht
  | SD.Tseq seq -> fpr ff "@[[[%a]]@]" pr_seqty seq
  | SD.PTyBind (id, ty) ->  fpr ff "@[%a@ %a@]" pr_ty ty Ident.print id
                               


let pr_ty_option ff = function
  | None -> fpr ff "VOID"
  | Some ty -> pr_ty ff ty

let toString_hostty hostty = 
  Printing.toString pr_hostty hostty

let toString_labty labty = 
  Printing.toString pr_labty labty

let toString_seqty ty = 
  Printing.toString pr_seqty ty

let toString_ty ty = 
  Printing.toString pr_ty ty

let toString_ty_option ty = 
  Printing.toString pr_ty_option ty

let toString_list_ty = function
  | [] -> ""
  | t :: r -> List.fold_left (fun acc ty -> acc^", "^(toString_ty ty))
    (toString_ty t) r



