module SD = SynDenot

module Tbl =
  Hashtbl.Make 
    (struct
      type t = Ta.state
      let equal = (==)
      let hash = Hashtbl.hash
    end)

(* count_refs : (int ref)Tbl.t -> Ta.state -> unit  *)
let rec count_refs tbl x =
  if x.Ta.desc <> None then ()
  else if Tbl.mem tbl x then
    incr (Tbl.find tbl x)
  else
    begin
      Tbl.add tbl x (ref 1);
      List.iter (fun (_,_,x1,x2,_) ->
	count_refs tbl x1;
	count_refs tbl x2)
	x.Ta.next
    end

(* recov_rec : (Ident.id)Tbl.t -> Ta.state -> SD.seqty *)
let rec recov_rec vtbl x =
  match x.Ta.desc with
    Some sty -> sty
  | None ->
      if Tbl.mem vtbl x then 
	SD.seqty (SD.TSeqRef (Tbl.find vtbl x))
      else
	let y = recovl vtbl x.Ta.next in
	if x.Ta.final then SD.alt [SD.sty_epsilon; y]
	else y
	
(* recovl : Ident.id Tbl.t -> (Label.t * Ta.state * Ta.state * 'a) -> SD.seqty*)
and recovl vtbl xl =
  SD.alt 
    (List.map (fun (c,a,x1,x2,_) ->
      SD.cat
	[SD.tree c a (recov_rec vtbl x1);
	 recov_rec vtbl x2])
       xl)

(* recov_rec' : Ident.id Tbl.t -> Ta.state -> SD.t *)
and recov_rec' vtbl x =
  let y = recovl vtbl x.Ta.next in
  if x.Ta.final then SD.alt [SD.sty_epsilon; y]
  else y

(* recov : Ta.state list -> SD.seqty *)
let rec recov x =
  let tbl = Tbl.create 121 in
  List.iter (count_refs tbl) x;
  let vtbl = Tbl.create 121 in
  Tbl.iter (fun x c ->
    if !c > 1 then Tbl.add vtbl x (Ident.create_internal "X"))
    tbl;
  Tbl.iter (fun x xtypid ->
    define_name xtypid (recov_rec' vtbl x))
    vtbl;
  let res = SD.alt (List.map (recov_rec' vtbl) x) in
  res

and define_name xtypid htype = 
  Store.id_to_seqty := Ident.add !Store.id_to_seqty (xtypid, htype)


	  

