module SD = SynDenot

type datctx = {
    dc_topclass: Ident.id;
    dc_parent: Ident.id Ident.tbl;
    dc_implements: (Ident.id list) Ident.tbl;
    dc_seqdef: SynDenot.seqty Ident.tbl;
    dc_seqdef_untranslated: SynSource.seqtbl;
    dc_any_id: Ident.id;
}
 
let setup (datctx: datctx) = 
  Store.reset();
  Store.topclass := datctx.dc_topclass;
  Store.parent := datctx.dc_parent;
  Store.implements := datctx.dc_implements;

  (* need to do this after the classes and interfaces have been set *)
  Store.id_to_seqty := 
    Ident.fold
      (fun seqtbl (id, (nctx, seqty)) -> 
	Ident.add seqtbl (id, Src2denot.seqty nctx seqty))
      datctx.dc_seqdef datctx.dc_seqdef_untranslated;
  Store.any_id := datctx.dc_any_id;
  (* set the type of Any in the engine *)
  Store.any_seqty_def := Ident.find_same !Store.id_to_seqty !Store.any_id;
  let any_ta = Buildta.build_ta !Store.any_seqty_def in
  Store.any_ta := any_ta


let parent cid =
  Ident.find_same !Store.parent cid

let implements cid =
  Ident.find_same !Store.implements cid

let sub ty1 ty2 = Sub.sub_ty ty1 ty2

let bsub = Sub.bsub_ty

let maximum tys = Sub.maximum_ty tys


let tyMatch ty pa env = 
  match TyMatch.tyMatch_ty ty pa with
  | None -> None
  | Some(isec, bind) -> 
      Some(isec, IdBind.fold TyEnv.addVar env bind)


(*## So far, this is an approximation: instead of the correct rule, we
   just match Any against the pattern -- this does not take the input
   type into account in determining the types of binders. *) 
let styChop sty spa env = 
  if not (Sub.bsub_seqty sty (SD.seqty (SD.TRep spa))) 
  then None 
(*    else match TyMatch.tyMatch_seqty SD.sty_any spa with  *)
  else match TyMatch.tyMatch_seqty sty (SD.cat [spa;SD.sty_any]) with 
  | None -> None 
  | Some (_, bind) -> Some(IdBind.fold TyEnv.addVar env bind)



