(** "Semantics" of Xtatic: provides functions that are sufficient for 
   implementing the "hard parts" of a typechecker and an evaluator. *)

(** The module is stateful in two ways: 
    (1) before any of the functions can be used, the module must be
        initialized by executing 'setup';
    (2) results of the previous calls to computation-intensive
        functions, like 'sub', are cached (memoized).
*)


(** datctx provides data sufficient to initialize this module; 
    this is the only intended use of it. *)
type datctx = {
    dc_topclass: Ident.id;
    dc_parent: Ident.id Ident.tbl;
    dc_implements: (Ident.id list) Ident.tbl;
    dc_seqdef: SynDenot.seqty Ident.tbl; 
    dc_seqdef_untranslated: SynSource.seqtbl;
    dc_any_id: Ident.id;  
}


(** Module initialization.  None of the functionality is available
    before this function is executed.  *)
val setup: datctx -> unit

val parent: Ident.id -> Ident.id

val implements: Ident.id -> Ident.id list

val sub: SynDenot.ty -> SynDenot.ty -> Typath.t Result.t

val bsub: SynDenot.ty -> SynDenot.ty -> bool

val maximum: SynDenot.ty list -> SynDenot.ty option


(** Operations for "whole-match" pattern matching 
   (as in 'let' and 'match' statements). *)
    

val tyMatch: SynDenot.ty -> SynDenot.pa -> 
      (TyEnv.t -> (SynDenot.ty * TyEnv.t) option)
      (** Returns None if no match detected, otherwise returns an
      extended type environment and the subtype of the input type that
      was captured by the pattern.  *)
    
(** Operations for "iterative match" (as in 'foreach' statement).  *)


val styChop: SynDenot.seqty -> SynDenot.seqpa -> (TyEnv.t -> TyEnv.t option)
  (** Checks if the type can be "exhausted" by chopping, one-by-one,
       its initial segments matching the pattern.  In case of success,
       returns the environment extended by the pattern's bindings.
       Each binding has the type which is the _union_ of types from
       all possible choppings.  *)
   (*## Check the implem; this description is the ultimate intention,
   but the current implem can be simpler .... *)
   (*## Figure out a way to give more info about the failing outcome. *)


