module SS = SynSource
module CN = Cnaming
module XN = Xnaming
module N = Naming

exception StartClassNotFound
exception ElaborationError of Finfo.t * string


(***************  Extension functions *********************)

(*--VG## It is ugly that extend_top_space and extend_cspace have 
    largely the same code... *)

let rec extend_top_space (nspace: N.space) tunit : N.space= 
  let (inf_tunit, tunit') = tunit in 
  let extender ns decl = 
    let (dn, def) = decl in  
    let dn_str = Finfo.drop dn in
    try (match def with
    | SS.TMemClas _ ->
	{ ns with 
	    N.cs = CN.spaceAlterChildLeaf ns.N.cs dn_str 
		     (function 
			| None -> (Ident.create_located [] dn_str, CN.LClas)
			| Some cid -> raise (CN.NameAlreadyUsed dn_str)) }
    | SS.TMemIntf _ -> 
	{ ns with 
	    N.cs = CN.spaceAlterChildLeaf ns.N.cs dn_str 
		     (function 
			| None -> (Ident.create_located [] dn_str, CN.LIntf)
			| Some cid -> raise (CN.NameAlreadyUsed dn_str)) }
    | SS.TMemSpace cu -> 
	{ ns with 
	    N.cs = CN.spaceAlterChildSpace ns.N.cs dn_str
		(function  
		   | None   -> extend_cspace ([dn_str]) CN.spaceEmpty cu
		   | Some n -> extend_cspace ([dn_str]) n cu  ) }

    | SS.TMemSeq _ ->  
	{ ns with N.xs = 
	    XN.spacesAddMember ns.N.xs (XN.NoUri, dn_str) (Ident.create dn_str)
	}
    | SS.TMemXspace xu -> 
	if dn_str = "" then raise
	  (ElaborationError(Finfo.get dn, 
		"Empty string cannot be a URI for an XML namespace"))
	else
	  { ns with 
	      N.xs = extend_xspace ns.N.xs dn_str xu }
	) 
    with 
      | CN.NameAlreadyUsed nm -> raise 
	 (ElaborationError(Finfo.get dn,
			   "The name '" ^ nm ^ "' is already used for " ^
			   "another item in global C# namespace."))
      | XN.NameAlreadyUsed nm -> raise
	  (ElaborationError(Finfo.get dn, 
			    "The name " ^ nm ^ " is already used for " ^
			    "another item in the global XML namespace"))
  in 
  List.fold_left  extender  nspace  tunit'.SS.tunit_decls


(* 'path' is the namespace path to the "current context", which is the 
   location of 'nspace'.  
*)
and extend_cspace (path: string list) (cspace: CN.space) (cunit: SS.cunit) 
  : CN.space = 
  let (inf_cunit, cunit') = cunit in 
  let extender cs decl = 
    let (dn, def) = decl in  
    let dn_str = Finfo.drop dn in
    try (match def with
    | SS.CMemClas _ ->  
	CN.spaceAlterChildLeaf cs dn_str 
	(function 
	   | None -> (Ident.create_located path dn_str, CN.LClas)
	   | Some(id, _) -> raise (CN.NameAlreadyUsed dn_str))
    | SS.CMemIntf _ -> 
	CN.spaceAlterChildLeaf cs dn_str 
	(function 
	   | None -> (Ident.create_located path dn_str, CN.LIntf)
	   | Some(id, _) -> raise (CN.NameAlreadyUsed dn_str))
    | SS.CMemSpace cu -> 
	CN.spaceAlterChildSpace cs dn_str
	(function  
	   | None   -> extend_cspace (path @ [dn_str]) CN.spaceEmpty cu
	   | Some n -> extend_cspace (path @ [dn_str]) n cu  )
	) 
    with CN.NameAlreadyUsed nm -> raise 
      (ElaborationError(Finfo.get dn,
			"The name '" ^ nm ^ "' is already used for " ^
			"another item in this declaration space."))
  in 
    List.fold_left  extender  cspace  cunit'.SS.cunit_decls

and extend_xspace (xspace: XN.spaces) (uri: string) (xunit: SS.xunit) 
  : XN.spaces = 
  let (inf_xunit, xunit') = xunit in
  let extender xs decl = 
    let (dn, def) = decl in  
    let dn_str = Finfo.drop dn in
      try (match def with
	  | SS.XMemSeq _ ->  
	      XN.spacesAddMember xs (XN.Uri uri, dn_str) (Ident.create dn_str)
	  )
      with
	| XN.NameAlreadyUsed nm -> raise
	    (ElaborationError(Finfo.get dn, 
			      "The name " ^ nm ^ " is already used for " ^
			      "another item in this XML namespace"))
  in
    List.fold_left extender xspace xunit'.SS.xunit_decls



(* prepare_imports: extends nimports with udir, where
     nctx -- used to resolve rhs of udir
     host_cspace -- used to check that lhs of udir does not shadow its member
*)
let prepare_imports (nctx: N.ctx) (host_cspace: CN.space) 
  (udirs: SS.usedir list) : N.imports  = 
  let rec loop cimports ximports = function
    | [] -> { N.ci = cimports;  N.xi = ximports; }
    | (inf, udir') :: rest -> (
	match udir' with 
	  | SS.CAlias (dn, refname) -> 
	      let alias = Finfo.drop dn in
              (* Check that alias is not used as a declared name 
		 in the host namespace *)
		(try 
		   ignore (CN.spaceFindChildMem host_cspace alias);
		   raise 
		     (ElaborationError (Finfo.get dn,
			 "Name '" ^ alias ^ 
		"' is already used to declare a member in this namespace " ^
		   "(perhaps, in a different namespace-body)."))
		 with Not_found -> () ); (* ok, name not used, proceed *)
		(* Resolve the path aliased *)
		let referent = N.resolve_member nctx refname in
		let newimport = CN.AliasMember(alias,referent) in
		  loop  (newimport :: cimports) ximports rest

	  | SS.COpen refname -> 
	      (match N.resolve_member nctx refname with
		 | CN.Leaf _ -> raise 
		     (ElaborationError
			(inf, "Using-open directive must refer to " ^ 
			   "a namespace, not to a type."))
		 | CN.Space ns -> 
		     loop (CN.OpenSpace(ns) :: cimports) ximports rest
	      )

	  | SS.XAlias (dn, qualname) -> 
	      let alias = Finfo.drop dn in
	      let referent = N.resolve_seqty nctx qualname in
	      let newimport = XN.AliasMember(alias, referent) in 
		loop cimports (newimport::ximports) rest

	  | SS.XPrefix(dn, uri) -> 
	      let pref = Finfo.drop dn in
	      let (inf_uri, str_uri) = uri in
	      let newimport = 
		if str_uri = "" 
		then XN.UnbindPrefix(XN.Pref pref) 
		else XN.BindPrefix(XN.Pref pref, XN.Uri str_uri)
	      in
		loop cimports (newimport :: ximports) rest

	  | SS.XOpen uri -> 
	      let (inf_uri, str_uri) = uri in
	      let newimport = 
		if str_uri = "" 
		then XN.UnbindDefault
		else XN.BindDefault(XN.Uri str_uri)
	      in
		loop cimports (newimport :: ximports) rest
      ) (*match*)
  in loop [] [] udirs


(**************************************************************************)

let rec add_top_unit (global_nctx: N.ctx) (global_cspace: CN.space)
  (prog: SS.prog) (tunit: SS.tunit) : SS.prog = 
  let (inf_tunit, tunit') = tunit in 
  let usedirs = tunit'.SS.tunit_usedirs in
  let nimports = prepare_imports global_nctx global_cspace usedirs in
  let curr_nctx = N.ctxApplyImports global_nctx nimports in 

  (* Process each member of tunit body in the curr_nctx context ... *)
  (* ... a class or a seqtype gets bundled with its context *)
  (* ... a namespace is processed recursively *)
  List.fold_left (fun prg (dn, def) -> 
    let str_dn = Finfo.drop dn in
    match def with 
    | SS.TMemClas clas -> 
	{ prg with SS.cls = 
	    match N.resolve_leaf curr_nctx [dn] with 
	      | id, CN.LClas -> Ident.add prg.SS.cls (id, (curr_nctx, clas))
	      | _ -> assert false  } 
    | SS.TMemIntf intf -> 
	{ prg with SS.intfs = 
	    match N.resolve_leaf curr_nctx [dn] with 
	      | id, CN.LIntf -> Ident.add  prg.SS.intfs (id, (curr_nctx, intf))
	      | _ -> assert false  }
    | SS.TMemSpace cu -> 
	let cu_nspace = 
	  match CN.spaceFindChildMem global_cspace str_dn with 
	  | CN.Leaf _ -> assert false (* we know this is namespace *)
	  | CN.Space s -> s
	in add_cunit curr_nctx cu_nspace prg cu

    | SS.TMemSeq seqty -> 
	{ prg with SS.seq = 
	    let id = XN.ctxLocateMember curr_nctx.N.xc (XN.NoUri, str_dn) in
	      Ident.add  prg.SS.seq (id, (curr_nctx, seqty)) }
    | SS.TMemXspace xu ->
	add_xunit curr_nctx str_dn prg xu 
		 )
    prog tunit'.SS.tunit_decls


(** Extends the program prog by types defined in the C# unit
   cunit.  prior_nctx is the naming context of the parent
   namespace-body of cunit (needed for resolving the 'using' directives
   in cunit), and curr_nspace is the namespace for which cunit
   constitutes one (of possibly many) namespace-bodies. 
*)
and add_cunit (prior_nctx: N.ctx) (curr_nspace: CN.space)
  (prog: SS.prog) (cunit: SS.cunit) : SS.prog = 
  let (inf_cunit, cunit') = cunit in 
  let usedirs = cunit'.SS.cunit_usedirs in
  let nimports = prepare_imports prior_nctx curr_nspace usedirs in
  let curr_nctx = N.ctxApplyImports prior_nctx nimports in 

  (* Make current namespace's members available for direct reference.
     This differs from add_top_unit, where they were already available *)
  let curr_nctx = N.ctxApplyImports curr_nctx 
		    { N.ci = [CN.OpenSpace curr_nspace];  N.xi=[]; } in

  (* Process each member of cunit body in the curr_nctx context ... *)
  (* ... a class or a seqtype gets bundled with its context *)
  (* ... a namespace is processed recursively *)
  List.fold_left (fun prg (dn, def) -> 
    let str_dn = Finfo.drop dn in
    match def with 
    | SS.CMemClas clas -> 
	{ prg with SS.cls = 
	    match N.resolve_leaf curr_nctx [dn] with 
	      | id, CN.LClas -> Ident.add prg.SS.cls (id, (curr_nctx, clas))
	      | _ -> assert false  } 
    | SS.CMemIntf intf -> 
	{ prg with SS.intfs = 
	    match N.resolve_leaf curr_nctx [dn] with 
	      | id, CN.LIntf -> Ident.add  prg.SS.intfs (id, (curr_nctx, intf))
	      | _ -> assert false  }
    | SS.CMemSpace cu -> 
	let cu_nspace = 
	  match CN.spaceFindChildMem curr_nspace str_dn with 
	  | CN.Leaf _ -> assert false (* we know this is namespace *)
	  | CN.Space s -> s
	in add_cunit curr_nctx cu_nspace prg cu
		 )
    prog cunit'.SS.cunit_decls


and add_xunit (prior_nctx: N.ctx) (curr_uri: string)
  (prog: SS.prog) (xunit: SS.xunit) : SS.prog = 
  let (inf_xunit, xunit') = xunit in 
  let usedirs = xunit'.SS.xunit_usedirs in
  let nimports = prepare_imports prior_nctx CN.spaceEmpty usedirs in
  let curr_nctx = N.ctxApplyImports prior_nctx nimports in 

  List.fold_left (fun prg (dn, def) -> 
    let str_dn = Finfo.drop dn in
      match def with 
	| SS.XMemSeq seqty -> 
	    { prg with SS.seq = 
		let id = XN.ctxLocateMember curr_nctx.N.xc (XN.Uri curr_uri, str_dn) in
		  Ident.add  prg.SS.seq (id, (curr_nctx, seqty)) }
		 )	
    prog xunit'.SS.xunit_decls
