(**  Subtyping and subtyping-based functions like maximums and minimums *)

(*## All this has to be in one module, since, e.g., subtyping on
   seqtypes depends on max/min for label types.  However, the module
   can be split following the sections below. *)


(************************** Host types ***************************)

val sub_class: Ident.id -> Ident.id -> bool

val sub_hostty: SynDenot.hostty -> SynDenot.hostty -> Typath.t Result.t 
val bsub_hostty: SynDenot.hostty -> SynDenot.hostty -> bool

val minimums_hostty: SynDenot.hostty list -> SynDenot.hostty list
val maximums_hostty: SynDenot.hostty list -> SynDenot.hostty list

val minimum_hostty: SynDenot.hostty list -> SynDenot.hostty option
val maximum_hostty: SynDenot.hostty list -> SynDenot.hostty option

val join_hostty: SynDenot.hostty -> SynDenot.hostty -> SynDenot.hostty

(************************** Label types ***************************)

val sub_labty: SynDenot.labty -> SynDenot.labty -> Typath.t Result.t
val maximums_labty: SynDenot.labty list -> SynDenot.labty list
val maximum_labty: SynDenot.labty list -> SynDenot.labty option
val minimum_labty: SynDenot.labty list -> SynDenot.labty option


(**************************  Automata ***************************)

val sub_ta:  Ta.state list -> Ta.state list -> Typath.t Result.t 
(**VG NB: correctness relies on the argument state lists being ordered
   by state.Ta.num (since sub uses Ta.state_set_diff) -- this must
   be ensured by the clients.  *)


(************************** Sequence types ***************************)

val sub_seqty: SynDenot.seqty -> SynDenot.seqty -> Typath.t Result.t
val bsub_seqty: SynDenot.seqty -> SynDenot.seqty -> bool

(**************************  Full types ***************************)

val sub_ty: SynDenot.ty -> SynDenot.ty -> Typath.t Result.t 
val bsub_ty: SynDenot.ty -> SynDenot.ty -> bool
val maximum_ty: SynDenot.ty list -> SynDenot.ty option

