(***********************************************************************
       Defines syntax of the "data model": values and types.
***********************************************************************)

(** See the modules included below for the real definitions.  This
    module just combines them together and adds a few convenience
    functions.  However, only this module is directly used in the rest
    of the program.  *)

(** Types: syntax and basic invariant-enforcing functions. *)
include SynTyp.PUBLIC

(** Patterns --  defined mostly for "documenting" purposes. *)
type hostpa = hostty
type pa = ty
type labpa = labty
type seqpa = seqty

val whole_hostty: hostty -> hostty
val whole_ty: ty -> ty
    (** e.g., "whole type" of 'Tint 3' is TInt, and 'Tseq ...' is TSeq *) 

val clasOf_ty: ty -> Ident.id

val isHostGEQSeq: hostty -> bool
val isHostLEQSeq: hostty -> bool


(** Convenience constructors for labels *)

val basiclab: hostty -> labty

val is_emptylab: labty -> bool
    (** Always false -- labels are singletons *)

val crack_labty: labty -> hostty * Ident.id list 
   (** For a labty, gives its underlying hostty and the list of bound
      variables.  The variables are returned sorted
      w.r.t. Pervasives.compare *)
val slap_labty: Ident.id list -> hostty -> labty
   (** Reverse of crack_labty *)

val attsty_any: attsty
val attsty_empty: attsty


(** Convenience constructors for sequence types.  *)
val tree: labty -> attsty -> seqty -> seqty
val leaf: hostty -> seqty
   (** A leaf is a tree with empty list of children and a simple label, 
       consisting of a single positively affirmed type.  *)

val cat: seqty list -> seqty
val alt: seqty list -> seqty 
val rep: seqty -> seqty
val seqref: Ident.id -> seqty
val seqbind: Ident.id -> seqty -> seqty
val tailbind: Ident.id -> seqty -> seqty


(** Convenience type definitions for often-used types *)
val hty_ClasRef: Ident.id -> hostty 
val hty_IntfRef: Ident.id -> hostty 
val hty_Object: hostty 
val hty_Null: hostty 
val hty_Bool: hostty 
val hty_Int: hostty 
val hty_Char: hostty 
val hty_String: hostty 
val hty_Tag: hostty 
val hty_Seq: hostty 

val ty_ClasRef: Ident.id -> ty
val ty_IntfRef: Ident.id -> ty
val ty_Object: ty
val ty_Null: ty 
val ty_Bool: ty 
val ty_Int: ty 
val ty_Char: ty
val ty_String: ty 
val ty_Tag: ty 
val ty_Seq: ty 

val sty_anyone : seqty
val sty_any : seqty
val sty_xml : seqty
val sty_pcdata : seqty
val sty_attseq : seqty

val ty_empty : ty
val ty_epsilon : ty
val ty_anyone : ty
val ty_any : ty
val ty_xml : ty
val ty_pcdata : ty
val ty_attseq : ty

(** Printing *)
val pr_hostty: Format.formatter -> hostty -> unit
val pr_ty: Format.formatter -> ty -> unit
val pr_seqty: Format.formatter -> seqty -> unit
val pr_labty: Format.formatter ->  labty -> unit

val toString_hostty: hostty -> string
val toString_labty: labty -> string
val toString_seqty: seqty -> string
val toTypeStampString_seqty: seqty -> string
val toString_ty: ty -> string
val toString_ty_option: ty option -> string
val toString_list_ty: ty list -> string


(** Checking existence of binders *)
val binders_ty: ty -> bool
val binders_seqty: seqty -> bool
val binders_labty: labty -> bool
val binders_attsty: attsty -> bool

val set_module_name: string -> unit
