(* ta.mli - tree automata *)

(*VG 'act' seems to correspond to different kinds of low-level variables 
   that can be generated by a variable bound in a pattern  *)
type act = 
  | Begin of Ident.id
  | End of Ident.id
  | Rest of Ident.id
  | Fixed of Ident.id * int
type pre = act list

type binder =
  | BNAttr of Ident.id
  | BNLabel of Ident.id * SynDenot.hostty
  | BNSeq of act

type state =
  { num : int;     (*VG was: id *)
    mutable final : bool;
    mutable finpre : pre;
    mutable next : trans list;
    mutable desc : SynDenot.seqty option;
    mutable empty : bool option;
    mutable binders : binder list option; (* closure of finpre AND label and attr binders *)
    mutable preclos : pre; (* closure of finpre sets *)
       (*VG: hypothesis: preclos is needed only for Infer *)
(*--VG not used...
    mutable nochk : bool;
    mutable exact : bool option;
    mutable always : bool;
*)
  }
and trans = (SynDenot.labty * SynDenot.attsty * state * state * pre) 

val number_of_states : unit -> int
val number_of_empty_states : unit -> int

val index_to_state : int -> state

val compare_state : state -> state -> int
val compare_trans : trans -> trans -> int
val compare_pre : pre -> pre -> int
val compare_act : act -> act -> int

(* create_state [final next desc] *)
val create_state :
    bool -> trans list -> SynDenot.seqty option -> state

val state_set_next : state list -> trans list
val state_set_final : state list -> bool

val empty : state (* pattern: [[any]], wildcard state *)
val epsilon : state (* pattern: () *)

(*?? val state_set_is_obviously_any : state list -> bool *)

val compare_state_sets : state list -> state list -> int
val union_state_sets : state list -> state list -> state list
val diff_state_sets : state list -> state list -> state list

val is_empty : state -> bool
val is_empty_l : trans list -> bool

val elim_empty1 : state -> bool
val elim_empty : state list -> state list
val elim_empty_l : 
    trans list -> trans list

val has_binders : state -> bool
val get_binders : state -> binder list

val print : Format.formatter -> state list -> unit


(***************** Map and Set modules ******************)
(*** that use state's as indices   *****)

module StSet : Set.S with type elt = state
module StMap : Map.S with type key = state

module StStSet : Set.S with type elt = state * state
module StStMap : Map.S with type key = state * state
(* used in Isect, Infer *)

module StLsMap : Map.S with type key = state * state list

module IntLsLsSet : Set.S with type elt = int * state list * state list
module IntLsLsMap : Map.S with type key = int * state list * state list
(* used in Sub *)

