(*## *)

type loctype = Var  |  Fld

type t = {
    status: SynOp.status;
    this: Ident.id;
    super: Ident.id;
    vars: (Ident.id * (loctype * SynDenot.ty)) list;
    returns: SynDenot.ty list;
  } 

let empty_static = 
  { status = SynOp.Static;
    this = Ident.dummyid;
    super = Ident.dummyid;
    vars = [];
    returns = [];
  } 

let empty_instance clasid parentid = {
  status = SynOp.Instance;
  this = clasid; 
  super = parentid;
  vars = [];
  returns = [];
} 

let addVar env (vid, typ) = 
  { status = env.status;
    this = env.this;
    super = env.super;
    vars =  (vid, (Var, typ)) :: env.vars;
    returns = env.returns;
  } 

let addFld env (vid, (status, typ)) = 
  (* we do not add static fields to the environment *)
  match status with
    | SynOp.Static -> env 
    | SynOp.Instance -> 
	{ status = env.status;
	  this = env.this;
	  super = env.super;
	  vars =  (vid, (Fld, typ)) :: env.vars;
	  returns = env.returns;
	} 


let find env str = 
  let (id, (loc, typ)) = 
    List.find (fun (id, _) -> str = (Ident.name id))  env.vars
  in (loc, id, typ)


let lookup env vid  = List.assq vid env.vars 
    (** NB: List.assq checks _physical_ equality of ids. *)

let get_this env = env.this

let get_super env = env.super


let set_return env va = 
  { status = env.status;
    this = env.this;
    super = env.super;
    vars =  env.vars;
    returns = va :: env.returns;
  } 

let is_returned env = ([] <> env.returns)

let get_returns env = env.returns
