module SD = SynDenot

type tybind = SynDenot.ty IdBind.t 


let rec tyMatch_hostty hty hpa  = 
  let t_sub_p = Sub.bsub_hostty hty hpa in
  let p_sub_t = Sub.bsub_hostty hpa hty in
  match t_sub_p,  p_sub_t  with
  | true,  true  -> Some (hty, IdBind.empty)
  | true,  false -> Some (hty, IdBind.empty)
  | false, true  -> Some (hpa, IdBind.empty)
  | false, false -> None
  (* The last clause relies on the (contraposition of the) property:
  if two host types intersect, then one is a subtype of another.  *)

let fpr = Format.fprintf
let outf = !Pref.output_formatter

let tyMatch_seqty sty spa = 
  let ta_sty = match Ta.elim_empty [Buildta.build_ta sty] with
  | [t] -> t | _ -> assert false in
  let ta_spa =  Buildta.build_ta spa in
  ignore(Ta.elim_empty1 ta_spa);

(*    Ta.print outf [ta_sty]; *)
(*    fpr outf "--- end of ta_sty@."; *)
(*    Ta.print outf [ta_spa]; *)
(*    fpr outf "--- end of ta_spa@."; *)

  let (ta_mtch, tyBind, stTbl) = Infer.tyMatch ta_sty ta_spa 
  in
  if Result.boolOf (Sub.sub_ta [ta_mtch] [Buildta.build_ta SD.sty_empty]) 
                                         (*## can this be precomputed? *)
      (* Here we rely on fact that ta_mtch is empty if there was actually no
         match of the type against the pattern. *)
  then None
  else
    let sty_mtch = Recov.recov [ta_mtch] in 
(*
    (Debug.fprOne_sty "Argument stype: %a@." sty); 
    (Debug.fprOne_sty "Pattern stype: %a@." spa); 
    (Debug.fprOne_sty "Matching stype: %a@." sty_mtch); 
    (Debug.fprOne_sty "Difference stype: %a@." sty_diff); 
*)
    let bind = Hashtbl.fold 
	(fun id ta b -> 
	  let sty = Recov.recov [ta] in
	  IdBind.add b (id, (SD.ty (SD.Tseq sty))) )
	stTbl tyBind
    in Some(sty_mtch, bind)



let rec tyMatch_ty ty pa =
  match (SD.ty' ty, SD.ty' pa) with 

  | SD.PTyBind(_, t1), _ -> 
      tyMatch_ty t1 pa
  (* Just skip the binder at this point, and
  continue matching the underlying type. *)

  (* For a binding pattern, make recursive calls until a non-binder
  pattern is reached, and use the encountered binders to extend the
  environment: *)
  | _, SD.PTyBind(vid, p1)  ->
      (match tyMatch_ty ty p1 with
      |	None -> None
      |	Some (isec, tybind) -> 
	  Some (isec, IdBind.add tybind (vid, isec))
      )

  | SD.Thost ht, SD.Thost hp ->  
	(match tyMatch_hostty ht hp with
	| None -> None
	| Some (isec, tybind) -> 
	    Some (SD.ty (SD.Thost isec), tybind)
	)

  (* For sequence patterns, use type matching from seqdenot *)
  | SD.Tseq st, SD.Tseq sp ->
      (match tyMatch_seqty st sp with
      |	None -> None
      |	Some (sisec, tybind) -> 
	  Some (SD.ty (SD.Tseq sisec), tybind)
      )

  (* Sequence type against host pattern *)
  | SD.Tseq _, SD.Thost ht -> 
      if SD.isHostGEQSeq ht 
      then Some(ty, IdBind.empty) else None

  (* Host type against sequence pattern *)
  | SD.Thost ht, SD.Tseq sp -> 
      if (SD.isHostLEQSeq ht) 
      then match tyMatch_seqty !Store.any_seqty_def sp with
      |	None -> None
      |	Some (sisec, tybind) -> 
	  Some (SD.ty (SD.Tseq sisec), tybind)
      else None

