open List

(* ------------------ input, output functions ------------------ *)

let call_with_input_file fname k =
  let ip = open_in fname in
  let res = k ip in
  close_in ip;
  res

let call_with_output_file fname k =
  let op = open_out fname in
  let res = k op in
  close_out op;
  res

let read_file fname =
  call_with_input_file fname
    (fun ip ->
       let len = in_channel_length ip in
       let str = String.make len ' ' in
       really_input ip str 0 len;
       str)

let write_file fname str =
  call_with_output_file fname
    (fun op -> output_string op str)


(* ------------------ list functions  ------------------ *)

let rec ormap f = function
  | [] -> false
  | x::rest -> f x || (ormap f rest)

(*  let andmap = List.for_all *)

let iota_ls (ls : 'a list) : int list =
  let r = ref 0 in List.map (fun _ -> incr r; !r) ls

let rec union2 (xs : 'a list) (ys : 'a list) : 'a list =
  match xs with
    | [] -> ys
    | x::xs -> if List.mem x ys then union2 xs ys else x :: union2 xs ys

let rec setify = function
  | [] -> []
  | x :: ls ->
      if List.mem x ls then setify ls
      else x :: setify ls

let rec map3 f ls1 ls2 ls3 =
  match (ls1,ls2,ls3) with
    | [],[],[] -> []
    | x1::ls1, x2::ls2, x3::ls3 -> f x1 x2 x3 :: map3 f ls1 ls2 ls3
    | _ -> assert false

let rec split3 (ls : ('a * 'b * 'c) list) : 'a list * 'b list * 'c list =
  match ls with
    | [] -> [],[],[]
    | (x,y,z)::ls ->
        let xs,ys,zs = split3 ls in
        x::xs, y::ys, z::zs

let split_where pred ls1 ls2 =
  let rec loop ls1 ls2 r1 r2 r3 r4 = match ls1 with
    | [] -> rev r1, rev r2, rev r3, rev r4
    | s::ls1 ->
        if pred s then loop ls1 (tl ls2) (s::r1) r2 (hd ls2 :: r3) r4
        else loop ls1 (tl ls2) r1 (s::r2) r3 (hd ls2 :: r4)
  in loop ls1 ls2 [] [] [] []

let split_before pred _ls1 _ls2 =
  let rec loop ls1 ls2 pre1 pre2 = match ls1 with
    | [] -> _ls1, [], _ls2, []
    | s::ls1 ->
        if pred s then rev pre1, s::ls1, rev pre2, ls2
        else loop ls1 (tl ls2) (s::pre1) (hd ls2 :: pre2)
  in loop _ls1 _ls2 [] []

let extract_where pred ls1 ls2 =
  let _,_,res,_ = split_where pred ls1 ls2 in
  res

let extract_before pred ls1 ls2 =
  let _,_,res,_ = split_before pred ls1 ls2 in
  res

let extract_after pred ls1 ls2 =
  let _,_,_,res = split_before pred ls1 ls2 in
  res

let rem_inner (ls : 'a list) (i : int) : 'a * 'a list =
  let rec loop ls i before =
    if i==0 then hd ls, rev before @ tl ls
    else loop (tl ls) (i-1) (hd ls::before)
  in loop ls i []


(* ------------------ array functions  ------------------ *)

let find_pos (x : 'a) (ls : 'a array) : int =
  let res = ref (-1) in
  for i = 0 to Array.length ls - 1 do
    if ls.(i) == x then res := i
  done;
  !res


(* ------------------ miscellaneous ------------------ *)

let ifopt sk alt = function
    None -> alt
  | Some(x) -> sk x

let doopt sk = function
    None -> None
  | Some(x) -> Some (sk x)


(* ------------------ symbol table ------------------ *)

let (_intern_table_ : (string,string) Hashtbl.t) = Hashtbl.create 97

let intern str =
    try
      Hashtbl.find _intern_table_ str
    with Not_found ->
      Hashtbl.add _intern_table_ str str;
      str
