(**  Support for XML namespaces.  *)

type mid = Ident.id     (* member id *)
type locname = string   (* local name *)

(** Qualified (prefixed) names *)
type pref = NoPref | Pref of string
type qname = pref * locname
  (** NB: NoPref is not (Pref ""); the latter is treated as a regular prefix
     (even though a programmer cannot specify such) *)

(** Expanded names *)
type uri = NoUri | Uri of string
type expname = uri * locname
  (** NB: NoUri is not (Uri "") *)

exception AmbiguousName of string
exception NameAlreadyUsed of string 


(*** Collection of types (or other members) defined in XML namespaces ******)

type spaces

val spacesEmpty: spaces

val spacesAddMember: spaces -> expname -> mid -> spaces

val spacesFindMember: spaces -> expname -> mid
      (** raises Not_found *)


(*****  Imports  *****)

type import =
  | BindPrefix of pref * uri       (* declaring a prefix for a namespace *)
  | UnbindPrefix of pref           (* undeclaring a namespace prefix *)
  | BindDefault of uri             (* declaring the default namespace *)
  | UnbindDefault                  (* undeclaring the default namespace *)
  | AliasMember of locname * mid   (* an alias for a namespace member *)


(***** Context can resolve a prefix to a namespace and a pname to member *****)

type ctx

(** Puts spaces as the collection of known namespace-member definitions. *)
val ctxInit: spaces -> ctx
  
val ctxApplyImports: ctx -> import list -> ctx

val ctxFindUri: ctx -> pref -> uri
  (** raises Not_found *)

val ctxResolveQname: ctx -> qname -> mid
  (** raises Not_found *)

val ctxLocateMember: ctx -> expname -> mid
  (** raises NotFound *)


