
(* 
 *  The lexical analyzer
 *)

{

module P = XtParser

exception Lexing of Finfo.t * string

(* stse 2003-0605 fixme
   c# comment does not nest
   ref: C# Spec: 2.3 Lexical analysis *)
let commentDepth    = ref 0
(* let labelsetJustStarted = ref false  *)

let info lexbuf =
  Finfo.make_finfo (!ParseGlobals.input_file) 
                   (Lexing.lexeme_start lexbuf) (Lexing.lexeme_end lexbuf)

let stringBuffer = ref (String.create 2048)
let stringEnd = ref 0

let resetStr () = stringEnd := 0

let addStr ch =
  let x = !stringEnd in
  let buffer = !stringBuffer
in
  if x = String.length buffer then
    begin
      let newBuffer = String.create (x*2) in
      String.blit buffer 0 newBuffer 0 x;
      String.set newBuffer x ch;
      stringBuffer := newBuffer;
      stringEnd := x+1
    end
  else
    begin
      String.set buffer x ch;
      stringEnd := x+1
    end

let getStr () = String.sub (!stringBuffer) 0 (!stringEnd)

let drop_startchars n str = 
  String.sub str n ((String.length str) - n)


}   (**** end functions section *****)


let space = [' ' '\t' '\n' '\r']

let float_literal =
  ['0'-'9']+ ('.' ['0'-'9']*)? (['e' 'E'] ['+' '-']? ['0'-'9']+)?

let tag_name = 
  ['A'-'Z' 'a'-'z' '_' ':']
  ['A'-'Z' 'a'-'z' '_' ':' '0'-'9' '.' '-']*


(***** nonlab -- entry point for non-label tokens  *****)
rule token = parse
|  [' ' '\009' '\012' '\n']+     
| "#!" [^ '\012' '\n']* ['\012' '\n']
    { token lexbuf }
| "*/" 
    { raise (Lexing(info lexbuf, "Unmatched end of comment")) }
| "/*" 
    { commentDepth := 1; comment lexbuf; token lexbuf }
| "//" [^ '\n']* "\n"
    { token lexbuf }
| ['0'-'9']+
| '-'?['0'-'9']+
    { P.INTV(int_of_string (Lexing.lexeme lexbuf)) }
| float_literal
| '-'float_literal
    { P.FLOATV (float_of_string(Lexing.lexeme lexbuf)) }
| '\''  
    { resetStr(); char lexbuf }

| '<' tag_name
    { P.LSTARTTAG(drop_startchars 1 (Lexing.lexeme lexbuf)) }
| "</" tag_name
    { P.LENDTAG(drop_startchars 2 (Lexing.lexeme lexbuf)) }
| "regtype" 
    { P.REGTYPE }
| "import_dtd"
    { P.IMPORT_DTD }
| "prefixby" 
    { P.PREFIXBY }
| "xmlns"
    { P.XMLNS }
| "regusing"
   { P.REGUSING }
| "switch"
    { P.SWITCH }
| "match"
    { P.MATCH }
| "case"
    { P.CASE }
| "break"
    { P.BREAK }
| "default"
    { P.DEFAULT }
| "if" 
    { P.IF }
| "else" 
    { P.ELSE }
| "while"
    { P.WHILE }
| "foreach"
    { P.FOREACH }
| "in"
    { P.IN }
| "using"
    { P.USING }
| "namespace"
    { P.NAMESPACE }
| "xmlspace"
    { P.XMLSPACE }
| "class"
    { P.CLASS } 
| "interface"
    { P.INTERFACE } 
| "sealed"
    { P.SEALED }
| "virtual"
    { P.VIRTUAL }
| "override"
    { P.OVERRIDE }
| "static"
    { P.STATIC }
| "base"
    { P.BASE }
| "this"
    { P.THIS }
| "null"
    { P.NULL }
|  "return"
    { P.RETURN }
|  "new"
    { P.NEW }
| "void"
    { P.VOID }
| "int"
    { P.INT }
| "bool"
    { P.BOOL }
| "false"
    { P.BOOLV(false) }
| "true"
    { P.BOOLV(true) }
| "as"
    { P.AS }
|  "[["
    { P.LREGPAREN }
|   "]]"
    { P.RREGPAREN }
|  "<[["
    { P.LTYPESTAMPPAREN }
|   "]]>"
    { P.RTYPESTAMPPAREN }
|  "~("
    { P.LUNSAFECASTPAREN }
|   ")~"
    { P.RUNSAFECASTPAREN }
|  "<("
    { P.LSTARTLAB }
|  "</>"
    { P.ENDTREE }
|  "/>"
    { P.REMPTYELTAG }
| "(" 
    { P.LPAREN }
| ")" 
    { P.RPAREN }
| "[" 
    { P.LBRACKET }
| "]" 
    { P.RBRACKET }
| "{"
    { P.LBRACE }
| "}"
    { P.RBRACE }

| "!="
    { P.BANGEQ }
| "=="
    { P.EQEQ }
|  "!"
    { P.BANG }
|  "=" 
    { P.EQ }
|  "||"
    { P.BARBAR }
| "|"
    { P.BAR }
| "*" 
    { P.STAR }
| "+" 
    { P.PLUS }
| "-" 
    { P.MINUS }
| "/" 
    { P.SLASH }
| "%" 
    { P.PERCENT }
| ":" 
    { P.COLON }
| ";" 
    { P.SEMICOLON }
| "?" 
    { P.QUESTION }
| "," 
    { P.COMMA }
| "->" 
    { P.ARROW }
| "^"
    { P.HAT }
| "~"
    { P.TILDE }
| "@"
    { P.AT }
| "<"
    { P.LT }
| ">"
    { P.GT }
| "<="
    { P.LEQ }
| ">="
    { P.GEQ }
| "&&"
    { P.AMPAMP }
| "&"
    { P.AMPERSAND }
| "."
    { P.DOT }

| ['A'-'Z' 'a'-'z' '_']
  ['A'-'Z' 'a'-'z' '_' '0'-'9' '\'']*
    { P.NAME (Lexing.lexeme lexbuf) }

(* we allow identifiers with a '-' inside, but only as attribute names
 * such as "http-equiv" *)
    
| ['A'-'Z' 'a'-'z' '_']
  ['A'-'Z' 'a'-'z' '_' '0'-'9' '-' '\'']*
    { P.ATNAME (Lexing.lexeme lexbuf) }


| "\"" { resetStr(); string lexbuf }

| "`" { resetStr(); pcdata lexbuf }

| eof { P.EOF }

| _  { raise (Lexing(info lexbuf, "Illegal character")) }


and comment = parse
  "/*"
    { commentDepth := succ !commentDepth; comment lexbuf }
| "*/"
    { commentDepth := pred !commentDepth; 
      if !commentDepth > 0 then comment lexbuf }
| eof
    { raise (Lexing(info lexbuf, "Comment not terminated")) }
| [^ '\n']
    { comment lexbuf }
| "\n"
    { comment lexbuf }


and string = parse
  "\""  { P.STRINGV (getStr()) }
| "\\"  { addStr(escape_in_string lexbuf); string lexbuf }
| "\n"  { addStr '\n'; string lexbuf }
| eof   { raise (Lexing(info lexbuf, "String not terminated")) }
| _     { addStr (Lexing.lexeme_char lexbuf 0); string lexbuf }

and escape_in_string = parse
  'n'    { '\n' }
| 't'    { '\t' }
| '\\'   { '\\' }
| '"'    { '\034'  }
(* | '\''   { '\'' } *)
| ['0'-'9']['0'-'9']['0'-'9']
    {
      let x = int_of_string(Lexing.lexeme lexbuf) in
      if x > 255 then
        raise (Lexing(info lexbuf, "Illegal numeric escape in string literal"))
      else
        Char.chr x
    }
(* | [^ '"' '\\' 't' 'n' '\''] *)
| [^ '"' '\\' 't' 'n']
    { raise (Lexing(info lexbuf, "Illegal character constant in string")) }


and pcdata = parse
  '`'  { P.PCDATA (getStr()) }
| '\\' { addStr(escape_in_pcdata lexbuf); pcdata lexbuf }
| "\n"  { addStr '\n'; pcdata lexbuf }
| eof  { raise (Lexing(info lexbuf, "PCDATA not terminated")) }
| _    { addStr (Lexing.lexeme_char lexbuf 0); pcdata lexbuf }


and escape_in_pcdata = parse
| '\\'   { '\\' }
(* | '"'    { '\034'  }
   | '\''   { '\'' }  *)
| '`'   { '`' }
| ['0'-'9']['0'-'9']['0'-'9']
    {
      let x = int_of_string(Lexing.lexeme lexbuf) in
      if x > 255 then
        raise (Lexing(info lexbuf, "Illegal numeric escape in pcdata literal"))
      else
        Char.chr x
    }
(* | [^ '"' '\\' 't' 'n' '\'']  *)
| [^ '`' '\\' 't' 'n']
    { raise (Lexing(info lexbuf, "Illegal character constant in pcdata")) }



and char = parse
  '\''  
    {  let data = getStr() in
    match String.length data with
    | 1 -> (P.CHARV (String.get data 0))
    | 0 -> raise (Lexing(info lexbuf, "No content in char literal"))
    | _ -> raise (Lexing(info lexbuf, "More than 1 character in char literal"))
     } 
| "\\"  { addStr(escape_in_char lexbuf); char lexbuf }
| "\n"  { addStr '\n'; char lexbuf }
| eof   { raise (Lexing(info lexbuf, "Character not terminated")) }
| _     { addStr (Lexing.lexeme_char lexbuf 0); char lexbuf }

and escape_in_char = parse
  'n'    { '\n' }
| 't'    { '\t' }
| '\\'   { '\\' }
| '\''    { '\034'  }
| ['0'-'9']['0'-'9']['0'-'9']
    {
      let x = int_of_string(Lexing.lexeme lexbuf) in
      if x > 255 then
        raise (Lexing(info lexbuf, "Illegal numeric escape in char literal"))
      else
        Char.chr x
    }
(* | [^ '"' '\\' 't' 'n' '\''] *)
| [^ '\'' '\\' 't' 'n']
    { raise (Lexing(info lexbuf, 
		    "Illegal character constant in character literal")) }
