/*  
 *  Yacc grammar for the parser.  The files parser.mli and parser.ml
 *  are generated automatically from parser.mly.
 */

/* stse 2003-0709 todo: 
   - array declaration, creation, access
   - conditional operator: (b ? x : y)
   - statement_without_trailer: switch, do, break, continue, synchronized, throw, try
   - statement_noshortif: labeled, for, while

   bug: while/break
*/

%{

module SS = SynSource
module SO = SynOp

let info datum =
  let inf = Finfo.make_finfo 
    (!ParseGlobals.input_file) (Parsing.symbol_start()) (Parsing.symbol_end())
  in (inf, datum)

let info_rhs s e datum = 
  let inf = Finfo.make_finfo 
      (!ParseGlobals.input_file) (Parsing.rhs_start s) (Parsing.rhs_end e)
  in (inf, datum)

let info_bogus datum = 
  let inf = Finfo.make_finfo (!ParseGlobals.input_file) (-1) (-1)
  in (inf, datum)

let attsty_any = info_bogus (SS.TAttAny)

let mkName symbolPos symbolYield = 
  info_rhs symbolPos symbolPos symbolYield


let mkDecName (name : SS.name) = name

(* !!! take care when fixing Ident related bugs in variable declaration
   and alternative patterns !!!
   See ident.ml for how bogus is used in function name *)

let mkBogusDecName () = (Finfo.bogus (), Ident.bogus_name)

let mkRefName (longname : SS.longname) = longname
let mkOneName (name : SS.name) = name

(*??
let rec array_spec_from_dims tspec = function
  | 1 -> SS.TArray(tspec)
  | n when (n > 1) -> array_spec_from_dims (info_bogus(SS.TArray tspec)) (n-1)
  | _ -> assert false
*)

(* transforms a string into a sequence of leaf trees in _reversed_ order *)
let pcdata (char_handler: char -> 'a) (literal : string) : 'a list = 
  let len = String.length literal in
  let rec go pchars ind = 
    if (ind < len) then
      let ch = String.get literal ind in
      let pch = char_handler ch
      in go (pch::pchars) (ind+1)
    else
      pchars
  in
  (go [] 0)

let pcdata_exp (chars : string) =
  if !Pref.pcdataencoding <> "char" then
    [info (SS.EPcdata chars)]
  else 
    pcdata (fun c -> info_bogus(SS.ETree(info_bogus (SS.Echar c), [], []))) chars

let pcdata_typ = 
  pcdata (fun c -> info_bogus(SS.TTree(
			      info_bogus(SS.TLab(info_bogus (SS.Tchar c))), 
			      info_bogus (SS.PAtt([],None)),
			      info_bogus (SS.TCat []))) )

%}

%token USING NAMESPACE XMLSPACE CLASS INTERFACE 
%token REGTYPE IMPORT_DTD PREFIXBY XMLNS REGUSING
%token SEALED VIRTUAL OVERRIDE STATIC
%token VOID BOOL INT
%token DOT COLON SEMICOLON COMMA
%token EQ
%token IF ELSE RETURN NEW
%token WHILE FOREACH IN
%token SWITCH MATCH CASE BREAK DEFAULT
%token LBRACE LBRACKET LPAREN
%token RBRACE RBRACKET RPAREN
%token THIS BASE NULL
%token BANG AMPAMP BARBAR    
%token MINUS PLUS STAR SLASH PERCENT 
%token GT LT GEQ LEQ EQEQ BANGEQ

%token <string> NAME ATNAME
%token <bool> BOOLV
%token <int> INTV
%token <char> CHARV
%token <string> STRINGV
%token <float> FLOATV
%token <string> PCDATA

%token LREGPAREN RREGPAREN
%token LTYPESTAMPPAREN RTYPESTAMPPAREN
%token LUNSAFECASTPAREN RUNSAFECASTPAREN

%token <string> LSTARTTAG
%token <string> LENDTAG
%token REMPTYELTAG

%token LSTARTLAB
%token ENDTREE

%token TILDE QUESTION BAR ARROW 
%token HAT AMPERSAND AT COERSE
%token AS

%token EOF


//%nonassoc  prec_let
//%right  prec_match
//%nonassoc  prec_coerse
%left   BAR
%left   AMPERSAND
%left   COMMA
%nonassoc QUESTION
%nonassoc AS
%left   BARBAR
%left   AMPAMP
%left   EQEQ BANGEQ
%left   LT GT LEQ GEQ
%left   PLUS MINUS      
%left   STAR SLASH PERCENT   // ops: mult, div, mod 
%nonassoc  UNARY    



%start top_unit
%type <SynSource.tunit> top_unit



%%


/*****  Identifiers   */

name:
  | NAME
      { mkName 1 $1 }
;

lname:
  | name
      { [$1] }
  | lname DOT name
      { $3 :: $1 } 
;

long_name:
  | lname 
      { (List.rev $1) }
;

long_name_comma_l:
  | long_name
       { [$1]}
  | long_name_comma_l COMMA long_name
       { $3 :: $1 }
;

long_name_comma_list:
  | long_name_comma_l
       { (List.rev $1)}

qual_name:
  | name
      { (None, $1) }
  | name COLON name 
      { (Some $1, $3) }

literal_val:
  | BOOLV     
      { info (SS.Ebool $1) }
  | INTV      
      { info (SS.Eint $1) }
  | CHARV
      { info (SS.Echar $1) }
  | STRINGV   
      { info (SS.Estring $1) }
;

literal_typ:
  | BOOLV     
      { info (SS.Tbool $1) }
  | INTV      
      { info (SS.Tint $1) }
  | CHARV
      { info (SS.Tchar $1) }
  | STRINGV   
      { info (SS.Tstring $1) }
;


/********************************************************************
     Java productions
*********************************************************************/
/* based on Java-1 LALR(1) grammar from 
   http://www.ugrad.cs.ubc.ca/~cs411/docs/modified/19.doc.html
*/


/********** Namespaces /aka modules, aka packages/  ***************/

/*** Using (import) directives ****/

using_directives:
  | using_dirs
      { List.rev $1 }
;

using_dirs:
  | using_directive
      { [$1] }
  | using_dirs using_directive
      { $2 :: $1 }
  ;

using_directive:
  | USING name EQ long_name SEMICOLON     // define a C# alias
      { info (SS.CAlias(mkDecName $2, mkRefName $4)) }
  | USING long_name SEMICOLON             // open a C# namespace 
      { info (SS.COpen(mkRefName $2)) }
  | XMLNS name EQ STRINGV SEMICOLON       // declare XML namespace prefix
      { info (SS.XPrefix(mkDecName $2, info_rhs 4 4 $4)) }
  | XMLNS EQ STRINGV SEMICOLON            // declare default XML namespace
      { info (SS.XOpen(info_rhs 3 3 $3)) }
  | REGUSING name EQ qual_name SEMICOLON  // define alias for regtype
      { info (SS.XAlias(mkDecName $2, $4)) }
;

/*** Top-level units (source file contents) ***/

top_unit:
  | using_directives  top_member_declarations
      { info {SS.tunit_usedirs = $1;  SS.tunit_decls = $2;}  
      } 
  |                  top_member_declarations
      { info {SS.tunit_usedirs = [];  SS.tunit_decls = $1;}
      }
  | /* empty */
    { info {SS.tunit_usedirs = [];  SS.tunit_decls = [];}
    } 
  ;

top_member_declarations:
  | top_member_declaration
      { [$1] }
  | top_member_declarations top_member_declaration
      { $2 :: $1 }
  | top_dtd_import
      { $1 }
  | top_member_declarations  top_dtd_import
      { $2 @ $1 }
;

top_member_declaration:
  | cspace_declaration
      { let (n,m) = $1 in (n, SS.TMemSpace m) }
  | class_declaration
      { let (n,m) = $1 in (n, SS.TMemClas m) }
  | interface_declaration
      { let (n,m) = $1 in (n, SS.TMemIntf m) }
  | xspace_declaration
      { let (n,m) = $1 in (n, SS.TMemXspace m) }
  | reg_type_declaration
      { let (n,m) = $1 in (n, SS.TMemSeq m) }
;

top_dtd_import:
  | dtd_import_declaration
      { List.map (fun (n,s) -> (n, SS.TMemSeq s)) $1 } 

/*** C# namespace declarations ***/

cspace_declaration:
  | NAMESPACE name cspace_body
      { (mkDecName $2, $3) }
;

cspace_body:
  | LBRACE cspace_unit RBRACE
      { $2 }

cspace_unit:
  | using_directives  cspace_member_declarations
      { info {SS.cunit_usedirs = $1;  SS.cunit_decls = $2;}  
      } 
  |                  cspace_member_declarations
      { info {SS.cunit_usedirs = [];  SS.cunit_decls = $1;}
      }
  | /* empty */
    { info {SS.cunit_usedirs = [];  SS.cunit_decls = [];}
    } 
  ;

cspace_member_declarations:
  | cspace_member_declaration
      { [$1] }
  | cspace_member_declarations cspace_member_declaration
      { $2 :: $1 }
;

cspace_member_declaration:
  | cspace_declaration
      { let (n,m) = $1 in (n, SS.CMemSpace m) }
  | class_declaration
      { let (n,m) = $1 in (n, SS.CMemClas m) }
  | interface_declaration
      { let (n,m) = $1 in (n, SS.CMemIntf m) }
;


/*** Regular (XML) namespace declarations ***/

xspace_declaration:
  | XMLSPACE STRINGV xspace_body
      { (mkName 2 $2, $3) }
;

xspace_body:
  | LBRACE xspace_unit RBRACE
      { $2 }

xspace_unit:
  | using_directives  xspace_member_declarations
      { info {SS.xunit_usedirs = $1;  SS.xunit_decls = $2;}  
      } 
  |                  xspace_member_declarations
      { info {SS.xunit_usedirs = [];  SS.xunit_decls = $1;}
      }
  | /* empty */
    { info {SS.xunit_usedirs = [];  SS.xunit_decls = [];}
    } 
  ;

xspace_member_declarations:
  | xspace_member_declaration
      { [$1] }
  | xspace_member_declarations xspace_member_declaration
      { $2 :: $1 }
  | xspace_dtd_import
      { $1 }
  | xspace_member_declarations  xspace_dtd_import
      { $2 @ $1 }
;

xspace_member_declaration:
  | reg_type_declaration
      { let (n,m) = $1 in (n, SS.XMemSeq m) }
;

xspace_dtd_import:
  | dtd_import_declaration
      { List.map (fun (n,s) -> (n, SS.XMemSeq s)) $1 } 



/*** old below ***/
/**--
compilation_unit:
  | using_directives  namespace_member_declarations
      { info {SS.cunit_usedirs = $1;  SS.cunit_decls = $2;}  
      } 
  |                  namespace_member_declarations
      { info {SS.cunit_usedirs = [];  SS.cunit_decls = $1;}
      }
  | // empty 
    { info {SS.cunit_usedirs = [];  SS.cunit_decls = [];}
    } 
  ;

namespace_member_declarations:
  | namespace_member_declaration
      { [$1] }
  | namespace_member_declarations namespace_member_declaration
      { $2 :: $1 }
  | dtd_import_declaration
      { $1 }
  | namespace_member_declarations  dtd_import_declaration
      { $2 @ $1 }
;

namespace_member_declaration:
  | namespace_declaration
      { $1 }
  | class_declaration
      { $1 }
  | interface_declaration
      { $1 }
  | reg_type_declaration
      { $1 }
;

namespace_declaration:
  | NAMESPACE name namespace_body
      { (mkDecName $2, SS.DefSpace $3) }
;

namespace_body:
  | LBRACE compilation_unit RBRACE
      { $2 }
*/



/*****  Types */

primitive_type:
  | BOOL  { info(SS.TBool) }   
  | INT   { info(SS.TInt) }   
;

defined_type:
  | long_name  { info (SS.TCSharp (mkRefName $1)) }
;

array_type:
  | primitive_type  LBRACKET RBRACKET  
      { info (SS.TArray (info_rhs 1 1 (SS.Thost $1))) }
  | long_name       LBRACKET RBRACKET  
      { info (SS.TArray (info_rhs 1 1 (SS.Thost (info_rhs 1 1 (SS.TCSharp (mkRefName $1))))) ) }
  | array_type      LBRACKET RBRACKET  
      { info (SS.TArray (info_rhs 1 1 (SS.Thost $1))) }
  | LPAREN reg_type_specifier RPAREN  LBRACKET RBRACKET  
      { info (SS.TArray ($2)) }
;

reference_type:
  | long_name     
      { info(SS.TCSharp (mkRefName $1)) }
  | array_type    { $1 }
;

item_type:
  | primitive_type 
      { $1 }
  | reference_type
      { $1 }
;
 
type_specifier:
  | item_type
      { info(SS.Thost $1)  }
//      { $1 }
  | reg_type_specifier
      { $1 }
;

/***** Patterns ******/

pattern_specifier:
  | literal_typ
      { info(SS.Thost $1)  }
  | item_type
      { info(SS.Thost $1)  }
//      { $1 }
  | reg_pattern
      { info(SS.Tseq $1) }
;

pattern:
  | literal_typ
      { info(SS.Thost $1)  }
  | item_type
      { info(SS.Thost $1)  }
  | item_type name
      { info(SS.PTyBind(mkDecName $2, info_rhs 2 2 (SS.Thost $1))) }
  | reg_pattern
      { info(SS.Tseq $1) }
  | reg_pattern name
      { info(SS.Tseq (info(SS.PSeqBind(mkDecName $2, $1)))) }
;

/***** Interface declarations */

interface_declaration:
  | INTERFACE name COLON long_name_comma_list interface_body
      {
        let intfname = mkDecName $2 in
	let intf = info_rhs 3 5 (Clas.mkIntf' intfname (List.map mkRefName $4) $5)
	in (intfname, intf) 
      }
  | INTERFACE name                            interface_body
      {
        let intfname = mkDecName $2 in
	let intf = info_rhs 3 3 (Clas.mkIntf' intfname [] $3)
	in (intfname, intf) 
      }
;
      
interface_body:
  | LBRACE interface_method_declarations RBRACE
      { $2 }    //!! keeping reversed order
  | LBRACE                           RBRACE
      { [] }
;

interface_method_declarations:
  | interface_method_declaration
      { [ $1 ] }
  | interface_method_declarations interface_method_declaration
      { $2 :: $1 }  //!! reversing 
;

interface_method_declaration:
  | type_specifier method_declarator SEMICOLON
      { let (mname, margs) = $2 in
        let (infty, ty) = $1 in
        (mname, info ((infty, Some ty), margs)) }
  | VOID           method_declarator SEMICOLON
      { let rettyp = info_rhs 1 1 (None) in
        let (mname, margs) = $2 in
	(mname, info (rettyp, margs))  }
;

/***** Class declarations */

class_declaration:
  | class_modifiers CLASS name COLON long_name_comma_list class_body
      { let clname = mkDecName $3 in
	let clas = info_rhs 4 6 (Clas.mkClas' $1 clname (List.map mkRefName $5) $6)
	in (clname, clas)  }
  | class_modifiers CLASS name                    class_body
      { let clname = mkDecName $3 in
	let object_loc = Parsing.rhs_end 3 in 
	let object_finfo = 
	  Finfo.make_finfo (!ParseGlobals.input_file) object_loc object_loc in
	let object_ref = [(object_finfo, "object")] in
	let clas = info_rhs 4 4 (Clas.mkClas' $1 clname [object_ref] $4) 
        in (clname, clas)  }
;

class_modifiers:
  | /* NOTHING */
    { [ ] }
  | class_modifier class_modifiers
    { $1 :: $2 }
;

class_modifier:
  | SEALED
    { Clas.Sealed }
;
    
class_body:
  | LBRACE class_member_declarations RBRACE
      { $2 }    //!! keeping reversed order
  | LBRACE                           RBRACE
      { [] }
;

class_member_declarations:
  | class_member_declaration
      { [ $1 ] }
  | class_member_declarations class_member_declaration
      { $2 :: $1 }  //!! reversing 
;

class_member_declaration:
  | fields_declaration
      { info (Clas.CmemFields $1) }
  | method_declaration
      { info (Clas.CmemMethod $1) }
  | constructor_declaration
      { info (Clas.CmemConstructor $1) }
;

fields_declaration:
  | type_specifier variable_declarators SEMICOLON
      { ($2, SO.Instance, $1) }  //!! keeping vars $2 reversed
  | STATIC type_specifier variable_declarators SEMICOLON
      { ($3, SO.Static, $2) }  //!! keeping vars $2 reversed
;

variable_declarators: 
  | variable_declarator 
      { [$1] }
  | variable_declarators COMMA variable_declarator 
      { $3 :: $1 }  //!! reversing
;

variable_declarator:
  | name
      { mkDecName $1, None }
  | name EQ expression
      { mkDecName $1, Some $3 }
;


method_declaration:
  | method_modifier type_specifier method_declarator method_body
      { let (mname, margs) = $3 in
        let (infty, ty) = $2 in
        info (mname, None, SO.Instance, $1, (infty, Some ty), margs, $4) }
  | method_modifier VOID           method_declarator method_body
      { let rettyp = info_rhs 2 2 (None) in
        let (mname, margs) = $3 in
	info (mname, None, SO.Instance, $1, rettyp, margs, $4)  }
  | STATIC type_specifier method_declarator method_body
      { let (mname, margs) = $3 in
        let (infty, ty) = $2 in
        info (mname, None, SO.Static, (info_rhs 1 1 SO.Normal), (infty, Some ty), margs, $4) }
  | STATIC VOID           method_declarator method_body
      { let rettyp = info_rhs 2 2 (None) in
        let (mname, margs) = $3 in
	info (mname, None, SO.Static, (info_rhs 1 1 SO.Normal), rettyp, margs, $4)  }
  | type_specifier method_declarator method_body
      { let (mname, margs) = $2 in
        let (infty, ty) = $1 in
        info (mname, None, SO.Instance, (info_rhs 1 1 SO.Normal), (infty, Some ty), margs, $3) }
  | VOID           method_declarator method_body
      { let rettyp = info_rhs 1 1 (None) in
        let (mname, margs) = $2 in
	info (mname, None, SO.Instance, (info_rhs 1 1 SO.Normal), rettyp, margs, $3)  }
;

method_modifier:
  | VIRTUAL
      { info SO.Virtual }
  | OVERRIDE
      { info SO.Override }
  | NEW
      { info SO.New }
;

method_declarator:
  | name LPAREN parameter_list RPAREN 
      { (mkDecName $1, $3)}
  | name LPAREN                RPAREN 
      { (mkDecName $1, []) }
;

parameter_list:
  | parameterlist
      { List.rev $1 } //!! reversing back

parameterlist:
  | parameter 
      { [$1] }
  | parameterlist COMMA parameter 
      { $3 :: $1 }  //!! reversing 
;

parameter:
  | type_specifier
      { (None, $1) }
  | type_specifier name
      { (Some (mkDecName $2), $1) }
;

method_body:
  | SEMICOLON
      { None }
  | block 
      { Some $1 }
/*??
  | LSEQPAR x_expression RSEQPAR  //sugar for a "pure XDuce" method
      { 
	info (SS.SBlock 
		[info_bogus (SS.SReturn(Some(info (SS.EXduce $2))))
	       ])  }
*/
;

constructor_declaration:
  | constructor_declarator constructor_initializer constructor_body
      { let (hostclassname, params) = $1 in
        info (hostclassname, params, $2, $3) }
;

constructor_declarator: 
  | name LPAREN parameter_list RPAREN 
      { (mkOneName $1, $3) }
  | name LPAREN RPAREN 
      { (mkOneName $1, []) }
;

constructor_initializer:
  | COLON THIS  LPAREN argument_list RPAREN 
      { (info (SS.InvkThis $4)) }
  | COLON THIS  LPAREN               RPAREN    
      { (info (SS.InvkThis [])) }
  | COLON BASE LPAREN argument_list RPAREN    
      { info (SS.InvkSuper $4) }
  | COLON BASE LPAREN               RPAREN    
      { info (SS.InvkSuper []) }
  | //empty//
      { info_bogus SS.InvkNone }
;

constructor_body:
  | SEMICOLON
      { None }
  | block
      { Some  $1 }
;


/***** Statements */

block:
  | LBRACE block_statements RBRACE
      { info ($2) }  
  | LBRACE            RBRACE
      { info ([]) }
;

block_statements:
  | block_stss
      { List.rev $1 }
;

block_stss:
  | block_statement
      { [$1] }
  | block_stss block_statement
      { $2 :: $1}
;

block_statement:
  | local_var_declaration_statement
      { $1 }
  | statement
      { $1 }  
;

local_var_declaration_statement:
  | local_var_declaration  SEMICOLON
      { info(SS.SLocVars $1) }
;

local_var_declaration:
  | reg_pattern EQ expression,
      { ([mkBogusDecName (), Some $3], info_rhs 1 1 (SS.Tseq $1)) }
  | pattern_specifier variable_declarators
      { (List.rev $2, $1) }
;

statement:
  | statement_without_trailer  { $1 }
  | if_then_statement          { $1 }
  | if_then_else_statement     { $1 }
  | WHILE LPAREN expression RPAREN statement 
      { info (SS.SWhile($3, info_rhs 5 5 [$5]) )  }
  | FOREACH LPAREN reg_pattern IN expression RPAREN statement
      { info (SS.SForeachRE($3, $5, info_rhs 7 7 [$7])) }
  // other: for, labeled
;

statement_without_trailer:
  | block                 { info (SS.SBlock $1) }
  | empty_statement       { $1 }
  | expression_statement  { $1 }
  | return_statement      { $1 }
  | match_statement       { $1 }
  | break_statement       { $1 }
  //other: switch, do, break, continue, synchronized, throw, try
;

statement_noshortif:
  | statement_without_trailer   { $1 }
  | if_then_else_noshortif     	{ $1 }
  | WHILE LPAREN expression RPAREN statement_noshortif 
      { info (SS.SWhile($3, info_rhs 5 5 [$5]) ) }
  //other noshortif: labeled, for, while
;


empty_statement:
  | SEMICOLON 
      { info(SS.SEmpty) }
;

expression_statement:
  | statement_expression SEMICOLON   
//      { $1 }
      { info(SS.SExp $1) }
;

statement_expression:
  | assignment   { $1 }
  //## pre/post-inc/dec exprs  ??
  | method_invocation   { $1 }
  | instance_creation   { $1 }
;

if_then_statement:
  | IF LPAREN expression RPAREN statement   
      { info (SS.SIf($3, info_rhs 5 5 [$5], info_bogus([]))) }
;

if_then_else_statement:
  | IF LPAREN expression RPAREN statement_noshortif ELSE statement   
      { info (SS.SIf ($3, info_rhs 5 5 [$5], info_rhs 7 7 [$7])) }
;

if_then_else_noshortif:
  | IF LPAREN expression RPAREN statement_noshortif ELSE statement_noshortif
      { info (SS.SIf ($3, info_rhs 5 5 [$5], info_rhs 7 7 [$7])) }
;

return_statement:
  | RETURN expression SEMICOLON 
      { let inf, ty = $2 in
        info(SS.SReturn (inf, Some ty)) }
  | RETURN            SEMICOLON 
      { info(SS.SReturn (info_rhs 1 1 None)) }
;

break_statement:
  | BREAK SEMICOLON 
      { info(SS.SBreak) }
;

match_statement:
  | MATCH LPAREN expression RPAREN match_block
      { info(SS.SMatch($3, $5)) }
;

match_block:
  | LBRACE match_sections RBRACE
      { List.rev $2 }
;

match_sections:
  | match_section
      { [$1] }
  | match_sections match_section
      { $2 :: $1 }
;

match_section:
  | match_label block_statements
      { info($1, info_rhs 2 2 ($2))  }
;

match_label:
  | CASE pattern COLON
      { $2 }
;


/***** Expressions */

argument_list:
  | argumentlist
      { List.rev $1 }

argumentlist:
  | expression 
      { [$1] }
  | argumentlist COMMA expression 
      { $3 :: $1 }
;

primary:
    | primary_nonewarray  {$1}
    | array_creation      {$1}

primary_nonewarray:
  | semi_atomic_expression               { $1 }
  | instance_creation               { $1 }
  | array_access                    { $1 }
;


semi_atomic_expression:
  | NULL                            { info (SS.Enull) }
  | literal_val                     { $1 }
  | THIS                            { info(SS.EThis) }
  | LPAREN expression_nonreg RPAREN        { $2 }
  | field_access                    { $1 }
  | method_invocation               { $1 }
;

atomic_expression:
  | semi_atomic_expression     { $1 }
  | long_name                  { info(SS.ENamedLoc (mkRefName $1)) }
  | reg_expression          { $1 }
;

instance_creation:
  | NEW long_name LPAREN argument_list RPAREN   
      { info(SS.ENew(mkRefName $2, $4)) }
  | NEW long_name LPAREN               RPAREN   
      { info(SS.ENew(mkRefName $2, [])) }
;

array_creation:
  | NEW primitive_type dim_expr dims   
      { info(SS.ENewArray(info_rhs 2 2 (SS.Thost $2), $3, $4)) }
  | NEW primitive_type dim_expr        
      { info(SS.ENewArray(info_rhs 2 2 (SS.Thost $2), $3, 0)) }
  | NEW defined_type   dim_expr dims   
      { info(SS.ENewArray(info_rhs 2 2 (SS.Thost $2), $3, $4)) }
  | NEW defined_type   dim_expr        
      {	info(SS.ENewArray(info_rhs 2 2 (SS.Thost $2), $3, 0)) }
  | NEW LPAREN reg_type_specifier RPAREN dim_expr dims   
      { info(SS.ENewArray($3, $5, $6)) }
  | NEW LPAREN reg_type_specifier RPAREN dim_expr        
      {	info(SS.ENewArray($3, $5, 0)) }
;

dim_expr:
  | LBRACKET expression RBRACKET  { $2 }
;

dims:
  | LBRACKET RBRACKET        {1}
  | dims LBRACKET RBRACKET   { $1 + 1}
;

field_access:
  | primary DOT name 
      { info(SS.EFieldAccess($1, mkOneName $3)) }
//  | reg_expression DOT name    //gives conflicts
//      { info(SS.EFieldAccess($1, mkOneName $3)) }
  | BASE   DOT name  
      { info(SS.EFieldAccessSuper(mkOneName $3)) }
;

method_invocation:
  | long_name         invocation_arguments 
      { match Fjxutil.split_last_elem $1 with 
        | ([], methname) -> (* "local" method *)
	    info (SS.EMethInvkLocal(mkOneName methname, $2))
	| (prefix, methname) -> (* method on an object or class *)
	    let host = info_rhs 1 1 (SS.ENamedLoc (mkRefName prefix)) in
	    info(SS.EMethInvkBase(host, mkOneName methname, $2)) }
  | primary DOT name  invocation_arguments 
      { info(SS.EMethInvkBase($1, mkOneName $3, $4)) }
//  | reg_expression DOT name  invocation_arguments   //gives conflicts
//      { info(SS.EMethInvkBase($1, mkOneName $3, $4)) }
  | BASE   DOT name  invocation_arguments 
      { info(SS.EMethInvkSuper(mkOneName $3, $4)) }
;

invocation_arguments:
  | LPAREN argument_list RPAREN { $2 }
  | LPAREN               RPAREN { [] }
;


array_access:
  | long_name LBRACKET expression RBRACKET   
      { info(SS.EArrayAccess(info_rhs 1 1 (SS.ENamedLoc (mkRefName $1)), $3)) }
  | primary_nonewarray   LBRACKET expression RBRACKET   
      { info(SS.EArrayAccess($1, $3)) }
;

//++++++++++++++++  Expressions core start +++++++++++++++

//Excluded forms: pre/post increment/decrement (++, --), 
//    bitwise ops (^, |, &), shift ops (<<, ...), 
//    conditional expression (.. ? .. : ..),
//     incremental assignments (+=, *=, ...)
//Unary +/- must be parenthesised: (- 5)

unary_expression:
  | primary          { $1 }
  | long_name         { info(SS.ENamedLoc (mkRefName $1)) }
  | cast_expression   { $1 }
  | typestamp_expression { $1 }
  | BANG unary_expression   { info(SS.EUnary (SO.UnNot, $2)) }    
//  |  PLUS unary_expression   %prec UNARY  {1}    //still gives a conflict...
//  |  MINUS unary_expression  %prec UNARY {1}   
  | LPAREN PLUS unary_expression  RPAREN  
      { info (SS.EUnary(SO.UnPlus, $3)) }    
  | LPAREN MINUS unary_expression RPAREN  
      { info (SS.EUnary(SO.UnMinus, $3)) }   
;

cast_expression:
  | LPAREN primitive_type       RPAREN unary_expression   
      { info(SS.ECast(info_rhs 2 2 (SS.Thost $2), $4, true))  }
  | LPAREN reg_type_specifier   RPAREN unary_expression 
      { info(SS.ECast($2, $4, true)) }
  | LUNSAFECASTPAREN reg_type_specifier RUNSAFECASTPAREN unary_expression 
      { info(SS.ECast($2, $4, false)) }
  | LPAREN expression_nonreg           RPAREN unary_expression   
      { match $2 with
      |	(inf, SS.ENamedLoc n) ->
	  let typ = (inf, SS.Thost(inf, SS.TCSharp n)) in
	  info (SS.ECast(typ, $4, true))
      |	_ -> raise Parsing.Parse_error
     } 
    //!! POST-PARSE: whether typ really refers to a type
;

typestamp_expression:
  | LTYPESTAMPPAREN regtype_sequence RTYPESTAMPPAREN unary_expression
      { info(SS.ETypeStamp($2, $4)) }

binary_expression:                //see priorities for binary symbols!
  | unary_expression   {$1}
  | binary_expression STAR binary_expression   
      { info(SS.EBinary($1, SO.BiMult, $3))  }
  | binary_expression SLASH binary_expression     
      { info(SS.EBinary($1, SO.BiDiv, $3))  }
  | binary_expression PERCENT binary_expression     
      { info(SS.EBinary($1, SO.BiMod, $3))  }
  | binary_expression PLUS binary_expression     
      { info(SS.EBinary($1, SO.BiPlus, $3))  }
  | binary_expression MINUS binary_expression     
      { info(SS.EBinary($1, SO.BiMinus, $3))  }
  | binary_expression LT binary_expression     
      { info(SS.EBinary($1, SO.BiLt, $3))  }
  | binary_expression GT binary_expression     
      { info(SS.EBinary($1, SO.BiGt, $3))  }
  | binary_expression LEQ binary_expression     
      { info(SS.EBinary($1, SO.BiLeq, $3))  }
  | binary_expression GEQ binary_expression     
      { info(SS.EBinary($1, SO.BiGeq, $3))  }
  | binary_expression EQEQ binary_expression     
      { info(SS.EBinary($1, SO.BiEq, $3))  }
  | binary_expression BANGEQ binary_expression     
      { info(SS.EBinary($1, SO.BiNeq, $3))  }
  | binary_expression AMPAMP binary_expression     
      { info(SS.EBinary($1, SO.BiAnd, $3))  }
  | binary_expression BARBAR binary_expression     
      { info(SS.EBinary($1, SO.BiOr, $3))  }
;


core_expression:
  | binary_expression { $1 }  //##
;

//++++++++++++++++  Expressions core end +++++++++++++++

assignment_expression:
  | core_expression   { $1 }  
  | assignment        { $1 }
;

assignment:
  | left_hand_side EQ assignment_expression  
      { info(SS.EAssign($1, $3)) }
  | left_hand_side EQ reg_expression
      { info(SS.EAssign($1, $3)) }
;

left_hand_side:
  | long_name     { info(SS.ENamedLoc (mkRefName $1)) }
  | field_access  { $1 }
  | array_access  { $1 }
;


expression:
  | expression_nonreg  { $1 }
  | reg_expression     { $1 }

expression_nonreg:
  | assignment_expression { $1 }
;


/********************************************************************
     XDuce productions
*********************************************************************/

dtd_import_declaration: 
  | IMPORT_DTD STRINGV SEMICOLON
      { (Parsedtd.parse_dtd "" $2)   }  
  | IMPORT_DTD STRINGV PREFIXBY STRINGV SEMICOLON
      { (Parsedtd.parse_dtd $4 $2)   }  
;

reg_type_declaration:
  | REGTYPE name LREGPAREN regtype_sequence RREGPAREN
      { (mkDecName $2, $4) } 
;

reg_type_specifier:
  | LREGPAREN regtype_sequence RREGPAREN
      { info (SS.Tseq $2)  }

label_type:
  | item_type
      { info(SS.TLab $1) }
  | literal_typ
      { info(SS.TLab $1) }
;

regtype_sequence:
  | regtype_seq 
      { info (SS.TCat (List.rev $1)) }
  | /* empty */
      { info (SS.TCat []) }
;

regtype_seq:
  | regtype
      { [$1] }
  | regtype_seq COMMA regtype 
      { $3 :: $1 }
  | regtypetree_seq treetype 
      { $2 :: $1 }
;

regtypetree_seq:
  | treetype
      { [$1] }
  | regtype_seq COMMA treetype 
      { $3 :: $1 }
  | regtypetree_seq treetype 
      { $2 :: $1 }
;

regtype:
  | qual_name
      { info (SS.TSeqRef $1) }
//      { info (SS.TSeqRef (mkRefName $1)) }
  | treetype
      { $1 }
  | PCDATA
      { info (SS.TCat (List.rev(pcdata_typ $1))) }  
                       //after rev the order is original
  | LPAREN regtype_sequence RPAREN
      { $2 }
  | regtype BAR regtype
      { info (SS.TAlt([$1;$3])) }
  | regtype STAR
      { info (SS.TRep $1) }
  | regtype PLUS
      { info(SS.TCat[$1; info(SS.TRep $1)]) }
  | regtype QUESTION
      { info(SS.TAlt[$1; info(SS.TCat[])]) }
;

treetype:
  | LSTARTLAB label_type RPAREN GT regtype_sequence ENDTREE
      { info (SS.TTree($2, attsty_any, $5)) }
  | LSTARTLAB label_type RPAREN REMPTYELTAG
      { info (SS.TTree($2, attsty_any,  info_bogus(SS.TCat[]))) }
  | LSTARTTAG GT regtype_sequence ENDTREE
      { let lty = info_rhs 1 1 (SS.TLab (info_rhs 1 1 (SS.Ttag $1))) in
	info (SS.TTree(lty, attsty_any, $3))   
                                                     (*##ATT *)
      } 
  | LSTARTTAG GT regtype_sequence LENDTAG GT
      { match compare $1 $4 with 
      |	0 -> 
	  let lty = info_rhs 1 1 (SS.TLab (info_rhs 1 1 (SS.Ttag $1))) in
	  info (SS.TTree(lty, attsty_any, $3))  
                                                       (*##ATT *)
      |	_ -> raise Parsing.Parse_error
     } 
  | LSTARTTAG REMPTYELTAG
      { let lty = info_rhs 1 1 (SS.TLab (info_rhs 1 1 (SS.Ttag $1))) in
	info (SS.TTree(lty, attsty_any, info_bogus(SS.TCat[])))  } 
;


/******** Reg patterns *****/

reg_pattern:
  | LREGPAREN regpat_sequence RREGPAREN
      { $2 }
;

label_pattern:        //no reg patterns in labels now
  | literal_typ
      { info (SS.TLab $1) }
  | item_type
      { info (SS.TLab $1) }
  | item_type name
      { let lty = info_rhs 1 1 (SS.TLab $1) in
	info(SS.PLabBind(mkDecName $2, lty)) }


regpat_sequence:
  | regpat_seq
      { info (SS.TCat (List.rev $1)) }
  | /* empty e */
    { info (SS.TCat []) }
;

regpat_seq:
  | bindpat
      { [$1] }
  | regpat_seq COMMA bindpat
      { $3 :: $1 }
  | regpattree_seq treepat 
      { $2 :: $1 }
;

regpattree_seq:
  | treepat
      { [$1] }
  | regpat_seq COMMA treepat 
      { $3 :: $1 }
  | regpattree_seq treepat 
      { $2 :: $1 }
;

bindpat:
  | regpat
      { info(SS.TCat $1) }
  | regpat name
      { info(SS.PSeqBind(mkDecName $2, info_rhs 1 1 (SS.TCat $1))) }
  | bindpat BAR bindpat
      { info (SS.TAlt([$1;$3])) }
;

regpat:
  | qual_name
      { [info (SS.TSeqRef $1)] }
//--      { [info (SS.TSeqRef (mkRefName $1))] }
  | treepat
      { [$1] }
  | PCDATA
      { List.rev(pcdata_typ $1) }  //after rev the order is original
  | LPAREN regpat_sequence RPAREN
      { [ $2 ] }
  | bindpat STAR
      { [ info (SS.TRep $1) ] }
  | bindpat PLUS
      { [ info(SS.TCat[$1; info(SS.TRep $1)]) ] }
  | bindpat QUESTION
      { [ info(SS.TAlt[$1; info(SS.TCat[])]) ] }


treepat:
  | LSTARTLAB label_pattern RPAREN attrpats GT regpat_sequence ENDTREE
    { info(SS.TTree($2, $4, $6)) } 
  | LSTARTLAB label_pattern RPAREN attrpats REMPTYELTAG
      { info (SS.TTree($2, $4, info_bogus(SS.TCat[]))) }
  | LSTARTTAG attrpats GT regpat_sequence ENDTREE
      { let lpa = info_rhs 1 1 (SS.TLab (info_rhs 1 1 (SS.Ttag $1))) in
	info (SS.TTree(lpa ,$2, $4))   } 
  | LSTARTTAG attrpats GT regpat_sequence LENDTAG GT
      { match compare $1 $5 with 
      |	0 -> 
	  let lpa = info_rhs 1 1 (SS.TLab (info_rhs 1 1 (SS.Ttag $1))) in
	  info (SS.TTree(lpa ,$2, $4))
      |	_ -> raise Parsing.Parse_error
     } 
  | LSTARTTAG attrpats REMPTYELTAG
      { let lpa = info_rhs 1 1 (SS.TLab (info_rhs 1 1 (SS.Ttag $1))) in
	info (SS.TTree(lpa, $2, info_bogus(SS.TCat[])))    } 
;

attrpats:
  | /* empty */                /* any attributes will be accepted */
    { info(SS.TAttAny) }
  | AT LPAREN RPAREN           /* absence of attributes is required */
      {info (SS.PAtt([], None)) }
  | attpats               /* there must be these named attrs, and only them */
      { info (SS.PAtt($1, None)) }
  | attpats AT LPAREN name RPAREN  /* these named attrs, and possibly more */
      { info (SS.PAtt($1, Some (mkDecName $4))) }
  | AT LPAREN name RPAREN      /* just catch all attrs as a sequence */
      { info (SS.PAtt([], Some (mkDecName $3))) }

attpats:
  | attrpat
    { [$1] }
  | attpats attrpat
    { $2 :: $1 }

attrpat:
  | NAME EQ name
    { info(mkName 1 $1, mkDecName $3) }


/******** Reg expressions ****/

reg_expression:
  | LREGPAREN reg_exp_unprotected RREGPAREN
      { info (SS.ESeq $2) } 
;

reg_exp_unprotected:
  | tree_expr_sequence   { $1 }
;

attrexps:
  | /* empty */
    { [] }
  | attrexps attrexp
    { $1 @ $2 }
;

attrexp:
  | NAME EQ treeexp_seq
      //single attribute as a string value to be coersed into pcdata
      { [info(SS.ETree(info_rhs 1 1 (SS.Etag $1), 
		       [], List.rev $3))] }
  | ATNAME EQ treeexp_seq
      //single attribute as a string value to be coersed into pcdata
      { [info(SS.ETree(info_rhs 1 1 (SS.Etag $1), 
		       [], List.rev $3))] }
//  | NAME EQ PCDATA
//      //single attribute as a pcdata literal
//      { [info(SS.ETree(info_rhs 1 1 (SS.Etag $1), 
//		       [], List.rev (pcdata_exp $3)))] }
  | AT LPAREN reg_exp_unprotected RPAREN   
      //a sequence to be treated attributively
      { $3 }
;

tree_expr_sequence:
  | /* empty */
    { [] }
  | treeexp_seq
      { List.rev $1 }  
;

treeexp_seq:
  | pcdataexp
      { $1 }
  | treeexp_seq COMMA pcdataexp
      { $3 @ $1 }
  | treeexptree_seq treeexp
      { $2 :: $1 }
;

treeexptree_seq:
  | pcdataexp
      { $1 }
  | treeexp_seq COMMA pcdataexp
      { $3 @ $1 }
  | treeexptree_seq treeexp 
      { $2 :: $1 }
;

pcdataexp:
  | treeexp
      { [$1] }
  | PCDATA
      { pcdata_exp $1 }
  | atomic_expression
      { [$1] }
;

treeexp:
  | LSTARTLAB expression RPAREN attrexps GT tree_expr_sequence ENDTREE
       { info (SS.ETree($2, $4, $6)) } 
  | LSTARTLAB expression RPAREN attrexps REMPTYELTAG
      { info (SS.ETree($2, $4, [])) }
  | LSTARTTAG attrexps GT tree_expr_sequence ENDTREE
      { info (SS.ETree(info_rhs 1 1 (SS.Etag $1), $2, $4))   } 
  | LSTARTTAG attrexps GT tree_expr_sequence LENDTAG GT
      { match compare $1 $5 with 
      |	0 -> info (SS.ETree(info_rhs 1 1 (SS.Etag $1), $2, $4))
      |	_ -> raise Parsing.Parse_error
     } 
  | LSTARTTAG attrexps REMPTYELTAG
      { info (SS.ETree(info_rhs 1 1 (SS.Etag $1),$2, []))    } 
;



/***********************************************************************
                     Xti files
***********************************************************************/
/*--VG??
xti_top_unit:
  | // empty 
    { info {SS.xti_tunit_usedirs = [];  SS.xti_tunit_decls = [];}
    } 

top_unit:
  | using_directives  top_member_declarations
      { info {SS.tunit_usedirs = $1;  SS.tunit_decls = $2;}  
      } 
  |                  top_member_declarations
      { info {SS.tunit_usedirs = [];  SS.tunit_decls = $1;}
      }
  | // empty 
    { info {SS.tunit_usedirs = [];  SS.tunit_decls = [];}
    } 
  ;

top_member_declarations:
  | top_member_declaration
      { [$1] }
  | top_member_declarations top_member_declaration
      { $2 :: $1 }
  | top_dtd_import
      { $1 }
  | top_member_declarations  top_dtd_import
      { $2 @ $1 }
;

top_member_declaration:
  | cspace_declaration
      { let (n,m) = $1 in (n, SS.TMemSpace m) }
  | class_declaration
      { let (n,m) = $1 in (n, SS.TMemClas m) }
  | interface_declaration
      { let (n,m) = $1 in (n, SS.TMemIntf m) }
  | xspace_declaration
      { let (n,m) = $1 in (n, SS.TMemXspace m) }
  | reg_type_declaration
      { let (n,m) = $1 in (n, SS.TMemSeq m) }
;

top_dtd_import:
  | dtd_import_declaration
      { List.map (fun (n,s) -> (n, SS.TMemSeq s)) $1 } 



cspace_declaration:
  | NAMESPACE name cspace_body
      { (mkDecName $2, $3) }
;

cspace_body:
  | LBRACE cspace_unit RBRACE
      { $2 }

cspace_unit:
  | using_directives  cspace_member_declarations
      { info {SS.cunit_usedirs = $1;  SS.cunit_decls = $2;}  
      } 
  |                  cspace_member_declarations
      { info {SS.cunit_usedirs = [];  SS.cunit_decls = $1;}
      }
  | // empty
    { info {SS.cunit_usedirs = [];  SS.cunit_decls = [];}
    } 
  ;

cspace_member_declarations:
  | cspace_member_declaration
      { [$1] }
  | cspace_member_declarations cspace_member_declaration
      { $2 :: $1 }
;

cspace_member_declaration:
  | cspace_declaration
      { let (n,m) = $1 in (n, SS.CMemSpace m) }
  | class_declaration
      { let (n,m) = $1 in (n, SS.CMemClas m) }
  | interface_declaration
      { let (n,m) = $1 in (n, SS.CMemIntf m) }
;




xspace_declaration:
  | XMLSPACE STRINGV xspace_body
      { (mkName 2 $2, $3) }
;

xspace_body:
  | LBRACE xspace_unit RBRACE
      { $2 }

xspace_unit:
  | using_directives  xspace_member_declarations
      { info {SS.xunit_usedirs = $1;  SS.xunit_decls = $2;}  
      } 
  |                  xspace_member_declarations
      { info {SS.xunit_usedirs = [];  SS.xunit_decls = $1;}
      }
  | // empty
    { info {SS.xunit_usedirs = [];  SS.xunit_decls = [];}
    } 
  ;

xspace_member_declarations:
  | xspace_member_declaration
      { [$1] }
  | xspace_member_declarations xspace_member_declaration
      { $2 :: $1 }
  | xspace_dtd_import
      { $1 }
  | xspace_member_declarations  xspace_dtd_import
      { $2 @ $1 }
;

xspace_member_declaration:
  | reg_type_declaration
      { let (n,m) = $1 in (n, SS.XMemSeq m) }
;

xspace_dtd_import:
  | dtd_import_declaration
      { List.map (fun (n,s) -> (n, SS.XMemSeq s)) $1 } 

*/


%%
