open List

let rec mapcat ~f = function
    [] -> []
  | x :: xl -> (f x) @ (mapcat ~f xl)

let mem_pred ~(f:'a -> bool) : 'a list -> bool =
  let rec loop = function
    | [] -> false
    | el :: rest -> if f el then true else loop rest in
  loop

let mem_pred_opt ~(f:'a -> bool) : 'a list -> 'a list option =
  let rec loop = function
    | [] -> None
    | (el :: rest as ls) -> if f el then Some ls else loop rest in
  loop

let filter ~(f:'a -> bool) : 'a list -> 'a list =
  let rec loop = function
    | [] -> []
    | el :: rest ->
        if f el then el :: loop rest
        else loop rest in
  loop

let split_list ~(f:'a -> bool) : 'a list -> 'a list * 'a list =
  let rec loop = function
    | [] -> [],[]
    | el :: rest ->
        let ls1, ls2 = loop rest in
        if f el then el :: ls1, ls2
        else ls1, el :: ls2 in
  loop

let union_many ~f l =
  let rec union2 = function
      s1 :: s2 :: rest -> f s1 s2 :: union2 rest
    | l -> l in
  let rec union_all = function
      [] -> []
    | [s] -> s
    | l -> union_all (union2 l) in
  union_all l



let split_into_classes ~(cmp: 'a -> 'a -> int) (lst: 'a list) : 'a list list = 
  let sorted_lst = List.sort cmp lst in
  List.fold_right 
    (fun x prevres -> 
      match prevres with 
      |	[] -> [[x]]
      |	((y::_) as cls)::rest ->
	  (match cmp x y with
	  | 0 -> (x::cls)::rest
	  | _ -> [x]::prevres )
      |	[]::rest -> assert false  (* an empty class is never created *)
    ) sorted_lst []


let rec filtermap_opt ~(f: 'a -> 'b option) = function 
  | [] -> [] 
  | x::xs -> 
      (match f x with
      |	None -> filtermap_opt f xs
      |	Some y -> y :: (filtermap_opt f xs))
